/******************************************************************************
* GhostQ: A QBF solver with ghost variables.
* Version 0.70 (Apr 2012)
* Author: Will Klieber (Carnegie Mellon University)
* Note: Sorry for the current untidy nature of this file; I still need to
* clean up all the dead code and reorganize the code for clarity.
******************************************************************************/

#define AIX 0

/* Debugging breakpoints */
void raw_nop() {}
void (*volatile volnop)() = &raw_nop;  // For debugging; optimizer won't remove.

void rawstop() {}    // Place a breakpoint here with gdb.
void (*volatile stop)() = &rawstop;  // For debugging; optimizer won't remove.


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <errno.h>

#include <algorithm>
#include <map>
#include <set>
#include <vector>
#include <list>
#include <queue>
using namespace std;

#include <ext/hash_set>
using namespace __gnu_cxx;

#define SORT_NAME int
#define SORT_TYPE int
#define SORT_CMP(x, y) ((x) - (y))
#include "sort.h"

#include "quickselect.cpp"



typedef unsigned long long uint64;


struct eq_int {
  bool operator()(int s1, int s2) const {
    return s1 == s2;
  }
};




/*****************************************************************************
* Utility Macros.    }}}{{{1
*****************************************************************************/

#define elif else if

#define auto(var, init) typeof(init) var = (init)

#define foreach(var, container) \
  for (auto(var, (container).begin());  var != (container).end();  ++var)


#define foreach_rev_begin(var, container) { \
    auto(var ## _BEGIN, (container).begin()); \
    auto(var, (container).end()); \
    if (var != var##_BEGIN) { \
        --var; \
        do {

#define foreach_rev_end(var) \
        } while ((var != var##_BEGIN) && ({--var; 1;}));\
    } \
}
            

#define printferr(fmt,args...) \
    ({fflush(stdout); fprintf(stderr, fmt, ##args);})

#define talloc(type, n) ((type*) calloc(n, sizeof(type)))

// Unicode character 0x0394 (GREEK CAPITAL LETTER DELTA)
#define UTF8_DELTA "\xCE\x94"



#if AIX

    #define getline my_getline

    /* getline.c */
    /* Copyright (C) 2004 Andy Goth <unununium@openverse.com> */
    /* Distributed under GNU GPL 2.0 or (at your option) any later version */
    ssize_t getline(char** buf, size_t* len, FILE* stream)
    {
        int i = 0, new_len, my_malloc = 0;
        char* nl, *new_buf;

        if (buf == NULL || len == NULL) {
            errno = EINVAL;
            return -1;
        }

        if (*buf == NULL || *len == 0) {
            *buf = NULL; *len = 0;
            my_malloc = 1;
        }

        if (*len <= 60) goto alloc;

        while (1) {
            if (fgets(*buf + i, *len - i, stream) == NULL) {
                if (!feof(stream) || i == 0) {
                    /* The read failed with an error, or the file is empty. */
                    goto error;
                } else {
                    /* The final line contains no newline, and the previous fgets()
                     * read exactly as many characters as remained in the line. */
                    return i;
                }
            }

            if (feof(stream)) {
                /* We were able to successfully read at least one byte before
                 * encountering EOF, but the file did not end in a newline.
                 * Let's hope the last line doesn't contain any NUL bytes. */
                return i + strlen(*buf + i);
            }

            if ((nl = (char*) memchr(*buf + i, '\n', *len - i - 1)) == NULL) {
                /* No newline found.  Either we're at the end of a file with no
                 * newline after its final line, or we need to grow the buffer.
                 * This chunk of code is also used to allocate the initial buffer,
                 * since realloc(NULL, x) works the same as malloc(x). */
                i = *len - 1;
    alloc:      new_len = *len < 60 ? 120 : *len * 2;
                if ((new_buf = (char*) realloc(*buf, new_len)) == NULL) goto error;
                *buf = new_buf;
                *len = new_len;
            } else {
                /* We have the newline, so we're done. */
                return nl - *buf + 1;
            }

            /* Try again. */
        }

    error:
        if (my_malloc) {
            free(*buf);
            *buf = NULL; *len = 0;
        }
        return -1;
    }

#endif





/*****************************************************************************
* Vim: highlight func name in func definition, folding.     }}}{{{1    
******************************************************************************
* syn match cppFuncDef "\~\?\zs\h\w*\ze([^)]*\()\s*\(const\)\?\)\?{ *$"
* hi def link cppFuncDef Todo
* set fdm=marker
*****************************************************************************/


/**************************************************************
* Global constants.    }}}{{{1
**************************************************************/

const short NO_WINR = 0x7f;
const short WINR_K = 'k';
int IsOneWinr(int winr) {return winr==0 || winr==1;}

const int DEFN_TAG = 1;
const int ORIG_TAG = 1;
const int LRN_TAG = 2;
const int NOT_DONE = 3;
#define ALL_PLY -2
#define ZERO_DECLEV 0
#define ZERO_LIT 0


/* Global command-line and compile-time options. */
int ExDbg = false;

#ifndef GloDbgLev
int GloDbgLev = 10;
#endif

#ifdef NO_SAN_CHK
#define GloSanChk 0
#else
int GloSanChk = 1;
#endif

int GloRandSeed;
int GloTimeOut = 0;
long long StartTime;
int RestartCycle = 32;
int NoRestart = false;
int OldCubeWay = false;
int UseLrnType4 = false;
int UseMonotone = 0;
int NoUnivGhost = 0;
int NoExpand = 1;
int AllowExpu = 0;
int expu = 0;
FILE* PrfLog = NULL;

double max_learnts;
double learntsize_factor = 1.0/3.0;
double learntsize_adjust_confl = 100;
double learntsize_adjust_inc = 1.5;
double learntsize_inc = 1.1;
int learntsize_adjust_cnt = (int)(learntsize_adjust_confl);

const int LINE_LEN = 90;



/*************************************************************/
typedef int LitT;


/*****************************************************************************
* Utility Functions.    }}}{{{1
*****************************************************************************/

inline int StrEqc(const char* s1, const char* s2) {
    // Precond: At least one of the strings is not null.
    if ((s1 == NULL) ^ (s2 == NULL)) {return 0;}
    return (strcasecmp(s1, s2) == 0);
}

// char* StrFmt(const char* fmt, ...) {
//     va_list ap;
//     char* pOut = NULL;
//     va_start(ap, fmt);
//     vasprintf(&pOut, fmt, ap);
//     va_end(ap);
//     return pOut;
// }
char* StrFmt(const char* fmt, ...) { 
    va_list ap;
    char* pOut = talloc(char, 16*1024);
    va_start(ap, fmt);
    vsprintf(pOut, fmt, ap);
    va_end(ap);
    return pOut;
}


int die_f(const char* fmt, ...) {
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    abort();
    return 0;
}

int die(const char* msg="") {
    fprintf(stderr, "%s", msg);
    abort();
    return 0;
}

long long GetTimeMicro() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return ((long long) tv.tv_sec) * 1000000 + (tv.tv_usec);
}

long long GetUserTimeMicro() {
    struct rusage usage;
    getrusage(RUSAGE_SELF, &usage);
    struct timeval tv = usage.ru_utime;
    return ((long long) tv.tv_sec) * 1000000 + (tv.tv_usec);
}

int RandSeedFromTime() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    int ret = tv.tv_sec * 10000 + (tv.tv_usec / 100);
    ret = (ret & (((uint)(~0)) >> 1));  // Set the MSB to zero.
    return ret;
}

void DumpIntSet(set<int>* pSet, const char* pEnd) {
    if (pSet == NULL) {printf("(null)%s", pEnd); return;}
    set<int>::iterator iter = pSet->begin();
    printf("[ ");
    for (; iter != pSet->end(); ++iter) {
        printf("%i ", *iter);
    }
    printf("]%s", pEnd);
}

/*template <typename T>
void set_extend(set<T>* pCurSet, set<T>& OtherSet) {
    typename set<T>::iterator iter = OtherSet.begin();
    for (; iter != OtherSet.end(); ++iter) {
        pCurSet->insert(*iter);
    }
}*/

bool gt_f(int const& x, int const& y) {
    return (x > y);
}

bool lt_f(int const& x, int const& y) {
    return (x < y);
}

template <typename K, typename V>
V condget(map<K,V>* pSet, K key, V Default) {
    typename map<K,V>::iterator it = pSet->find(key);
    if (it == pSet->end()) {
        return Default;
    }
    return it->second;
}

/*struct gt_t {
    bool operator() (const int& x, const int& y) const {
        return (x > y);
    }
};*/



/*****************************************************************************
* Fixed-Size Array with Bounds Info.    }}}{{{1
*****************************************************************************/
template <typename T>
struct ap {  // "Array Pointer"
    T* arr;
    int n;

    /* Constructor */
    ap() : arr(NULL), n(0) { }; 
    ap(int* _arr, int _n) : arr(_arr), n(_n) { }; 

    T& operator[](unsigned ix) const {return arr[ix];}
    T& last() {return arr[n-1];}
    void alloc(int size) {
        n = size;
        arr = talloc(T, size);
    }

    T* begin() {return &arr[0];}
    T* end() {return &arr[n];}

    #define FPROTO_ap
    #include "ghostq.proto"
    #undef FPROTO_ap

};

template <typename T>
int ap<T>::indexof(T x) {
    for (int i=0; i < n; i++) {
        if (arr[i] == x) {
            return i;
        }
    }
    return -1;
}

template <typename T>
void ap<T>::sort(int (*cmp)(const T*, const T*)) {
    qsort(this->arr, this->n, sizeof(T), (int (*)(const void*, const void*))(cmp));
    // Note: Use the STL sort algorithm instead; it's faster because it's inlined.
}

template <typename T>
void ap<T>::remove_adj_dups() {
    T* pNewEnd = unique(begin(), end());
    n = pNewEnd - begin();
}

template <typename T>
int DeepEq(ap<T> a1, ap<T> a2) {
    if (a1.n != a2.n) return false;
    for (int i=0; i < a1.n; i++) {
        if (a1.arr[i] != a2.arr[i]) return false;
    }
    return true;
}

template <typename T>
void ap<T>::dealloc() {
    n = 0;
    free(arr);
    arr = NULL;
}

template <typename T>
ap<T> ap<T>::dup() {
    ap<T> ret = ap<T>();
    ret.n = n;
    ret.arr = talloc(T, n);
    for (int i=0; i < n; i++) {
        ret[i] = arr[i];
    };
    return ret;
}



/*****************************************************************************
* Growable Array.    }}}{{{1
* We use this instead of the STL vector class to make debugging a less horrible
* experience and to manually optimize memory usage.
*****************************************************************************/
template <typename T>
struct garr {
    T* arr;
    int n;
    int MaxSize;

    #define FPROTO_garr
    #include "ghostq.proto"
    #undef FPROTO_garr

    /* Constructor */
    garr() : arr(NULL), n(0), MaxSize(0) { }; 
    garr(int _size) : n(_size), MaxSize(_size) {
        arr = talloc(T, _size);
    }; 
    garr(int _size, int prealloc) : n(_size), MaxSize(prealloc) {
        arr = talloc(T, prealloc);
    }; 

    void setval(int ix, T val) {
        if (ix >= n) {
            resize_bzero(ix + 1);
        }
        arr[ix] = val;
    }
    
    T& operator[](unsigned ix) const {return arr[ix];}

    T& last() const {return arr[n-1];}

    void init(int _size=0, int prealloc=8)  {
        n = _size;
        MaxSize = prealloc = max(prealloc, _size);
        if (MaxSize == 0) {arr = NULL;}
        else {
            arr = talloc(T, prealloc);
            if (arr == NULL) {
                int memsize = prealloc * sizeof(T);
                die_f("\nFatal error: Unable to allocate contiguous chunk of %i bytes of memory!\n", memsize);
            }
        }
    };

    void resize(int NewSize) {
        if (NewSize > MaxSize) {realloc_grow(NewSize);}
        n = NewSize;
    };

    void resize_fill(int NewSize, T pad) {
        if (NewSize > MaxSize) {realloc_grow(NewSize);}
        for (int i = n; i < NewSize; i++) {
            arr[i] = pad;
        }
        n = NewSize;
    };

    void grow_fill(int NewSize, T pad) {
        if (n >= NewSize) return;
        resize_fill(NewSize, pad);
    };

    void resize_bzero(int NewSize) {
        if (NewSize > MaxSize) {
            realloc_grow(NewSize);
        }
        if (NewSize > n) {
            memset(arr + n, 0, (NewSize - n) * sizeof(T));
        }
        n = NewSize;
    };

    T* begin() {return &arr[0];}
    T* end() {return &arr[n];}
    int size() const {return n;}

    bool bin_search(T val) {
        return binary_search(this->begin(), this->end(), val);
    }

    void append(T item) {resize(n+1); arr[n-1] = item;};

    T pop() {
        assert(n > 0);
        n--;
        return arr[n];
    }

    typedef T* iterator;

};


template <typename T>
int garr<T>::realloc_grow(int DemandSize) {
    int NewSize;
    if (MaxSize < 65536) {
        if (MaxSize < 8) {
            MaxSize = 8;
        }
        NewSize = MaxSize * 2;
    } else {
        NewSize = (MaxSize + (MaxSize / 4)) & ~4096;
    }
    if (NewSize < DemandSize) {
        NewSize = DemandSize;
    }
    MaxSize = NewSize;
    arr = (T*) realloc(arr, NewSize*sizeof(T));
    if (arr == 0) throw "Fatal error: alloc failed!";
    return 0;
};

template <typename T>
void garr<T>::realloc_shrink(int NewSize=-1) {
    if (NewSize == -1) {
        NewSize = n;
    }
    MaxSize = NewSize;
    arr = (T*) realloc(arr, NewSize*sizeof(T));
};

template <typename T>
T* garr<T>::AppendBlank() {
    resize(n+1);
    arr[n-1] = T();
    memset(&arr[n-1], 0, sizeof(arr[n-1]));
    return &arr[n-1];
}

template <typename T>
void garr<T>::extend(garr<T> items) {
    int base = n;
    int m = items.n;
    resize(n + m);
    for (int i=0; i < m; i++) {
        arr[base+i] = items[i];
    }
};

template <typename T>
void garr<T>::extend(ap<T> items) {
    int base = n;
    int m = items.n;
    resize(n + m);
    for (int i=0; i < m; i++) {
        arr[base+i] = items[i];
    }
};

template <typename T>
void garr<T>::dealloc() {
    n = MaxSize = 0;
    free(arr);
    arr = NULL;
}

template <typename T>
garr<T> garr<T>::dup() {
    garr<T> ret;
    ret.init(n, n);
    for (int i=0; i < n; i++) {
        ret[i] = arr[i];
    };
    return ret;
}

template <typename T>
int garr<T>::indexof(T x, int start=0) {
    for (int i=start; i < n; i++) {
        if (arr[i] == x) {
            return i;
        }
    }
    return -1;
}

template <typename T>
T garr<T>::RemoveAndFill(int ix) {
    T ret = arr[ix];
    arr[ix] = arr[n-1];
    n--;
    return ret;
}

template <typename T>
T garr<T>::PopLast() {
    assert(n > 0);
    T ret = arr[n-1];
    n--;
    return ret;
}

template <typename T>
void garr<T>::RemoveVal_AndFill(T val) {
    for (int i=0; i < n; i++) {
        if (arr[i] == val) {
            RemoveAndFill(i);
            return;
        }
    }
    assert(false);
}

template <typename T>
garr<T> garr<T>::dup_shuffle() {
    garr<T> tmp = this->dup();
    garr<T> ret;
    ret.init(n);
    for (int i=0; i < n; i++) {
        ret[i] = tmp.RemoveAndFill(rand() % (n-i));
    }
    tmp.dealloc();
    return ret;
}

template <typename T>
garr<T> garr<T>::dup_apply(T (*fn)(T)) {
    garr<T> ret;
    ret.init(n, n);
    for (int i=0; i < n; i++) {
        ret[i] = fn(arr[i]);
    };
    return ret;
}

template <typename T>
void garr<T>::apply(T (*fn)(T)) {
    for (int i=0; i < n; i++) {
        arr[i] = fn(arr[i]);
    };
}

template <typename T>
void garr<T>::sort(int (*cmp)(const T*, const T*)) {
    qsort(this->arr, this->n, sizeof(T), (int (*)(const void*, const void*))(cmp));
    // Note: Use the STL sort algorithm instead; it's faster because it's inlined.
}

template <typename T>
void garr<T>::stl_sort() {
    sort(this->begin(), this->end());
}

template <typename T>
void garr<T>::remove_adj_dups() {
    T* pNewEnd = unique(begin(), end());
    n = pNewEnd - begin();
}

template <typename T>
void garr<T>::remove_value(int badval) {
    T* pNewEnd = remove(begin(), end(), badval);
    n = pNewEnd - begin();
}
template <typename T>
void ap<T>::remove_value(int badval) {
    T* pNewEnd = remove(begin(), end(), badval);
    n = pNewEnd - begin();
}


template <typename T, typename C>
void init_garr_from_set(garr<T>* pGarr, set<T,C>& OrigSet) {
    pGarr->resize(OrigSet.size());
    pGarr->n = 0;
    typename set<T>::iterator iter = OrigSet.begin();
    for (; iter != OrigSet.end(); ++iter) {
        pGarr->append(*iter);
    }
}

template <typename T, typename C>
void init_garr_from_list(garr<T>* pGarr, list<T,C>& OrigSet) {
    pGarr->resize(OrigSet.size());
    pGarr->n = 0;
    typename list<T>::iterator iter = OrigSet.begin();
    for (; iter != OrigSet.end(); ++iter) {
        pGarr->append(*iter);
    }
}

template <typename T>
set<T> garr<T>::to_set() {
    set<T> ret(this->arr, this->arr+n);
    return ret;
}

template <typename T>
void garr<T>::reverse() {
    for (int i=0; i < n/2; i++) {
        T tmp1 = arr[i];
        T tmp2 = arr[(n-1)-i];
        arr[i] = tmp2;
        arr[(n-1)-i] = tmp1;
    }
}

template <typename T>
T garr<T>::find_median_destructively() {
    T ret = quick_select(this->arr, this->n);
    return ret;
}

typedef garr<int> garri;

template <typename T>
garr<T> dup_as_garr(ap<T>& cur) {
    garr<T> ret;
    ret.init(cur.n, cur.n);
    for (int i=0; i < cur.n; i++) {
        ret[i] = cur.arr[i];
    };
    return ret;
}

template <typename T>
ap<T> dup_as_ap(garr<T>& cur) {
    ap<T> ret;
    ret.arr = talloc(T, cur.n);
    ret.n = cur.n;
    for (int i=0; i < cur.n; i++) {
        ret[i] = cur.arr[i];
    };
    return ret;
}


void timsort(garri arr) {
    int_tim_sort(arr.arr, arr.n);
}

void timsort(ap<int> arr) {
    int_tim_sort(arr.arr, arr.n);
}



/*****************************************************************************
* BitSet.    }}}{{{1
* Unlike the STL bitset, this doesn't require that the number
* of bits be a compile-time constant.
* TODO: Optimize this so that each bit only consumes 1 bit of memory. 
*****************************************************************************/
struct BitSetT {
    garr<char> bmap;

    void init(int n) {bmap = garr<char>(n);}
    void dealloc() {bmap.dealloc();}
    char operator[](uint ix) {return bmap[ix];}
    void set(uint ix, bool val) {bmap[ix] = val;}
    int  count(int ix) {return bmap[ix];}
    void insert(int ix) {bmap[ix] = 1;}
    void erase(int ix) {bmap[ix] = 0;}

    void erase_vals(garri vals) {
        foreach (itVal, vals) {
            erase(*itVal);
        }
    }
    //void clear(int n) {memset(bmap, 0, n);}

    #define FPROTO_BitSetT
    #include "ghostq.proto"
    #undef FPROTO_BitSetT

};

class FakeBitSetT {
    public:
    set<LitT> data;

    void init(int n) {}
    void dealloc() {}
    int  count(int ix) {return data.count(ix);}
    void insert(int ix) {data.insert(ix);}
    void erase(int ix) {data.erase(ix);}
    void erase_vals(garri vals) {
        foreach (itVal, vals) {
            assert(!*itVal || data.count(*itVal));
            data.erase(*itVal);
        }
        assert(data.size() == 0);
    }
};



/*****************************************************************************
* Permanent Allocator.    }}}{{{1
*****************************************************************************/
// template <typename T>
// class PermAlloc {
//     public:
//     T* mem;
// };


/*****************************************************************************
* Encoding of Literals.    }}}{{{1
*****************************************************************************/

/* Literals are encoded as numbers.  The least-significant bit (LSB)
 * is 0 if the literal is in its positive form, and
 * is 1 if the literal is negated.
 * Currently, we require that the input file uses only even numbers;
 * this makes debugging easier.
 */
int encv(int x) {
    if (x >= 0) {return x;}
    else {return (abs(x) + 1);}
}
int decv(int x) {
    if ((x & 1) == 1) {return -(x & ~1);}
    else {return x;}
}
inline int negv(int x) {return (x ^ 1);}
inline int absv(int x) {return (x & ~1);}
inline int signv(int x) {return (x & 1);}
inline int IsNegLit(int x) {return (x & 1);}
inline int IsPosLit(int x) {return !IsNegLit(x);}


int DumpLitRaw(int lit);

template <typename T>
void DumpLitsEnc(T& arr, const char* newl="\n") {
    const char* sep = "";
    printf("[");
    int n = 0;
    foreach (it, arr) {
        n += printf("%s", sep) + DumpLitRaw(*it);  //printf("%s%i", sep, decv(*it));
        if (n > 85) {
            if (n < 95) printf("\n");
            n = 0;
        }
        sep = " ";
    }
    printf("]%s", newl);
}


void DumpLitsEnc(garri& arr, const char* newl="\n") {
    const char* sep = "";
    printf("[");
    int n = 0;
    foreach (it, arr) {
        n += printf("%s", sep) + DumpLitRaw(*it);  //printf("%s%i", sep, decv(*it));
        if (n > 85) {
            if (n < 95) printf("\n");
            n = 0;
        }
        sep = " ";
    }
    printf("]%s", newl);
}


/*void DumpLitsEnc(ap<int>& arr, const char* newl="\n") {
    printf("[");
    for (int i=0; i < arr.n; i++) {
        printf("%i%s", decv(arr[i]), (i == arr.n-1 ? "" : " "));
    }
    printf("]%s", newl);
}*/


/*****************************************************************************
* Lexer/Parser.    }}}{{{1
*****************************************************************************/
enum {
    INT_TOK_TYPE = 1,
    IDFR_TOK_TYPE,
    OTHER_TOK_TYPE
};

#define LEX_ERR(str) {fprintf(stderr, "%s\n", str); abort();}

/* String Array Pointer */
struct sap {
    char* p;    // pointer
    int n;      // size
};


class Lexer {
    public:
    const char* filename;
    FILE* infile;
    sap CurLine;
    size_t CurLineBufrSize;
    sap CurTok;
    int CurPos;
    int LineNum;
    int CurTokType;

    #define FPROTO_Lexer
    #include "ghostq.proto"
    #undef FPROTO_Lexer

    /* Constructor */
    Lexer(const char* filename_) : filename(filename_) {
        init();
    }

    void init() {
        CurLine = (sap){NULL, 0}; CurLineBufrSize = 0;
        CurTok  = (sap){NULL, 0};
        CurPos = 0;
        LineNum = 0;
        if (StrEqc(filename, "-")) {
            infile = stdin;
        } else {
            infile = fopen(filename, "r");
        }
        if (infile == 0) {
            LEX_ERR(StrFmt("File '%s' does not exist.", filename));
        }
        AdvLine();
        AdvTok();
        while (StrEqc(CurTok.p, "\n")) {AdvTok();}
    };

    void AdvLine() {
        CurLine.n = getline(&CurLine.p, (size_t*)&CurLineBufrSize, infile);
        CurPos = 0;
        LineNum++;
    }
};


void Lexer::AdvTok() {
    int InitPos;
    CurTokType = OTHER_TOK_TYPE;
    again:
    if (CurLine.n == -1) {
        CurTok = (sap){NULL, -1};
        return;
    }
    if (CurPos >= CurLine.n) {
        AdvLine();
        goto again;
    }
    //CurTok.p = &CurLine.p[CurPos];
    InitPos = CurPos;
    switch (CurLine.p[CurPos++]) 
    {
        case '\r':  // For MS-DOS text files
            goto again;

        case ' ': case '\t': case ',':
            goto again;

        case '#':
            // Lines beginning with "#" are comment lines
            CurPos = CurLine.n - 1;
            goto again;

        case '\n': case '(': case ')': case '=':
            CurTok.n = 1;
            goto copytok;

        case '0' ... '9': case '-':
            CurTokType = INT_TOK_TYPE;
            while (true) {
              assert(CurPos <= CurLine.n);
              switch (CurLine.p[CurPos++]) {
                case '0' ... '9': continue;
                default: goto done_idfr;
              }
            }

        case 'a' ... 'z':
        case 'A' ... 'Z': case '_':
            CurTokType = IDFR_TOK_TYPE;
            while (true) {
              assert(CurPos <= CurLine.n);
              switch (CurLine.p[CurPos++]) {
                case '0' ... '9':  case 'a' ... 'z':  case 'A' ... 'Z':  case '_':
                  continue;
                default:
                  goto done_idfr;
              }
            }

        case '<': 
            CurPos++;
            goto done_idfr;

        done_idfr:
        CurPos--;
        CurTok.n = CurPos - InitPos;
        goto copytok;


        default: {
            char* pChrStr;
            unsigned char chr = CurLine.p[CurPos-1];
            if ((32 <= chr) && (chr <= 126)) {
                pChrStr = StrFmt("'%c'", chr);
            } else {
                pChrStr = StrFmt("0x%X", chr);
            }
            fprintf(stderr, "Bad character (%s) on line %i, column %i.\n",
                pChrStr, LineNum, CurPos);
            exit(1);
        }
    }
    copytok:
    CurTok.p = (char*) realloc(CurTok.p, CurTok.n+1);
    memcpy(CurTok.p, &CurLine.p[InitPos], CurTok.n);
    CurTok.p[CurTok.n] = 0;
        
}

char* Lexer::StrCurTokPos() {
    return StrFmt("line %i, col %i-%i", LineNum, CurPos - CurTok.n + 1, CurPos);
}

void Lexer::ChkTok(const char* expected) {
    if (!StrEqc(CurTok.p, expected)) {
        LEX_ERR(StrFmt("Parsing error (%s): Expected '%s'.",
            StrCurTokPos(), expected));
    }
    AdvTok();
}

void Lexer::ReadLineSep() {
    if (!StrEqc(CurTok.p, "\n")) {
        LEX_ERR(StrFmt("Parsing error (%s): Expected a line break.", StrCurTokPos()));
    }
    while (StrEqc(CurTok.p, "\n")) {
        AdvTok();
    }
}

void Lexer::SkipLine() {
    AdvLine();
    AdvTok();
    while (StrEqc(CurTok.p, "\n")) {
        AdvTok();
    }
}
    

int Lexer::ReadIntRaw() {
    if (CurTokType != INT_TOK_TYPE) {
        LEX_ERR(StrFmt("Parsing error (%s): Expected an integer.", StrCurTokPos()));
    }
    int ret = atoi(CurTok.p);
    AdvTok();
    return ret;
}

int Lexer::ReadIntEnc() {
    int x = ReadIntRaw();
    if (x % 2 != 0) {
        LEX_ERR(StrFmt("Parsing error (%s): Expected an even int, but found an odd.",
            StrCurTokPos()));
    }
    return encv(x);
}

const char* Lexer::OpNameSym(char* pNameStr) {
    if (StrEqc(pNameStr, "AND"))    {return "AND";}
    if (StrEqc(pNameStr, "OR"))     {return "OR";}
    return NULL;
}

const char* Lexer::ReadOpIdfr() {
    if (CurTokType != IDFR_TOK_TYPE) {
        LEX_ERR(StrFmt("Parsing error (%s): Expected an identifier.", StrCurTokPos()));
    }
    const char* ret = OpNameSym(CurTok.p);
    if (ret == NULL) {
        LEX_ERR(StrFmt("Parsing error (%s): Unrecognized operator name '%s'.",
            StrCurTokPos(), CurTok.p));
    }
    AdvTok();
    return ret;
}

garri Lexer::ReadArgs() {
    garri ret;
    ret.init(0, 8);
    ChkTok("(");
    while (CurTokType == INT_TOK_TYPE) {
        ret.append(ReadIntEnc());
    }
    if (StrEqc(CurTok.p, ")")) {
        AdvTok();
    } else {
        LEX_ERR(StrFmt("Parsing error (%s): Expected an integer or a ')'.",
            StrCurTokPos()));
    }
    return ret;
}

int Lexer::ReadIntParamEnc(const char* pParamName) {
    ChkTok(pParamName);
    int ret = ReadIntEnc();
    ReadLineSep();
    return ret;
}




/*****************************************************************************
* Decision Level stuff, and other stuff.    }}}{{{1
*****************************************************************************/
/* Note: QbFromLit[x] is 16 bits long. */
#define MAX_QB ((1 << 15) - 2)
#define MAX_VARS ((1 << 24) - 1)

#define NO_DECLEV ((uint)(~0) >> 1)


typedef int FmlaT;

typedef int VsidsT;




typedef map<int, garr<struct GsT*> >  QbDepListT;  // Maps quantifier index to list of dependent game states. 
typedef QbDepListT::iterator QbDepListIterT;


#define OR_GATE 0x77777777

typedef map<LitT, vector<LitT> > SubstCircuitDefn;


/*****************************************************************************
* Headers.    }}}{{{1
*****************************************************************************/
struct QuantBlkT {
    int     qtype;      // 'A' or 'E'.
    int     ix;         // index number
    int     bat;        // gate of which this is a prefix
    garri   lits;
    QuantBlkT() : qtype(0), lits() { }; 
};

int InnermostQB = 0;
int InnermostInputQb = 0;
int ExtraQb[2] = {0,0};

struct QuantPfxT {
    garr<QuantBlkT> blks;
};

struct ExprT {
    const char* sOp;
    garri args;
    struct GsT* pLongLs;
};

struct GateDefT {
    int IsAnd;
    ap<LitT> args;
};


/*****************************************************************************
* ProbInstT Header.    }}}{{{1
*****************************************************************************/
struct ProbInst {
    int LastInpLit;
    int RawLastInpLit;
    int FirstGateVar;
    int LastOrigGlit;
    int NextNewVar;
    int NoMoreVars;
    int LastGlit;
    int LastPlit;   // This should now always be equal to LastGlit.
    int OutGlit;
    garr<QuantBlkT> QuantBlk;   // indexed by qb index number
    garr<GsT*> *LitHave;   
    garr<GsT*> clauses;
    int NumDoneClauses;
    garr<struct GsT*> PhantLitSets;
    //garri *ParentLit;
    int PreprocTimeMilli;
    int LastQtypeBlock[2];

    #define FPROTO_ProbInst
    #include "ghostq.proto"
    #undef FPROTO_ProbInst

    int IsValidInpVar(int x) {return (2 <= x  &&  x <= LastInpLit);}
    int IsValidExprIx(int x) {return (2 <= x  &&  x <= LastGlit);}
    int IsValidQbIx(int x)   {return (1 <= x  &&  x < QuantBlk.n);}

};

ProbInst GloProb = ProbInst();
#define pGloProb (&GloProb)

struct LvCacheEntry {
    int IsAlive;
    long long timestamp;
};


/*****************************************************************************
* Global vars and utility funcs.    }}}{{{1
*****************************************************************************/

/* Global Variables. */
int GloStep = 1;
int* pNumClauses;
int GloIsPostMortem = false;
bool NeedNewl = false;
int NumImpBlocked = 0;
int NumExecAttempts = 0;
long long NumProps = 0;
long long NumWatchFixes = 0;
int NumReducs = 0;
int NumImpUnripe = 0;
int NumImpRevives = 0;
long long NumStepBt = 1;
int NumBigBt = 0;
long long NumDecs = 0;

int LastInOrderDL = 0;

int LastPickGate = 0;

int NumResolves;

// #define AND_GATE 'c'    // "c" for "conjunction"
// #define OR_GATE 'd'     // "d" for "disjunction"
#define INPUT_LIT ((GsT*) (-1))
#define NON_INPUT ((GsT*) 1)
int LastPhant[2];
int PhantOffset[2];
short *QbFromLit;          // maps a lit to the index of its quant block.
garr<garri> GateDefn;
//int *LitToBat;
//int *GlitToBat;
int *QbToBat;
GsT** GlitToGs;
garr<GsT*> GhostToGs;
GsT** GlitToDerivGs;
QuantPfxT QuantPfx;
short* ReEvalArr;
//int* OrigGateFromDeriv;
garri LastWinrChlits;
int LastWinr;
int WinStreak;
int LastChoice;
int NumCegLrn = 0;
int AllowCegar = 0;
int LastOrigClauseID = 0;
unsigned int CurPathMask = 1;
int CurPathNum = 0;
#define PathMaskFromNum(i) (1 << (i))
#define PathNum1 1
#define MaxPathNum 1
#define MaxPathMask PathMaskFromNum(MaxPathNum)

double cla_inc = 1;
double var_inc = 1;
set<GsT*> TempLearntClauses = set<GsT*>();
int NumLearntLocked = 0;

int MinPropQb = 0;

uint* LitToDL;       // maps lit to the decision level in which it was chosen or forced
garr<double> LitHit;
double LitHitInc = 1;


#define USE_PROP_SET 1
#define UseAltProp 1
#define NO_UNIV_REDUC 0

//hash_set<LitT> WaitingPropSet;
BitSetT WaitingPropSet;
garri WaitingPropLits;

//int* TimesInConflict;
//VsidsT* vsids;
set<LitT> AddVsids = set<LitT>();
// set<LitT> CegarLits = set<LitT>();

GsT* AbsGlitToGs(int glit) {
    assert(GlitToGs[glit]==0 || GlitToGs[negv(glit)]==0);
    return (GsT*)((long)(GlitToGs[glit]) | (long)(GlitToGs[negv(glit)]));
}

GsT* AbsGlitToDerivGs(int glit) {
    assert(GlitToDerivGs[glit]==0 || GlitToDerivGs[negv(glit)]==0);
    return (GsT*)((long)(GlitToDerivGs[glit]) | (long)(GlitToDerivGs[negv(glit)]));
}

GsT* NegGlitToGs(int glit) {
    return GlitToGs[negv(glit)];
}

void FixNeedNewl() {
    if (NeedNewl) {printf("\n"); NeedNewl = false;}
}

int IsAndGate(int lit) {
    return ((long) GhostToGs[lit]) > 1;
}

inline int IsOrGate(int lit) {
    int ret = (GhostToGs[lit] == NON_INPUT);
    if (ret) {
        assert(IsAndGate(negv(lit)));
    }
    return ret;
}

int GloQtype(int ixQb) {
    return pGloProb->QuantBlk[ixQb].qtype;
}

int QtypeFromLit(int lit) {
    return GloQtype(QbFromLit[lit]);
}

int QtypeNumFromLit(int lit) {
    switch (QtypeFromLit(lit)) {
        case 'A':  return 0;
        case 'E':  return 1;
        default:   assert(false);
    }
}

int QtypeNumFromLetr(char letr) {
    switch (toupper(letr)) {
        case 'A':  return 0;
        case 'E':  return 1;
        default:   assert(false);
    }
}

int QtypeLetrFromNum(int num) {
    switch (num) {
        case 0:  return 'A';
        case 1:  return 'E';
        default:   assert(false);
    }
}


int find_var_pol(garri a, int var) {
    /* precond: array a must be sorted. */
    var = absv(var);
    int* pRet = lower_bound(a.begin(), a.end(), var);
    if (absv(*pRet) == absv(var)) {
        return *pRet;
    } else {
        return 0;
    }
}




struct LessGsById {
    bool operator() (const GsT* const& pLs1, const GsT* const& pLs2) const;
};

int NumWatchCleanups = 0;

//typedef set<struct GsT*,LessGsById> GwSetT;
#define USE_WATCH_SET 0
#define USE_WATCH_LIST 0
#define USE_WATCH_ARR 1

#if USE_WATCH_SET
    typedef set<pair<int, struct GsT*> > GwSetT;
    #define GwSetT_erase_iterator GwSetT::iterator
#elif USE_WATCH_ARR
    #define GwSetT_erase_iterator int
#elif USE_WATCH_LIST
    typedef list<struct GsT*> GwSetT;
    #define GwSetT_erase_iterator GwSetT::iterator
#endif




#define LRN_TYPE_1 1
#define LRN_TYPE_2 2
#define LRN_TYPE_3 3
#define LRN_TYPE_4 4

/*****************************************************************************
* Header for Game-State Entry data structure.    }}}{{{1
*****************************************************************************/
struct GsT {
    int id;
    char FreeFmla;
    char LrnType;
    char inactive;
    char IsLearned;
    int GlitDefd;
    //int IsBattle;
    //int GateLit;
    //int ctlr;
    //int IsInit;
    //set<LitT> ctlrs;
    //garri* pChoices;
    //int WatchedLit[2];

    int NumFix;
    int WatStartPos;
    //int WatStopPos;
    int WatStopBt;
    int LastUse;
    unsigned int InUse;
    double activity;
    int NumFixes;
    int WatchedRes;
    int WatchRawAE[2];
    #define WatchRawArr WatchRawAE
    //#define WatchRawAE WatOrd
    //#define WatchRawArr WatOrd.arr
    GwSetT_erase_iterator itWatchRes;
    GwSetT_erase_iterator itWatchRawAE[2];

    ap<int> ReqLits;    // Triggering Lits (corresponding to $L^{now}$ in the paper)
    ap<int> WatOrd;    
    // ap<int> TrigAE[2];  // Triggering universal lits (TrigAE[0]) and
    //                     // triggering existential lits (TrigAE[1]), sorted.
    ap<int> ImpLits;    // Reserved Lits (corresponding to $L^{fut}$ in the paper)

    int& WatchLosrAE(int x) {
        return WatchRawAE[x];
    }

    double get_adj_act() {
        return activity;
        // return activity / (NumFixes + 1);
    }

    #define FPROTO_GsT
    #include "ghostq.proto"
    #undef FPROTO_GsT

};


bool GsGtByAct(GsT* const p1, GsT* const p2) {
    return p1->activity > p2->activity;
}


void GsT::init() {
    WatStartPos = 0;
    //ctlrs = set<LitT>();
    //IsInit = true;
}

vector<LitT>* GsT::AsLitVec() {
    GsT* pCurLs = this;
    vector<LitT>* pRawInts;
    pRawInts = new vector<LitT>(pCurLs->ReqLits.begin(), pCurLs->ReqLits.end());
    foreach (itImp, pCurLs->ImpLits) {
        pRawInts->push_back(*itImp);
    }
    return pRawInts;
}

#if USE_WATCH_SET == 1
int set_count(GwSetT& s, GsT* item) {return s.count(pair<int, struct GsT*>(item->id, item));}  // for debugging
#endif

bool LessGsById::operator() (const GsT* const& pLs1, const GsT* const& pLs2) const {
    //assert(pLs1->id != 0 && pLs2->id != 0);
    if (pLs1->id == pLs2->id) {return pLs1 < pLs2;}
    return (pLs1->id < pLs2->id);
}

bool FnGreaterGsById(const GsT* const& pLs1, const GsT* const& pLs2) {
    return (pLs1->id > pLs2->id);
}

bool FnLessGsById(const GsT* const& pLs1, const GsT* const& pLs2) {
    return (pLs1->id < pLs2->id);
}

int CmpGsById(GsT* const* p1, GsT* const* p2) {
    return ((*p1)->id - (*p2)->id);
}


int CmpLsByPtr(GsT* const* ppLs1, GsT* const* ppLs2) {
    GsT* pLs1 = *ppLs1;
    GsT* pLs2 = *ppLs2;
    return (pLs1 - pLs2);
}




inline GwSetT_erase_iterator* get_gst_itwatch(GsT* pCurLs, LitT wlit) {
    GwSetT_erase_iterator* ret = NULL;
    if (pCurLs->WatchRawAE[0] == wlit) {return &pCurLs->itWatchRawAE[0];}
    if (pCurLs->WatchRawAE[1] == wlit) {return &pCurLs->itWatchRawAE[1];}
    if (pCurLs->WatchedRes    == wlit) {return &pCurLs->itWatchRes;}
    return ret;
}

inline void update_gst_itwatch(GsT* pCurLs, LitT wlit, GwSetT_erase_iterator it) {
    if      (pCurLs->WatchRawAE[0] == wlit) {pCurLs->itWatchRawAE[0] = it;}
    else if (pCurLs->WatchRawAE[1] == wlit) {pCurLs->itWatchRawAE[1] = it;}
    else if (pCurLs->WatchedRes    == wlit) {pCurLs->itWatchRes = it;}
}


#if USE_WATCH_ARR
    class GwSetT {
        public:
        garr<GsT*> arr;
        int num_blanks;
        int wlit;

        /* Constructor */
        GwSetT() {
            arr.init();
            num_blanks = 0;
        }

        typedef GsT** iterator;
        typedef GwSetT_erase_iterator erase_iterator;

        erase_iterator insert(GsT* pCurLs) {
            int num_good = arr.n - num_blanks;
            if (num_blanks > 0 && (num_good == 0 || num_blanks > 4 + num_good)) {
                remove_blanks();
            }
            arr.append(pCurLs);
            erase_iterator ret = arr.n - 1;
            return ret;
        }

        void erase(erase_iterator it, GsT* pCurLs) {
            int pos = it;
            assert (arr[pos] == pCurLs);
            arr[pos] = NULL;
            /*else {
                int flag = false;
                for (int ix=0; ix < arr.n; ix++) {
                    if (arr[ix] == pCurLs) {
                        arr[ix] = NULL;
                        flag = true;
                        break;
                    }
                }
                assert(flag);
            }*/
            num_blanks++;
            while (arr.n != 0 && arr.last() == NULL) {
                arr.n--;
                num_blanks--;
            }
        }
        void remove_blanks() {
            NumWatchCleanups++;
            auto(src,  arr.begin());
            auto(dest, arr.begin());
            auto(endptr,  arr.end());
            int NumHits = 0;
            while (src != endptr) {
                auto(pCurLs, *src);
                if (pCurLs != NULL) {
                    if (src != dest) {
                        *dest = *src;
                        update_gst_itwatch(pCurLs, wlit, dest - arr.begin());
                        // erase_iterator* erit = get_gst_itwatch(pCurLs, wlit);
                        // *erit = dest - arr.begin();
                    }
                    dest++;
                } else {
                    NumHits++;
                }
                src++;
            }
            arr.n = dest - arr.begin();
            assert(NumHits == num_blanks);
            num_blanks = 0;
        }

        iterator begin() {return arr.begin();}
        iterator end() {return arr.end();}
    };
#endif



    
//GsT EmptyLs;



/*****************************************************************************
* ChlitListT: List of chosen (non-forced) literals.    }}}{{{1
*****************************************************************************/
class ChlitListT {
    public:
    garri flat;
    #define FPROTO_ChlitListT
    #include "ghostq.proto"
    #undef FPROTO_ChlitListT
};

void ChlitListT::append(int lit) {
    flat.append(lit);
}

int ChlitListT::pop() {
    return flat.PopLast();
}

int ChlitListT::last() {
    return flat.last();
}

void ChlitListT::dump() {
    int CurQb = 0;
    printf("[");
    for (int i=1; i < flat.n; i++) {
        int lit = flat[i];
        if (i != 1) {printf("%c ", (QbFromLit[lit]==CurQb ? ',' : ';'));}
        CurQb = QbFromLit[lit];
        printf("%i", decv(lit));
    }
    printf("]\n");
}

/*****************************************************************************
* Stuff  }}}{{{1
*****************************************************************************/


#define UNIT_PROP_ANTECED 1
#define MONO_ANTECED 2
#define CONTRADICT_ANTECED 3

struct AntecedT {
    GsT* pDepGs;

    /* Constructor */
    AntecedT() : pDepGs(0) {};
    AntecedT(GsT* _gs) : pDepGs(_gs) {};

    bool operator==(AntecedT x) {
        return (this->pDepGs == x.pDepGs);
    }
};

AntecedT EmptyAnteced;


bool AntecedByGreaterGsId(AntecedT const& a1, AntecedT const& a2) {
    if (a1.pDepGs->id <  a2.pDepGs->id) return true;
    if (a1.pDepGs->id != a2.pDepGs->id) return false;
    return false;
}


/****************************************************************************/
struct GsPriQ {
    garr<AntecedT> a;
    int first;
    //set<GsT*> s;

    void init() {
        a.init();
        //s = set<GsT*>();
    }
    void append(AntecedT ante) {
        //int IsNew = s.insert(pGs).second;
        //if (!IsNew) {return;}
        a.append(ante); 
        push_heap(a.begin(), a.end(), AntecedByGreaterGsId);
    }
    AntecedT pop()  {
        if (a.n == 0) return EmptyAnteced;
        pop_heap(a.begin(), a.end(), AntecedByGreaterGsId);
        AntecedT ret = a.PopLast();
        //s.erase(ret);
        return ret;
    }
    AntecedT peek() {return (a.n == 0 ? EmptyAnteced : a[0]);}
    bool is_empty() {return (a.n == 0);}

    void clear() {
        a.n=0; 
        //s.clear();
    }

    /*void init() {first=0; a.init();}
    void append(AntecedT ante) {a.append(ante);}
    AntecedT pop()  {return (first >= a.n ? EmptyAnteced : a[first++]);}
    AntecedT peek() {return (first >= a.n ? EmptyAnteced : a[first]);}
    bool is_empty() {return (first >= a.n);}
    void clear() {first=0; a.n=0;}*/
};


typedef pair<int,int> IntIntPair;
typedef pair<GsT*,int> GsIntPair;

struct GsSetByPly {
    set<GsT*,LessGsById> ply[2];
};


template <typename T, bool (*lt)(T const& x, T const& y)>
struct heap {
    garr<T> h;

    void insert(T x) {
        h.append(x);
        push_heap(h.begin(), h.end(), lt);
    }

    T pop() {
        assert(h.n > 0);
        T ret = h[0];
        while (h.n > 0 && h[0] == ret) {
            pop_heap(h.begin(), h.end(), lt);
            h.n--;
        }
        return ret;
    }

    T peek() {
        return h[0];
    }

    bool empty() {return (h.n == 0);}

    T top() {return h[0];}
    void push(T x) {insert(x);}
    
};



class GsBuilder;


#define CONFLICT_GST_PRI 1
//#define MONO_GST_PRI 2
#define FORC_GST_PRI 3
//#define WIN_GST_PRI 4
#define MAX_GST_PRI 4

   
#define EM_DRYRUN 0
#define EM_SCHED 1
#define EM_WIN 3


typedef int ChronoT;

/*****************************************************************************
* AsgnT Header.    }}}{{{1
*****************************************************************************/
struct AsgnT {
    ChlitListT chlits;   // Chosen lits, arranged according to quantifier block
    GwSetT *LitWatch;   // LitWatch[lit] lists the game-states that are watching lit.
    //set<LitT> FreeVars;
    garri* UndoListByDL;
    //LitT* UndoTarg;
    AntecedT* DepByLit;
    ChronoT* LitToChrono;
    garr<char> PolPref[MaxPathNum+1];
    //long long* LitToAbsTime;  // For debugging only
    LitT* ChronoToLit;
    FmlaT FreeFmla;
    set<LitT> LitsInConflict;
    //set<LitT> BattleVars;
    //garri BattleVarList;

    GsT* pConflictGs;

    //vector<set<LitT> > FreeVarsByQb;
    

    set<LitT> VarsExpanded;
    //garri *ParentLit;
    //garr<int> NestLev;
    //garr<char> HasInner;
    // garr<garri> VarEsc;
    // garr<garri> VarPars;
    // garr<garri> PossAnc;
    // garr<garri> Lca;
    //map<LitT, char > GateZap;  // map to 0, 't', or 'f'.
    map<LitT, LitT> ExpSing[2];
    map<LitT, char > ExpCh[2]; // Either (1) ptr to a set, (2) 't', or (3) 'f'.
    #define EXPDEP_T ((char) 't')
    #define EXPDEP_F ((char) 'f')
    #define EXPDEP_0 ((char) 'o') 
    #define EXPDEP_MOD ((char) 'M') 
    #define EXPDEP_DONTCARE ((char) 'x')
    #define IsConstExpDep(e) ((e)==EXPDEP_T || (e)==EXPDEP_F)
    int CurExpSplit;
    int NumNewNodes;

    int TargGate;
    int CurChrono;
    int LastInitGs;
    int LastDefnGs;
    int DeathChrono;
    long long NumAppends;

    GsPriQ PendingExecs[MAX_GST_PRI+1]; 

    //garr<GsT*> DoneLitSets;
    //set<GsT*> DoneGsSet;
    //garr<GsT*> DoneGsTmp;
    //GsT* pBestDoneLs;
    //int NumDoneGSes;

    //int IsRestarting;
    garri InitChoices;
    //garri TargStack;
    int ixChoi;
    int NumSinceCut;
    int NumUntilRestart;
    long long LastPrintTime;
    int NumRestarts;
    set<pair<int, vector<int> > > runs;

    //LvCacheEntry* LvCache;
    //LitT* LitToLiveParent;  // maps lit to nondead orig clause with it
    //set<LitT>* ParentToWatchers;   // maps parent literal to set of watching lits
    // int NumMonoProps;  // for use as a timestamp
    //set<LitT> PendingMonotoneLits;
    //heap<LitT,lt_f> PendingDeadPars;
    //long long *LastChkPar;

    //int* TriedLit;
    //int *LastHitDL;  // for debugging



    #define FPROTO_AsgnT
    #include "ghostq.proto"
    #undef FPROTO_AsgnT
    
    inline bool HasLit(int x) {return LitToDL[x] != NO_DECLEV;}
    inline bool HasLitOrNeg(int x) {return HasLit(x) || HasLit(negv(x));}
    inline bool IsAlreadyExp(int x) {return VarsExpanded.count(absv(x));}

    inline int GetLastChlit() {
        return chlits.last();
    }

    inline uint GetCurDL() {
        return LitToDL[GetLastChlit()];
    }

    bool IsChlit(int lit) {
        return (ChlitByDL(LitToDL[lit]) == lit);
    }

    inline GwSetT_erase_iterator LitWatchInsert(int lit, GsT* pGs) {
        //assert(LitWatch[lit].count(pGs) == 0);
        #if USE_WATCH_SET
        return LitWatch[lit].insert(pair<int,GsT*>(pGs->id, pGs)).first;
        #elif USE_WATCH_ARR
        return LitWatch[lit].insert(pGs);
        #elif USE_WATCH_LIST
        return LitWatch[lit].insert(LitWatch[lit].end(), pGs);
        #endif
    }
    inline void LitWatchErase(int lit, GsT* pGs, GwSetT_erase_iterator it) {
        #if USE_WATCH_ARR
        LitWatch[lit].erase(it, pGs);
        #else
        LitWatch[lit].erase(it);
        #endif
    }
};

/****************************************************************************/

AsgnT GloAsgn = AsgnT();
#define pGloAsgn (&GloAsgn)

/****************************************************************************/
#include "Heap.h"

garr<double> VarActivity;

struct VarActLt {
    bool operator () (int x, int y) const {
        if (VarActivity[x] == VarActivity[y]) {
            return x < y;
        }
        return VarActivity[x] > VarActivity[y];
    }
};

VarActLt VarActLtInst;

template <class Lt>
class VarOrderByQb {
    public:
    garr<Heap<Lt> > order_heap;
    Lt comp;

    VarOrderByQb(Lt c) {
        comp = c;
        order_heap = garr<Heap<Lt> >();
    }

    void add_qb(int max_qb) {
        while (order_heap.n <= max_qb) {
            order_heap.append(Heap<Lt>(comp));
        }
    }

    void insert(int var) {
        assert(absv(var) == var);
        int qb = QbFromLit[var];
        if (!order_heap[qb].inHeap(var)) {
            order_heap[qb].insert(var);
        }
    }

    bool inHeap(int var) {
        int qb = QbFromLit[var];
        return order_heap[qb].inHeap(var);
    }

    LitT removeMin() {
        for (int qb=0; qb < order_heap.size(); qb++) {
            if (order_heap[qb].size() > 0) {
                return order_heap[qb].removeMin();
            }
        }
        return 0;
    }

    void decrease_key(int var) {
        int qb = QbFromLit[var];
        order_heap[qb].decrease(var);
    }
};

VarOrderByQb<VarActLt> VarActOrder = VarOrderByQb<VarActLt>(VarActLt());




GsT* AsgnT::GetDepByLit(int lit) {
    return DepByLit[lit].pDepGs;
}

void AsgnT::SetDepByLit(int lit, GsT* pDepGs) {
    assert((pDepGs->InUse & CurPathMask) == 0);
    pDepGs->InUse |= CurPathMask;
    if (pDepGs->IsLearned) {
        NumLearntLocked++;
    }
    DepByLit[lit].pDepGs = pDepGs;
}

void AsgnT::ResetDepByLit(int lit, bool KeepDepLocked) {
    GsT* pDepGs = DepByLit[lit].pDepGs;
    if (!pDepGs) return;
    assert((pDepGs->InUse & CurPathMask));
    if (!KeepDepLocked) {
        pDepGs->InUse &= ~CurPathMask;
        if (pDepGs->IsLearned) {
            NumLearntLocked--;
        }
    }
    DepByLit[lit].pDepGs = NULL;
}

int PlyFromLit(int lit) {
    return QtypeNumFromLit(lit);
}

int frob_raw_gate(int glit) {
    int pol = signv(glit);
    int ret = (absv(glit)*2) ^ pol;
    return ret;
}

int ghost_to_orig(int glit) {
    int pol = signv(glit);
    int ret = absv(absv(glit)/2) ^ pol;
    return ret;
}

int ghosted(int glit, int ply) {
    int pol = signv(glit);
    int gvar = absv(glit);
    assert((gvar & 0x2) == 0);
    int ret = (gvar | (ply*2)) ^ pol;
    assert(pGloProb->RawLastInpLit < ret);
    return ret;
}

bool is_ghost(int lit) {
    int ret = (GhostToGs[lit] != INPUT_LIT);
    //assert(ret == (pGloProb->RawLastInpLit < lit));
    return ret;
}

int ghosted_or_pass(int lit, int ply) {
    if (is_ghost(lit)) {
        return ghosted(lit, ply);
    } else {
        return lit;
    }
}

bool IsInputLit(int glit) {
    return (GhostToGs[glit] == INPUT_LIT);
}

inline bool IsOrigLit(int lit) {
    return (lit <= pGloProb->LastOrigGlit);
}

bool IsReallyGateLit(int glit) {
    return (GhostToGs[glit] != INPUT_LIT);
}

void DumpOrigGate(int lit) {
    int OrigLit = decv(is_ghost(lit) ? ghost_to_orig(lit) : lit);
    const char* PlyName = (is_ghost(lit) ? (QtypeNumFromLit(lit) ? "e" : "a") : "");
    printf("%i%s\n", OrigLit, PlyName);
}

int DumpLitRaw(int lit) {
    return printf("%i", decv(lit));
}

void DumpLit(int lit) {
    DumpLitRaw(lit);
    printf("\n");
}
    

// prototype declaration
void add_gate_to_qb(int ixExpr, garri args);
void add_var_to_qblock(int qvar, int ixQb);

// int Phantomize(int lit, int ply) {
//     return lit + PhantOffset[ply];
// }

// int Unphantomize(int lit) {
//     if (lit >= PhantOffset[0]) {
//         if (lit >= PhantOffset[1]) {
//             lit -= PhantOffset[1];
//         } else {
//             lit -= PhantOffset[0];
//         }
//     }
//     return lit;
// }


/* DEBUGGING ****************************************************************/

void VarDL(int var) {
    if (var < 0) {printf("Error.\n"); return;}
    if (var > pGloProb->LastPlit) {printf("Error.\n"); return;}
    int pos = LitToDL[var];
    int neg = LitToDL[var ^ 1];
    if (pos < (1 << 30)) {printf("Pos at DL %u.\n", pos);}
    else if (neg < (1 << 30)) {printf("Neg at DL %u.\n", neg);}
    else {printf("Unassigned.\n");}
}


/****************************************************************************/


// int NarfType(int lit) {
//     if (QtypeFromLit(lit) == 'E') {return 3;}
//     if (QtypeFromLit(lit) == 'A') {return 2;}
//     assert(false);
// }
// 
// int CmpLitByNarf(const int *pLit1, const int *pLit2) {
//     int x1 = (NarfType(*pLit1) << 26) + *pLit1;
//     int x2 = (NarfType(*pLit2) << 26) + *pLit2;
//     return x1 - x2;
// }
//         
// int CmpLitByPoit(const int *pLit1, const int *pLit2) {
//     assert(NarfType(*pLit1) == NarfType(*pLit2));
//     int x1 = *pLit1;
//     int x2 = *pLit2;
//     return x1 - x2;
// }
        
int CmpLitByQb(const int *pLit1, const int *pLit2) {
    int q1 = QbFromLit[*pLit1];
    int q2 = QbFromLit[*pLit2];
    if (q1 == q2) {return 0;}
    else if (q1 < q2) {return -1;}
    else {return 1;}
}
        
int CmpLitByDL(const int *pLit1, const int *pLit2) {
    int d1 = LitToDL[*pLit1];
    if (d1==NO_DECLEV &&  LitToDL[negv(*pLit1)]==NO_DECLEV) {d1--;}
    int d2 = LitToDL[*pLit2];
    if (d2==NO_DECLEV &&  LitToDL[negv(*pLit2)]==NO_DECLEV) {d2--;}
    if (d1 == d2) {return 0;}
    else if (d1 < d2) {return -1;}
    else {return 1;}
}

int CmpLitByChrono(const int *pLit1, const int *pLit2) {
    ChronoT d1 = pGloAsgn->LitToChrono[*pLit1];
    ChronoT d2 = pGloAsgn->LitToChrono[*pLit2];
    //return d1 - d2; // NO!
    if (d1 == d2) {return (*pLit1 - *pLit2);}
    else if (d1 < d2) {return -1;}
    else {return 1;}
}

struct LitLessByDL {
    bool operator() (const int& Lit1, const int& Lit2) const {
        int d1 = LitToDL[Lit1];
        int d2 = LitToDL[Lit2];
        if (d1 == d2) {
            if (Lit1 == Lit2) return false;
            assert(!((pGloAsgn->IsChlit(Lit1) && (pGloAsgn->IsChlit(Lit2)))));
            if (pGloAsgn->IsChlit(Lit1)) return true;
            if (pGloAsgn->IsChlit(Lit2)) return false;
            return (Lit1 < Lit2);
        }
        return (d1 < d2);
    }
};


/*bool LitLessByChronof(const int& Lit1, const int& Lit2) {
    ChronoT d1 = pGloAsgn->LitToChrono[Lit1];
    ChronoT d2 = pGloAsgn->LitToChrono[Lit2];
    if (d1 == d2) {assert(Lit1 == Lit2);}
    return (d1 < d2);
}*/

struct LitLessByChrono {
    bool operator() (const int& Lit1, const int& Lit2) const {
        ChronoT d1 = pGloAsgn->LitToChrono[Lit1];
        ChronoT d2 = pGloAsgn->LitToChrono[Lit2];
        //if (d1 == d2) {assert(Lit1 == Lit2);}
        return (d1 < d2);
    }
};

struct LitGtByChrono {
    bool operator() (const int& Lit1, const int& Lit2) const {
        ChronoT d1 = pGloAsgn->LitToChrono[Lit1];
        ChronoT d2 = pGloAsgn->LitToChrono[Lit2];
        //if (d1 == d2) {assert(Lit1 == Lit2);}
        return (d1 > d2);
    }
};



// struct HashedApInt {
//     int hash;
//     ap<int> x;
// 
//     void init (ap<int> a_list) {
//         x = a_list;
//         hash = 0;
//         foreach (itArg, a_list) {
//             hash = hash * 3 + *itArg;
//         }
//     }
//     HashedApInt(garr<int> a_list) {
//         init(dup_as_ap(a_list));
//     }
// 
//     bool operator<(const HashedApInt& y) const;
//     bool operator==(const HashedApInt& y) const;
//     void dump();
// 
// };
// 
// void HashedApInt::dump() {
//     printf("(hash %8i) ", hash);
//     DumpLitsEnc(x);
// }
// 
// bool HashedApInt::operator<(const HashedApInt& y) const {
//     printf("CMP "); this->dump();
//     printf(" to "); y.dump();
//     if ((*(HashedApInt*)this).x.n < y.x.n) {return true;}
//     if ((*this).hash < y.hash) {return true;}
//     int n = (*this).x.n;
//     for (int i=0; i < n; i++) {
//         if (((*this).x)[i] < (y.x)[i]) {
//             return true;
//         }
//     }
//     printf("RESULT false\n");
//     return false;
// }
// 
// bool HashedApInt::operator==(const HashedApInt& y) const {
//     if ((*this).x.n != y.x.n) {return false;}
//     if ((*this).hash != y.hash) {return false;}
//     int n = (*this).x.n;
//     for (int i=0; i < n; i++) {
//         if (((*this).x)[i] != (y.x)[i]) {
//             return false;
//         }
//     }
//     return true;
// }
// 
// map<HashedApInt, LitT> ArgsToGlit;
map<vector<int>, GsT*> DupLrnTest;

map<vector<int>, LitT> ArgsToGlit;
set<vector<int> > CegLrnSet;

vector<int> HashedApInt(garri a) {
    vector<int> ret = vector<int>(a.arr, a.arr + a.n);
    return ret;
}

vector<int> VectorFromAp(garri a) {
    vector<int> ret = vector<int>();
    foreach (itLit, a) {
        ret.push_back(*itLit);
    }
    return ret;
}



/****************************************************************************/

void AsgnT::init() {
    //FreeVars = set<LitT>();
    UndoListByDL =  talloc(garri, pGloProb->LastPlit+1);
    //UndoTarg =      talloc(LitT,  pGloProb->LastInpLit+1);
    //LastHitDL =     talloc(int,   pGloProb->LastInpLit+1);
    LitToDL =       talloc(uint,        pGloProb->LastPlit+1);
    LitHit  =       garr<double>(pGloProb->LastPlit+1);
    DepByLit =      talloc(AntecedT,    pGloProb->LastPlit+1);
    LitToChrono =   talloc(ChronoT,     pGloProb->LastPlit+1);
    WaitingPropSet.init(pGloProb->LastPlit + 1);
    for (int path=0; path <= MaxPathNum; path++) {
        PolPref[path] = garr<char>(pGloProb->LastPlit+1);
    }
    //LitToAbsTime =  talloc(long long,   pGloProb->LastPlit+1);
    ChronoToLit =   talloc(LitT,        pGloProb->LastPlit+1);
    //LvCache =       talloc(LvCacheEntry,pGloProb->LastPlit+1);
    //TriedLit =      talloc(int,         pGloProb->LastPlit+1);
    LitsInConflict = set<LitT>();
    //BattleVars = set<LitT>();
    //BattleVarList = garri();
    pConflictGs = NULL;
    pGloAsgn->FreeFmla = NOT_DONE;
    chlits.flat.init();
    LastInitGs = (1 << 30);
    NumAppends = 0;
    for (int i=0; i <= MAX_GST_PRI; i++) {
        PendingExecs[i].init();
    }
    for (int i=0; i <= pGloProb->LastPlit; i++) {
        LitToDL[i] = NO_DECLEV;
        LitToChrono[i] = ((ChronoT) NO_DECLEV);
    }
    this->AppendChlit(ZERO_LIT);
    LitToDL[ZERO_LIT] = ZERO_DECLEV;
    CurChrono = 0;
    
    //ParentLit = pGloProb->ParentLit;
    //NestLev.init(pGloProb->LastGlit+2, 0);
    //HasInner.init(pGloProb->LastGlit+2, 0);
    // VarEsc.init(pGloProb->LastGlit+2, 0);
    // VarPars.init(pGloProb->LastGlit+2, 0);
    // PossAnc.init(pGloProb->LastGlit+2, 0);
    // Lca.init(pGloProb->LastGlit+2, 0);
    VarsExpanded = set<LitT>();

    ExpCh[0] = map<LitT, char >();
    ExpCh[1] = map<LitT, char >();
    ExpSing[0] = map<LitT, LitT>();
    ExpSing[1] = map<LitT, LitT>();
    //GateZap = map<LitT, char >();
    // for (int v = 0; v <= pGloProb->LastGlit; v += 1) {
    //     VarEsc[v].init(0, 0);
    //     VarPars[v].init(0, 0);
    //     PossAnc[v].init(0, 0);
    //     Lca[v].init(0, 0);
    // }
}

void AsgnT::init2() {
    //FreeVarsByQb = vector<set<LitT> >(pGloProb->QuantBlk.n);
    foreach (itQb, pGloProb->QuantBlk) {
        //int ixQb = itQb->ix;
        foreach (itLit, itQb->lits) {
            int lit = *itLit;
            if (absv(lit) != lit) continue;
            //FreeVarsByQb[ixQb].insert(lit);
        }
    }
}

void AsgnT::PrintFormula(char const* outname) {
    FILE* outf = fopen(outname, "w");
    fprintf(outf, "CktQBF\n");
    fprintf(outf, "LastInputVar %i\n", absv(pGloProb->LastInpLit));
    fprintf(outf, "LastGateVar %i\n", absv(pGloProb->LastGlit));
    fprintf(outf, "OutputGateLit %i\n\n", absv(pGloProb->OutGlit));
    fprintf(outf, "<q gate=%i>\n", absv(pGloProb->OutGlit));
    for (int i=1; i < pGloProb->QuantBlk.n; i++) {
        QuantBlkT* pQb = &pGloProb->QuantBlk[i];
        assert(pQb->bat == pGloProb->OutGlit);
        fprintf(outf, "    %c", tolower(pQb->qtype));
        foreach (itLit, pQb->lits) {
            if (IsAlreadyExp(*itLit)) continue;
            fprintf(outf, " %i", decv(*itLit));
        }
        fprintf(outf, "\n\n");
    }
    fprintf(outf, "</q>\n\n");

    for (int g = absv(pGloProb->NextNewVar)-2; g >= pGloProb->FirstGateVar; g -= 2) {
        //if (VarPars[g].n == 0 && absv(g) != absv(pGloProb->OutGlit)) continue;
        const char* op = (IsAndGate(g) ? "and" : "or");
        auto(args, dup_as_garr(AbsGlitToGs(g)->ReqLits));
        sort(args.begin(), args.end());
        fprintf(outf, "%i = %s(", g, op);
        bool NoComma = true;
        for (int i=0; i < args.n; i++) {
            int CurArg = args[i];
            if (IsOrGate(g)) {CurArg = negv(CurArg);}
            if (HasLitOrNeg(CurArg)) continue;
            fprintf(outf, "%s%i", (NoComma ? "" : ", "), decv(CurArg));
            NoComma = false;
        }
        args.dealloc();
        // int OrigGlit = OrigGateFromDeriv[g];
        // if (OrigGlit && OrigGlit != g) {
        //     fprintf(outf, ")  #orig %i, ", OrigGlit);
        // } else {
        //     fprintf(outf, ")  # ");
        // }
        fprintf(outf, ")  \n");
        //fprintf(outf, "[OccPol=%i, Lev=%i]\n", OccPolarity(g), NestLev[g]);
    }
    fclose(outf);
}

void AsgnT::PrintFormulaQdimacs(char const* outname) {
    FILE* outf = fopen(outname, "w");
    assert(IsAndGate(pGloProb->OutGlit));
    fprintf(outf, "p cnf 0 0                    \n");
    for (int i=1; i < pGloProb->QuantBlk.n; i++) {
        QuantBlkT* pQb = &pGloProb->QuantBlk[i];
        assert(pQb->bat == pGloProb->OutGlit);
        fprintf(outf, "%c", tolower(pQb->qtype));
        foreach (itLit, pQb->lits) {
            if (IsAlreadyExp(*itLit)) continue;
            fprintf(outf, " %i", decv(*itLit));
        }
        if (i == pGloProb->QuantBlk.n - 1) {
            if (tolower(pQb->qtype) == 'a') {
                fprintf(outf, " 0\n");
                fprintf(outf, "e ");
            }
            for (int g = pGloProb->FirstGateVar; g < pGloProb->NextNewVar; g += 2) {
                fprintf(outf, " %i", decv(g));
            }
        }
        fprintf(outf, " 0\n");
    }

    int NumClauses = 1;
    fprintf(outf, "%i 0\n", decv(pGloProb->OutGlit));

    for (int g = absv(pGloProb->NextNewVar)-2; g >= pGloProb->FirstGateVar; g = absv(g) - 2) {
        //if (VarPars[g].n == 0 && absv(g) != absv(pGloProb->OutGlit)) continue;
        if (IsOrGate(g)) {
            g = negv(g);
        }
        auto(args, dup_as_garr(GlitToGs[g]->ReqLits));
        sort(args.begin(), args.end());
        // (v <==> (x1 & x2 & x3)) expands to
        // (v | ~x1 | ~x2 | ~x3)  &  (~v | x1)  &  (~v | x2)  &  (~v | x3)
        fprintf(outf, "%i ", decv(g));
        for (int i=0; i < args.n; i++) {
            int CurArg = args[i];
            if (HasLitOrNeg(CurArg)) continue;
            fprintf(outf, "%i ", decv(negv(CurArg)));
        }
        NumClauses++;
        fprintf(outf, " 0\n");
        for (int i=0; i < args.n; i++) {
            int CurArg = args[i];
            if (HasLitOrNeg(CurArg)) continue;
            fprintf(outf, "%i %i 0\n", decv(negv(g)), decv(CurArg));
            NumClauses++;
        }
        args.dealloc();
    }
    fflush(outf);
    volnop();
    rewind(outf);
    fprintf(outf, "p cnf %i %i", pGloProb->NextNewVar - 2, NumClauses);
    fclose(outf);
}

// void AsgnT::CalcVarPars(int g, bool* hit) {
//     if (hit[g]) return;
//     hit[g] = true;
//     assert(IsPosLit(g));
//     assert(AbsGlitToGs(g) != NULL);
//     int IsOr = IsOrGate(g);
//     int Lev = 0;
//     foreach (itLit, AbsGlitToGs(g)->ReqLits) {
//         int lit = *itLit;
//         if (HasLitOrNeg(lit)) continue;
//         //if (lit == 5044) {stop();}
//         VarPars[absv(lit)].append(g);
//         ParentLit[lit ^ IsOr].append(g);
//         ParentLit[negv(lit) ^ IsOr].append(negv(g));
//         if (!IsInputLit(lit)) {
//             CalcVarPars(absv(lit), hit);
//             if (HasInner[absv(lit)]) {
//                 HasInner[g] = true;
//             }
//         } else {
//             if (QbFromLit[lit] == InnermostQB) {
//                 HasInner[g] = true;
//             }
//         }
//         Lev = max(Lev, NestLev[absv(lit)]+1);
//     }
//     NestLev[g] = Lev;
//     NestLev[negv(g)] = Lev;
// }

// void AsgnT::WipeOldDefn(int glit) {
//     GsT* pDefLs = GlitToGs[glit];
//     assert(pDefLs != NULL);
//     for (int i=0; i < pDefLs->ReqLits.n + 1; i++) {
//         GsT* pCurLs = pGloProb->clauses[pDefLs->id + i];
//         //assert(absv(pCurLs->GateLit) == absv(glit));
//         pCurLs->inactive = 1;
//         EraseWatchedLits(pCurLs);
//     }
//     //pDefLs->ReqLits.n = 0;
// }

// int AsgnT::LcaCommonPar(int v) {
//     int subroot = Lca[v][0];
//     if (Lca[v].n > 1) {
//         assert(VarPars[Lca[v][0]].n == 1);
//         subroot = VarPars[Lca[v][0]][0];
//         foreach (itCh, Lca[v]) {
//             assert(subroot == VarPars[*itCh][0]);
//         }
//     }
//     if (IsOrGate(subroot)) {
//         subroot = negv(subroot);
//     }
//     return subroot;
// }



SubstCircuitDefn* AsgnT::NewCiruitDefn() {
    SubstCircuitDefn* pDef = new SubstCircuitDefn();
    SubstCircuitDefn& ckt = *pDef;
    for (int glit = 0; glit < GateDefn.n; glit++) {
        garri args = GateDefn[glit];
        if (args.arr == NULL) continue;
        ckt[glit] = vector<LitT>(args.begin(), args.end());
        ckt[negv(glit)] = vector<LitT>(1, OR_GATE);
        foreach (itLit, args) {
            int lit = *itLit;
            if (IsInputLit(lit)) {
                ckt[lit] = vector<LitT>(1, lit);
                ckt[negv(lit)] = vector<LitT>(1, negv(lit));
            }
        }
    }
    return pDef;
}

LitT AsgnT::CktRestrict(SubstCircuitDefn& ckt, int glit) {
    int ret;
    if (ckt.count(glit)) {
        vector<LitT>& sublits = ckt[glit];
        if (sublits.size() == 1) {
            return *sublits.begin();
        } else {
            return glit;
        }
    }
    if (IsOrGate(glit)) {
        return negv(CktRestrict(ckt, negv(glit)));
    }
    if (IsInputLit(glit)) {
        ret = glit;
        ckt[glit] = vector<LitT>(1, glit);
        return ret;
    }
    assert(0 < glit && glit < GateDefn.n);
    assert(GateDefn[glit].arr != NULL);
    vector<LitT> sublits = vector<LitT>();
    foreach (itArg, GateDefn[glit]) {
        sublits.push_back(CktRestrict(ckt, *itArg));
    }
    ckt[glit] = vector<LitT>(sublits.begin(), sublits.end());
    if (sublits.size() == 1) {
        return *sublits.begin();
    } else {
        return glit;
    }
}   

bool AsgnT::HasMoreVars() {
    int align = 4;
    return (pGloProb->NextNewVar + align < pGloProb->LastPlit);
}

int IsInCegar = false;
garr<GsT*> TempNewClauses = garr<GsT*>(0);

int AsgnT::CegarExtend(garri AsgnLits, map<LitT,LitT>* hit, int ElimQb) {
    int NewOut = 0;
    int winr = QtypeNumFromLetr(pGloProb->QuantBlk[ElimQb].qtype);
    assert(winr==0 || winr==1);
    garr<garri> AsgnInQb = garr<garri>(pGloProb->QuantBlk.n);
    timsort(AsgnLits);
    foreach (itLit, AsgnLits) {
        AsgnInQb[QbFromLit[*itLit]].append(*itLit);
    }
    garri AsgnReps = garri(0);
    foreach (itBlk, AsgnInQb) {
        if (true || itBlk->n <= 3) {
            foreach (itLit, *itBlk) {
                AsgnReps.append(*itLit);
            }
        } else {
            AsgnReps.append(NewConj(0, *itBlk));
        }
    }
    assert(GetCurDL() == 0);
    assert(MinPropQb == 0);
    foreach (itLit, AsgnReps) {
        if (HasLit(negv(*itLit))) {
            static int AlreadyWarned = false;
            if (AlreadyWarned == false) {
                AlreadyWarned = true;
                printf("Oh no!  Cegar failed on line %i.\n", __LINE__);
            }
            goto CegarFail;
        }
        if (HasLit(*itLit)) continue;
        if (HasConflict()) break;
        assert(MinPropQb <= QbFromLit[*itLit]);
        MinPropQb = QbFromLit[*itLit];
        AppendChlit(*itLit);
        SchedForProp(*itLit);
        Propagate();
    }

    IsInCegar = true;

    {
    {
    int flag = 1;
    MinPropQb = ElimQb;
    while (flag) {
        flag = 0;
        garr<signed char>& MonoForce = *GetMonoLits();
        for (int lit=0; lit <= pGloProb->RawLastInpLit; lit++) {
            if (HasConflict()) {
                flag = false;
                break;
            }
            if (MonoForce[lit] && MinPropQb <= QbFromLit[lit]) {
                //if (GloDbgLev >= 2000) printf("Mono forced %i\n", lit);
                if (HasLitOrNeg(lit)) continue;
                AppendChlit(lit);
                SchedForProp(lit);
                Propagate();
                flag = true;
            }
        }
        MonoForce.dealloc();
    }
    MinPropQb = 0;
    }
    foreach (itChlit, chlits.flat) {
        garri& CurUndo = UndoListByDL[LitToDL[*itChlit]];
        assert(CurUndo[0] == *itChlit);
        foreach (itLit, CurUndo) {
            int lit = *itLit;
            if (lit == 0) continue;
            if (IsInputLit(lit)) {
                // if (LitToDL[lit] != 0 && !AsgnLits.bin_search(lit)) {
                //     volnop();
                // }
                (*hit)[lit] = 1;
                (*hit)[negv(lit)] = 0;
            }
        }
    }
    try {
        if (HasConflict()) {
            goto CegarFail;
        }
        NewOut = Restrict(pGloProb->OutGlit, hit, ElimQb);
    } catch (char const* s) {
        fprintf(stderr, "%s\n", s);
        goto CegarFail;
    }
    restart();
    MinPropQb = 0;
    garri NewArgs = garri(0);
    int NumNewLrn = 0;
    /*
    foreach (itLit, AsgnLits) {
        assert(QtypeNumFromLit(*itLit) == winr);
    }
    foreach (itCl, pGloProb->clauses) {
        GsT* pCurLs = *itCl;
        if (pCurLs == NULL) continue;
        if (!pCurLs->IsLearned) continue;
        NewArgs.n = 0;
        bool MakeNew = false;
        foreach (itLit, pCurLs->ReqLits) {
            if ((*hit).count(*itLit) == 0) {
                MakeNew = false;
                break;
            }
            if (HasLit(negv(*itLit)) || (*hit).count(*itLit) == 0) {
                MakeNew = false;
                break;
            }
            if (HasLit(*itLit)) {
                if (LitToDL[*itLit] != 0 && !AsgnLits.bin_search(*itLit)) {
                    MakeNew = true;
                }
            } else {
                NewArgs.append(*itLit);
            }
        }
        foreach (itLit, pCurLs->ImpLits) {
            if (HasLit(negv(*itLit))) {
                MakeNew = false;
                break;
            }
        }
        if (MakeNew) {
            foreach (itRep, AsgnReps) {
                if (QbFromLit[*itRep] == ElimQb) continue;
                else assert(QbFromLit[*itRep] < ElimQb);
                NewArgs.append(*itRep);
            }
            NewArgs.extend(pCurLs->ImpLits);
            GsT* pLrn = NewLrnGs_Mixed(NewArgs, pCurLs->FreeFmla);
            pLrn->IsLearned = true;
            pLrn->bump_activity();
            TempNewClauses.append(pLrn);
            NumNewLrn++;
        }
    }
    */
    foreach (itCl, TempNewClauses) {
        SimpFixWatch(*itCl);
    }
    max_learnts += NumNewLrn;
    //stop();
    TempNewClauses.n = 0;
    IsInCegar = false;
    NewArgs.dealloc();
    }

    CegarFail: ;
    restart();
    MinPropQb = 0;

    foreach (itQb, AsgnInQb) {
        (*itQb).dealloc();
    }
    AsgnInQb.dealloc();
    AsgnReps.dealloc();
    return NewOut;
}

garr<signed char>* AsgnT::GetMonoLits() {
    static garr<signed char> LitPos;
    static garr<signed char> MonoForce;
    if (HasConflict()) {
        return &MonoForce;
    }
    assert(LitPos.n == 0);
    assert(MonoForce.n == 0);
    MonoForce.resize_bzero(pGloProb->LastOrigGlit + 1);
    LitPos.resize_bzero(pGloProb->LastPlit + 1);
    LitPos[pGloProb->OutGlit] = 1;
    for (int glit = pGloProb->LastOrigGlit; glit > 0; glit--) {
        // if ((glit & ~3) == 28984*2) {
        //     volnop();
        // }
        if (!IsAndGate(glit)) continue;
        if (HasLit(glit) && HasLit(glit ^ 2)) continue;
        if (HasLit(negv(glit)) && HasLit(negv(glit ^ 2))) continue;
        int LiveGlitPos = (LitPos[glit] == 1);
        int LiveGlitNeg = (LitPos[negv(glit)] == 1);
        foreach (itArg, GateDefn[glit]) {
            int arg = *itArg;
            assert(LitPos[arg] != -1);       
            assert(LitPos[negv(arg)] != -1); 
            if (absv(arg) == absv(glit)) abort();
            if (IsInputLit(arg)) assert(!HasLit(negv(arg)));
            assert(arg < glit);
            if (LiveGlitPos) {LitPos[arg] = 1;}
            if (LiveGlitNeg) {LitPos[negv(arg)] = 1;}
        }
        if (!LiveGlitPos) {LitPos[glit] = -1;}
        if (!LiveGlitNeg) {LitPos[negv(glit)] = -1;}
    }
    MonoForce.resize_bzero(pGloProb->LastOrigGlit + 1);
    for (int lit = 0; lit <= pGloProb->RawLastInpLit; lit++) {
        if (HasLitOrNeg(lit)) continue;
        if (LitPos[lit] && !LitPos[negv(lit)]) {
            int flit = (QtypeNumFromLit(lit) ? lit : negv(lit));
            MonoForce[flit] = 1;
        }
    }

    LitPos.n = 0;
    return &MonoForce;
}


int AsgnT::Restrict(int glit, map<LitT,LitT>* hit, int ElimQb) {
    int ret = 0xdeadbeef;
    {
    if (hit->count(glit)) {
        if (hit->count(negv(glit))) {
            assert((*hit)[glit] != (*hit)[negv(glit)]);
        }
        return (*hit)[glit];
    }
    if (IsOrGate(glit)) {
        ret = negv(Restrict(negv(glit), hit, ElimQb));
        return ret;
        //goto finish_restrict;
    }
    //int WinrQtype = pGloProb->QuantBlk[ElimQb].qtype;
    if (IsInputLit(glit)) {
        if (QbFromLit[glit] < ElimQb) {
            ret = glit;
        } else if (QbFromLit[glit] == ElimQb) {
            // int plit = find_var_pol(AsgnLits, glit);
            // if      (plit == glit)       {ret = 1;}
            // else if (plit == negv(glit)) {ret = 0;}
            // else 
            //ret = glit; 
            //ret = rand() % 2;
            ret = 0;
        } else {
            ret = GenNewVar(glit);
            int NewQb = QbFromLit[glit] - 2;
            add_var_to_qblock(absv(ret), NewQb);
            GhostToGs[ret] = INPUT_LIT;
            GhostToGs[negv(ret)] = INPUT_LIT;
            assert(QtypeFromLit(ret) == QtypeFromLit(glit));
            if (GloDbgLev >= 2002) {
                printf("  New CEGAR input var ");
                DumpLitRaw(ret);
                printf(" from ");
                DumpLitRaw(glit);
                printf("\n");
            }
        }
        goto finish_restrict;
    }
    foreach (itArg, GateDefn[glit]) {
        auto(itCached, hit->find(*itArg));
        if (itCached != hit->end()) {
            int NewArg = itCached->second;
            if (NewArg == 0) {
                ret = 0;
                goto finish_restrict;
            }
        }
    }
    garri NewArgs = garri(0,0); 
    bool IsDiff = false;
    foreach (itArg, GateDefn[glit]) {
        int NewArg = Restrict(*itArg, hit, ElimQb);
        if (NewArg == 0) {
            NewArgs.dealloc();
            ret = 0;
            goto finish_restrict;
        }
        if (NewArg == 1) {
            IsDiff = true;
            continue;
        }
        if (NewArg != *itArg) {IsDiff = true;}
        NewArgs.append(NewArg);
    }
    if (NewArgs.n == 0) {
        NewArgs.dealloc();
        ret = 1;
        goto finish_restrict;
    }
    //NewArgs.sort(CmpLitByNarf);
    timsort(NewArgs);
    NewArgs.remove_adj_dups();
    for (int i=1; i < NewArgs.n; i++) {
        if (NewArgs.arr[i-1] == negv(NewArgs.arr[i])) {
            NewArgs.dealloc();
            ret = 0;
            goto finish_restrict;
            //(*hit)[glit] = 0;
            //return 0;
        }
    }
    if (NewArgs.n == 1) {
        ret = NewArgs[0];
        NewArgs.dealloc();
        goto finish_restrict;
    }
    if (!IsDiff) {
        NewArgs.dealloc();
        ret = glit;
        goto finish_restrict;
    }
    ret = NewConj(glit, NewArgs);
    NewArgs.dealloc();
    }

    finish_restrict:
    //printf("%i => %i\n", glit, ret);
    assert(hit->count(glit) == 0);
    assert(hit->count(negv(glit)) == 0);
    (*hit)[glit] = ret;
    (*hit)[negv(glit)] = negv(ret);
    return ret;
}

garr<GsT*> KillTries;

LitT AsgnT::RestrictVar(LitT split_var, int val) {
    map<LitT,LitT>* pHit = new map<LitT,LitT>();
    assert(val == 0 || val == 1);
    (*pHit)[split_var] = val;
    (*pHit)[negv(split_var)] = negv(val);
    int ElimQb = MAX_QB + 1;
    int ret = Restrict(pGloProb->OutGlit, pHit, ElimQb);
    delete pHit;
    return ret;
}

void AsgnT::SplitOnVar(LitT split_var) {
    GsT* pCurLs[4];
    LitT tbra = RestrictVar(split_var, 1);
    LitT fbra = RestrictVar(split_var, 0);
    int winr[4];
    garri args[4] = {garri(0), garri(0), garri(0), garri(0)};
    int i = 0;
    winr[i] = 1; args[i].append(ghosted(tbra, 1 - winr[i]));       args[i].append(split_var); i++;
    winr[i] = 1; args[i].append(ghosted(fbra, 1 - winr[i]));       args[i].append(negv(split_var)); i++;
    winr[i] = 0; args[i].append(ghosted(negv(tbra), 1 - winr[i])); args[i].append(split_var); i++;
    winr[i] = 0; args[i].append(ghosted(negv(fbra), 1 - winr[i])); args[i].append(negv(split_var)); i++;
    for (i=0; i < 4; i++) {
        pCurLs[i] = NewLrnGs_Mixed(args[i], winr[i]);
        ExecLitSet(pCurLs[i], EM_SCHED);
        Propagate();
    }
    //stop();
}


GsT* AsgnT::Augment(int glit, garri AsgnLits, map<LitT,LitT>* hit, int ElimQb) {
    int winr = QtypeNumFromLetr(pGloProb->QuantBlk[ElimQb].qtype);
    assert(glit == pGloProb->OutGlit);
    // int WinrQtype = pGloProb->QuantBlk[ElimQb].qtype;
    foreach (itLit, AsgnLits) {
        assert(QbFromLit[*itLit] <= ElimQb);
        if (QbFromLit[*itLit] != ElimQb) continue;
        // if (QtypeFromLit(*itLit) != WinrQtype) continue;
        // ElimQb = min(ElimQb, (int) QbFromLit[*itLit]);
        (*hit)[*itLit] = 1;
        (*hit)[negv(*itLit)] = 0;
    }
    //int NewOut = Restrict(glit, hit, ElimQb);
    int NewOut = CegarExtend(AsgnLits, hit, ElimQb);
    garri args = garri(0, 16);
    int WinrOut = -1, LosrOut = -1;
    int FrobOut = (winr==1 ? NewOut : negv(NewOut));
    if (FrobOut == 0) {
        if (GloDbgLev >= 2002) {
            printf("WARNING: Opposing player won in CEGAR learning!\n");
            foreach (itPair, *hit) {
                if (IsInputLit(itPair->first)) {
                    if (itPair->first == itPair->second) continue;
                    int OrigLit = (itPair->first);
                    assert((*hit)[OrigLit] != (*hit)[negv(OrigLit)]);
                    DumpLitRaw(itPair->first);
                    printf(": ");
                    printf("%i", itPair->second);
                    printf(",  ");
                }
            }
            printf("\n");
        }
        stop();
        args.dealloc();
        return 0;
    } else if (FrobOut == 1) {
        args.n = 0;
    } else {
        if (IsReallyGateLit(FrobOut)) {
            LosrOut = ghosted(FrobOut, winr ^ 1);
            WinrOut = ghosted(FrobOut ^ 1, winr);
        } else {
            LosrOut = FrobOut;
            WinrOut = FrobOut;
        }
        args.append(LosrOut);
    }
    foreach (itPair, *hit) {
        int lit = itPair->first;
        int NewLit = itPair->second;
        if (NewLit != 1) continue;
        if (ElimQb <= QbFromLit[lit]) continue;
        //if (!IsInputLit(lit)) continue;
        args.append(lit);
    }
    vector<int> entry = vector<int>(args.arr, args.arr + args.n);
    if (CegLrnSet.count(entry)) {
        args.dealloc();
        return 0;
    } 
    CegLrnSet.insert(entry);
    // GsT* ret = NewLrnGs_Mixed(args, winr);
    GsT* ret = NULL;

    if (FrobOut == 1) {
        static int AlreadyWarned = false;
        if (AlreadyWarned == false) {
            AlreadyWarned = true;
            printf("WARNING: Case where FrobOut==1 is not implemented, aborting CEGAR!\n");
        }
    }

    garri WinrGuard = garri(0);
    foreach (itLit, AsgnLits) {
        if (QbFromLit[*itLit] != ElimQb) {
            WinrGuard.append(*itLit);
        }
    }

    if (FrobOut != 1) {
        //ret = NewLrnGs_Top(LosrOut, winr);
        WinrGuard.append(LosrOut);
        ret = NewLrnGs_Mixed(WinrGuard, winr);
        assert(args[0] == LosrOut);
        args[0] = WinrOut;
        // {
        // args.n = 0;
        // for (int lit=0; lit <= pGloProb->LastPlit; lit += 1) {
        //     if (!IsInputLit(lit)) continue;
        //     if (hit->count(lit) == 0) continue;
        //     int NewLit = (*hit)[lit];
        //     if (NewLit != 1) continue;
        //     args.append(lit);
        // }
        // }
        if (0) {
            //foreach (itLit, AsgnLits) {assert(QbFromLit[*itLit] == ElimQb);}
            GsT* pTriedGs = NewLrnGs_Mixed(args, 1 ^ winr);
            KillTries.append(pTriedGs);
        }
    }

    args.dealloc();
    WinrGuard.dealloc();
    return ret;
}

int AsgnT::NewConj(int OrigGate, garri args) {
    timsort(args);
    args.remove_adj_dups();
    if (args.n == 1) {return args[0];}
    auto(entry, HashedApInt(args));
    if (ArgsToGlit.count(entry)) {
        int ret = ArgsToGlit[entry];
        //stop();
        return ret;
    }
    int ret = GenNewVar(OrigGate, 4);
    add_gate_to_qb(ret, args);
    // for (int qtype=0; qtype <= 1; qtype++) {
    //     add_var_to_qblock(ghosted(absv(ret), qtype), ExtraQb[qtype]);
    // }
    NewDefnGsAlt(ret, args);
    if (GloDbgLev >= 2002) {
        printf("  New CEGAR gate ");
        DumpLitRaw(ret);
        printf(" from ");
        DumpLitRaw(OrigGate);
        printf(", def_gs: %4i", GhostToGs[ret]->id);
        printf(", args: ");
        DumpLitsEnc(args, "");
        printf("\n");
    }
    //assert(QbFromLit[ret] != 2);
    //ArgsToGlit[entry] = ret;
    assert(ArgsToGlit.count(entry));
    //if (GloSanChk > 1) {
        auto(test, HashedApInt(args));
        assert(ArgsToGlit.count(test));
    //}
    return ret;
}

// int AsgnT::ConsiderExpansion(int JustDump=false, int begin=2) {
//     static bool FirstTime = true;
//     if (HasLitOrNeg(pGloProb->OutGlit)) return 0;
//     InitPart2();
//     int BestSplit = 0;
//     int BestScore = -(1<<30);
//     for (int s=begin; s <  pGloProb->LastInpLit; s += 2) {
//         if (HasLitOrNeg(s)) continue;
//         if (IsAlreadyExp(s)) continue;
//         if (VarPars[s].n == 0) continue;
//         if (QbFromLit[s] < InnermostQB) continue;
//         //if (Lca[s][0] == pGloProb->OutGlit && QbFromLit[s] < InnermostQB) continue;
//         //printf(" (narf %i) ", s);
//         PickExpSplit(s);
//         int subroot = LcaCommonPar(s);
//         uint NumOldNodes = ExpCh[0].size();
//         int CurScore = -(NumNewNodes - NumOldNodes);
//         // if (NumBigBt < 20) CurScore -= 900;
//         // if (TimesInConflict[s] < 1) CurScore -= 900;
//         // if (TimesInConflict[s] < 2) CurScore -= 2;
//         // if (TimesInConflict[s] < 4) CurScore -= 1;
//         if (JustDump) {
//             if (GloDbgLev >= 1100) {
//                 assert(ExpCh[1].size() == NumOldNodes);
//                 printf("Split[%3i]: %4i - %4i = %4i  (LCA: %5i) (TimesInConflict: %i)\n", 
//                     s, NumNewNodes, NumOldNodes, -CurScore, subroot, TimesInConflict[s]);
//             }
//             continue;
//         }
//         if (IsConstExpDep(GetExpCh(0, subroot)) || IsConstExpDep(GetExpCh(1, subroot))) {
//             ExecSplit();
//             return s;
//         }
//         bool LcaHasInner = false;
//         foreach (itLca, Lca[s]) {
//             if (IsInputLit(*itLca)) {
//                 if (absv(*itLca) != absv(s)) {
//                     LcaHasInner = true;
//                     break;
//                 }
//                 continue;
//             }
//             foreach (itArg, AbsGlitToGs(*itLca)->ReqLits) {
//                 if (absv(*itArg) == absv(s)) continue;
//                 LcaHasInner |= HasInner[*itArg];
//             }
//         }
//         if (CurScore >= -10 || (FirstTime && (VarPars[s].n==2 || !LcaHasInner))) {
//             //PickExpSplit(s);
//             ExecSplit();
//             return s;
//         }
//         if (BestScore <= CurScore) {
//             BestScore = CurScore;
//             BestSplit = s;
//         }
//         // PrintAllExpCh(0);
//         // PrintAllExpCh(1);
//     }
//     if (!JustDump) FirstTime = false;
//     if (BestScore > (NumBigBt > 800 ? -30 : -10)) {
//         //printf("Best SPLIT score: %i\n", BestScore);
//         return 0;
//         PickExpSplit(BestSplit);
//         ExecSplit();
//     } else {
//         return 0;
//     }
// }

// void AsgnT::PickExpSplit(int s) {
//     CalcLca(s);
//     CurExpSplit = s;
//     NumNewNodes = 0;
//     // for (int pol=0; pol <= 1; pol++) {
//     //     foreach (itCur, ExpCh[pol]) {
//     //         auto(pCur, itCur->second);
//     //         if (IsConstExpDep(pCur)) continue;
//     //         delete pCur;
//     //     }
//     // }
//     ExpCh[0].clear();
//     ExpCh[1].clear();
//     ExpSing[0].clear();
//     ExpSing[1].clear();
//     int subroot = LcaCommonPar(s);
//     GetExpCh(0, subroot);
//     GetExpCh(1, subroot);
//     // foreach (itLca, Lca[CurExpSplit]) {
//     //     GetExpCh(0, *itLca);
//     //     GetExpCh(1, *itLca);
//     // }
//     //GateZap.clear();
// }

// garri AsgnT::GetPossAnc(int v) {
//     assert(IsPosLit(v));
//     if (PossAnc[v].n > 0) {return PossAnc[v];}
//     if (v == pGloProb->OutGlit) {
//         PossAnc[v].init(0,1);
//         PossAnc[v].append(absv(v));
//         return PossAnc[v];
//     }
//     //set<LitT> AncSet = set<LitT>();
//     PossAnc[v].n = 0;
//     foreach (itPar, VarPars[v]) {
//         PossAnc[v].append(absv(*itPar));
//         auto(ParAnc, GetPossAnc(*itPar));
//         //AncSet.insert(absv(*itPar));
//         //AncSet.insert(ParAnc.begin(), ParAnc.end());
//         foreach (itAnc, ParAnc) {
//             PossAnc[v].append(*itAnc);
//         }
//     }
//     sort(PossAnc[v].begin(), PossAnc[v].end());
//     auto(OldEnd, PossAnc[v].end());
//     auto(NewEnd, unique(PossAnc[v].begin(), PossAnc[v].end()));
//     PossAnc[v].n -= (OldEnd - NewEnd);
//     // assert(AncSet.size() == PossAnc[v].n);
//     // int i=0;
//     // foreach (itAnc, AncSet) {
//     //     assert(PossAnc[v][i++] == *itAnc);
//     // }
//     return PossAnc[v];
// }
// 
// int AsgnT::OccPolarity(int v) {
//     if (v == pGloProb->OutGlit) {return 1;}
//     if (VarPars[absv(v)].n != 1) {return 0;}
//     int par = VarPars[absv(v)][0];
//     bool is_pos = (ParentLit[v][0] == par);
//     bool is_neg = (ParentLit[negv(v)][0] == par);
//     assert(is_pos != is_neg);
//     int ParPol = OccPolarity(par);
//     return is_pos ? ParPol : -ParPol;
// }

// int AsgnT::GetCommonParent(set<LitT>* pAnc) {
//     if (pAnc->size() == 1) {
//         int ret = *pAnc->begin();
//         if (true /*pNonElig->count(ret)==0*/) {
//             int pol = OccPolarity(ret);  // 1 (pos), -1 (neg), or 0 (both)
//             if (pol == 1) {return ret;}
//             if (pol == -1) {return ret;}
//         }
//         if (pAnc->count(pGloProb->OutGlit)) {
//             return pGloProb->OutGlit;
//             //return 999999999;
//         }
//         return 0;
//     }
//     if (pAnc->count(pGloProb->OutGlit)) {
//         if (*pAnc->begin() == pGloProb->OutGlit) {
//             return -(*(++(pAnc->begin())));
//         }
//         return 0;
//     }
//     assert(!pAnc->empty());
//     int first = *pAnc->begin();
//     int common_par = VarPars[first][0];
//     if (AbsGlitToGs(common_par)->ReqLits.n == (int) pAnc->size()) {
//         return 0;
//     }
//     foreach (itAnc, *pAnc) {
//         assert(*itAnc != 0);
//         //if (*itAnc != pGloProb->OutGlit && pNonElig->count(*itAnc)) {return -*itAnc;}
//         if (VarPars[*itAnc].n != 1) {return -*itAnc;}
//         if (NestLev[VarPars[*itAnc][0]] < NestLev[common_par]) {return -*itAnc;}
//         if (VarPars[*itAnc][0] != common_par) {return 0;}
//     }
//     if (pAnc->size() > 1) {
//         //return 0;
//     }
//     if (OccPolarity(common_par) == 0) return 0;
//     //if (pNonElig->count(common_par) == 1) return 0;
//     return common_par;
// }

// int ShowLca = false;

// void AsgnT::CalcAllLca() {
//     for (int ixQb = pGloProb->QuantBlk.n - 1; ixQb > 0; ixQb--) {
//         foreach (itVar, pGloProb->QuantBlk[ixQb].lits) {
//             if (IsAlreadyExp(*itVar)) continue;
//             if (VarPars[*itVar].n == 0) continue;
//             CalcLca(*itVar);
//         }
//     }
//     if (ShowLca) {
//         ShowLca = 0;
//     }
//     // for (int v = 2; v <= pGloProb->LastInpLit; v += 2) {
//     //     printf("%i: ", v);
//     //     DumpLitsEnc(Lca[v], "\n");
//     // }
// }

// void AsgnT::CalcLca(int var) {
//     set<LitT> anc = set<LitT>();
//     foreach (itPar, VarPars[var]) {
//         anc.insert(*itPar);
//     }
//     while (true) {
//         // if (var == 112) {
//         //     DumpIntSet(&anc, "\n");
//         // }
//         int x = GetCommonParent(&anc);
//         //printf(" (x=%i) ", x);
//         if (x > 0) break;
//         if (x == 0) {
//             x = *anc.begin();
//         } else {
//             x = -x;
//         }
//         assert(VarPars[x].n > 0);
//         anc.erase(x);
//         VarEsc[x].append(var);
//         foreach (itAA, VarPars[x]) {
//             anc.insert(*itAA);
//         }
//     }
//     if (var == 66) {
//         //stop();
//     }
//     if (GloDbgLev > 1200 && ShowLca) {
//         printf("Lca[%i]: ", var); DumpIntSet(&anc, "\n");
//     }
//     if (anc.size() > 1) {
//         foreach (itAnc, anc) {
//             VarEsc[*itAnc].append(var);
//         }
//     }
//     //Lca[var].realloc_grow(anc.size());
//     foreach (itAnc, anc) {
//         Lca[var].append(*itAnc);
//     }
// }

int AsgnT::GenNewVar(int orig, int align=2) {
    if (align == 4 and pGloProb->NextNewVar % 4 != 0) {
        pGloProb->NextNewVar += 2;
    }
    int ret = pGloProb->NextNewVar;
    // if ((ret & ~3) == 5612) {
    //     volnop();
    // }
    if (ret + align > pGloProb->LastPlit) {
        throw "Out of space for new variables!";
    }
    assert(QbFromLit[ret] == 0);
    assert(ret % align == 0);
    pGloProb->NextNewVar += align;
    ret = ret ^ signv(orig);
    // OrigGateFromDeriv[ret] = OrigGateFromDeriv[orig];
    // OrigGateFromDeriv[negv(ret)] = negv(OrigGateFromDeriv[orig]);
    // if (align == 4) {
    //     OrigGateFromDeriv[ghosted(ret, 1)] = OrigGateFromDeriv[orig];
    //     OrigGateFromDeriv[negv(ghosted(ret, 1))] = negv(OrigGateFromDeriv[orig]);
    // }
    // NestLev[ret] = NestLev[orig];
    // NestLev[negv(ret)] = NestLev[negv(orig)];
    return ret;
}

// char AsgnT::GetExpCh(int pol, int g) {
//     assert((pol==0) || (pol==1));
//     int s = CurExpSplit;
//     if (IsInputLit(g)) {
//         if (g == negv(s)) {
//             return (pol ? EXPDEP_F : EXPDEP_T);}
//         else if (g == s)  {return (pol ? EXPDEP_T : EXPDEP_F);}
//         else return EXPDEP_0;
//         assert(false);
//     }
//     if (IsOrGate(g)) {
//         auto(InvRet, GetExpCh(pol, negv(g)));
//         if (InvRet == EXPDEP_T) return EXPDEP_F;
//         if (InvRet == EXPDEP_F) return EXPDEP_T;
//         return InvRet;
//     }
//     map<LitT, char>& ExpChPol = ExpCh[pol];
//     auto(itRet, ExpChPol.find(g));
//     if ((itRet != ExpChPol.end()) && (itRet->second != EXPDEP_0)) {
//         auto(ret, itRet->second);
//         return ret;
//     }
//     if (PossAnc[s].bin_search(absv(g)) == false) {
//         return EXPDEP_0;
//     }
//     if (HasLit(g)) {return EXPDEP_T;}
//     if (HasLit(negv(g))) {return EXPDEP_F;}
//     else {
//         ExpChPol[g] = EXPDEP_MOD;
//         int NumCh = 0;
//         int LastCh = 0;
//         foreach (itChild, AbsGlitToGs(g)->ReqLits) {
//             GetExpCh(pol, *itChild);
//         }
//         foreach (itChild, AbsGlitToGs(g)->ReqLits) {
//             int CurChild = *itChild;
//             if (IsOrGate(g)) {CurChild = negv(CurChild); assert(0);}
//             char ChDeps = GetExpCh(pol, CurChild);
//             if (ChDeps == EXPDEP_0) {
//                 //ExpChPol[g]->insert(CurChild);
//                 if (NumCh==1 && LastCh != CurChild) NumCh = 2;
//                 else if (NumCh==0) NumCh = 1;
//                 LastCh = CurChild;
//                 continue;
//             }
//             bool ChT = (ChDeps == EXPDEP_T);
//             bool ChF = (ChDeps == EXPDEP_F);
//             assert(IsAndGate(g));
//             if ((ChT && IsAndGate(g)) || (ChF && IsOrGate(g))) {
//                 continue;
//             }
//             if ((ChF && IsAndGate(g)) || (ChT && IsOrGate(g))) {
//                 //delete ExpChPol[g];
//                 ExpChPol[g] = ChDeps;
//                 return ChDeps;
//             }
//             //ExpChPol[g]->insert(CurChild);
//             if (NumCh==1 && LastCh != CurChild) NumCh = 2;
//             else if (NumCh==0) NumCh = 1;
//             LastCh = CurChild;
//         }
//         // TODO: Check for contradiction (both polarities of a var in the same AND gate)
//         if (NumCh == 0) {
//             auto(ret, (IsAndGate(g) ? EXPDEP_T : EXPDEP_F));
//             //delete ExpChPol[g];
//             ExpChPol[g] = ret;
//             return ret;
//         }
//         if (NumCh == 1) {
//             while (ExpSing[pol].count(LastCh)) {
//                 LastCh = ExpSing[pol][LastCh];
//             }
//             auto(ret, GetExpCh(pol, LastCh));
//             if (ret == EXPDEP_T || ret == EXPDEP_F) {
//                 ExpChPol[g] = ret;
//                 return ret;
//             }
//             ExpSing[pol][g] = LastCh;
//             ExpSing[pol][negv(g)] = negv(LastCh);
//         }
//         /*if (ExpChPol[g]->size() == 1) {
//             int ch = *ExpChPol[g]->begin();
//             auto(ret, GetExpCh(pol, ch));
//             assert(!IsConstExpDep(ret));
//             if (ret == NULL) {
//                 ExpChPol[g]->clear();
//                 ExpChPol[g]->insert(ch);
//             } else if (ExpChPol[g]->size() == 1) {
//                 delete ExpChPol[g];
//                 ExpChPol[g] = ret;
//             }
//         }*/
//         NumNewNodes++;
//         return ExpChPol[g];
//     }
// }

// void AsgnT::PrintExpCh(int pol, int g) {
//     auto(ret, GetExpCh(pol, g));
//     if (ret == EXPDEP_T) {
//         printf("True\n"); return;
//     }
//     if (ret == EXPDEP_F) {
//         printf("False\n"); return;
//     }
//     if (ret == EXPDEP_0) {
//         printf("Unmodified\n"); return;
//     }
//     if (ret == EXPDEP_MOD) {
//         printf("Modified\n"); return;
//     }
//     assert(0);
// }
// 
// void AsgnT::PrintAllExpCh(int pol) {
//     assert((pol==0) || (pol==1));
//     foreach (itPair, ExpCh[pol]) {
//         printf("ExpCh[%i][%i]: ", pol, itPair->first);
//         PrintExpCh(pol, itPair->first);
//     }
// }

// void PrintTimesInConflict() {
//     for (int i=2; i <= pGloProb->LastPlit; i += 2) {
//         printf("TiC[%3i] = %4i\n", i, TimesInConflict[i]);
//     }
// }
//         
// void PrintVsids() {
//     for (int i=2; i <= pGloProb->LastPlit; i += 2) {
//         printf("vsids[%3i(%c)] = %4i\n", i, QtypeFromLit(i), vsids[i]);
//     }
// }
        




/*****************************************************************************
* Functions dealing mainly with game-state objects.    }}}{{{1
*****************************************************************************/

bool GsT::IsBaseGs() {
    return (id <= pGloAsgn->LastInitGs);
}

// int GsT::GetFreeResQb() {
//     foreach (itLit, ImpLits) {
//         if (pGloAsgn->HasLit(*itLit)) continue;
//         return QbFromLit[*itLit];
//     }
//     return MAX_QB;
// }

int GsT::GetWinner() {
    if (FreeFmla==1) return 'E';
    if (FreeFmla==0) return 'A';
    if (FreeFmla==WINR_K) return WINR_K;
    abort();
}

int GsT::IsBlocked() {
    if (HasBlockedImps()) return true;
    return false;
}

int GsT::HasBlockedImps() {
    for (int i=0; i < ImpLits.n; i++) {
        if (pGloAsgn->HasLit(negv(ImpLits[i]))) {
            return ImpLits[i];
        }
    }
    return false;
}

ap<LitT>* GsT::GetWinrLits(int winr) {
    GsT* pCurLs = this;
    return &pCurLs->ImpLits;
}

int AsgnT::GetUpstreamMissingLit(int ForcedLit, GsT* pCurLs, int winr, bool OrNeg) {
    ap<int> WinrLits = *pCurLs->GetWinrLits(winr);
    return GetUpstreamMissingLitAp(ForcedLit, WinrLits, OrNeg);
}

int AsgnT::GetUpstreamMissingLitAp(int flit, ap<int> LitList, bool OrNeg) {
    int WatchQb = QbFromLit[flit];
    for (int i=0; i < LitList.n; i++) {
        int CurLit = LitList[i];
        if (QbFromLit[CurLit] < WatchQb && !(OrNeg ? pGloAsgn->HasLitOrNeg(CurLit) : pGloAsgn->HasLit(CurLit))) {
            return CurLit;
        }
    }
    return 0;
}

int AsgnT::GetFirstFreeLit(ap<int>& LitList) {
    for (int i=0; i < LitList.n; i++) {
        int CurLit = LitList[i];
        if (!pGloAsgn->HasLit(CurLit)) {
            return CurLit;
        }
    }
    return 0;
}

int GsT::HasAllWatchedTrigs() {
    for (int i=0; i < 2; i++) {
        if (!pGloAsgn->HasLit(WatchRawAE[i])) return false;
    }
    return true;
}

int GsT::TrigRipe() {
    if (!pGloAsgn->HasLit(WatchLosrAE(1))) {return 0;}
    if (pGloAsgn->HasLit(negv(WatchLosrAE(0)))) {return 0;}
    if (pGloAsgn->HasLit(WatchLosrAE(0))) {return 2;}
    return 1;
}

int GsT::Is_ExactlyOne_LosrLit_Free() {
    return (pGloAsgn->HasLit(WatchLosrAE(0)) ^ pGloAsgn->HasLit(WatchLosrAE(1)));
}

void GsT::InitTrigLits(ap<int> args) {
    int MaxReqQb = 0, MaxFutQb = 0;
    foreach (itLit, ReqLits) {
        MaxReqQb = max(MaxReqQb, (int)QbFromLit[*itLit]);
    }
    foreach (itLit, ImpLits) {
        MaxFutQb = max(MaxFutQb, (int)QbFromLit[*itLit]);
    }
    //assert(MaxReqQb==0 || MaxFutQb==0 || MaxFutQb < MaxReqQb);
}

int GsT::PrintIndex() {
    // This is for debugging.
    int ret = pGloProb->clauses.indexof(this);
    printf(  "pGloProb->clauses.arr[%i]\n", ret);
    return ret;
}

void GsT::RegisterLitsHave() {
    for (int i=0; i < ReqLits.n; i++) {
        int lit = ReqLits[i];
        pGloProb->LitHave[lit].append(this);
    }
    //volnop();
    // if (pGloAsgn->BattleVars.count(absv(this->GateLit)) == 1) {
    //     IsBattle = true;
    //     assert(this->FreeFmla == 1^signv(this->GateLit));
    // }
}

void GsT::dealloc() {
    assert(!InUse);
    assert(pGloProb->clauses[this->id] == this);
    TempLearntClauses.erase(this);
    auto(pRawInts, this->AsLitVec());
    DupLrnTest.erase(*pRawInts);
    delete pRawInts;
    pGloAsgn->EraseWatchedLits(this);
    this->ReqLits.dealloc();
    this->ImpLits.dealloc();
    pGloProb->clauses[this->id] = NULL;
}

bool GsT::is_subsumed_by(GsT* pOther) {
    if (this->ReqLits.n < pOther->ReqLits.n) {
        return false;
    }
    auto(cur1,  pOther->ReqLits.begin());
    auto(last1, pOther->ReqLits.end());
    auto(cur2,  this->ReqLits.begin());
    auto(last2, this->ReqLits.end());
    while (cur1 != last1 && cur2 != last2) {
        if (*cur1 < *cur2) {return false;}
        else if (*cur2 < *cur1) {cur2++;}
        else { cur1++; cur2++; }
    }
    return true;
}

void GsT::bump_activity() {
    //if (!this->IsLearned) return;
    this->activity += cla_inc;
    if (this->activity > 1e20) {
        pGloAsgn->cut_clause_act();
    }
}

void AsgnT::bump_var_act(LitT var) {
    assert(absv(var) == var);
    VarActivity[var] += var_inc;
    if (VarActivity[var] > 1e20) {
        cut_var_act();
    }
    if (VarActOrder.inHeap(var)) {
        VarActOrder.decrease_key(var);
    }
}

void AsgnT::InsertVarActOrder(LitT var) {
    if (!VarActOrder.inHeap(var)) {
        VarActOrder.insert(var);
    }
}


/*****************************************************************************
* Stuff  }}}{{{1
*****************************************************************************/

// int AsgnT::IsBattleLit(int glit) {
//     return binary_search(BattleVarList.begin(), BattleVarList.end(), absv(glit));
// }


GsT* GsById(int id) {
    /* This function is for use in the debugger. */
    return pGloProb->clauses[id];
}

inline int AsgnT::GetWinner() {
    return (FreeFmla==1 ? 'E' : (FreeFmla==0 ? 'A' : die()));
}


void AsgnT::cut_clause_act() {
    foreach (it, pGloProb->clauses) {
        GsT* pCurLs = *it;
        if (!pCurLs) continue;
        pCurLs->activity *= 1e-20;
    }
    cla_inc *= 1e-20;
    if (cla_inc < 1e-30) {
        cla_inc = 1e-30;
    }
}

void AsgnT::cut_var_act() {
    for (int var=0; var <= pGloProb->LastPlit; var += 2) {
        VarActivity[var] *= 1e-20;
    }
    var_inc *= 1e-20;
    if (var_inc < 1e-30) {
        var_inc = 1e-30;
    }
}

void AsgnT::cut_lit_hit() {
    for (int lit=0; lit <= pGloProb->LastPlit; lit += 1) {
        LitHit[lit] *= 1e-20;
    }
    LitHitInc *= 1e-20;
    if (LitHitInc < 1e-30) {
        LitHitInc = 1e-30;
    }
}

void AsgnT::cla_decay_activity() {
    double clause_decay = 1.003;
    cla_inc *= clause_decay;
}

void AsgnT::var_decay_activity() {
    double var_decay = 1.05;
    var_inc *= var_decay;
}

void DumpIntSet(set<int,struct LitLessByChrono>* pSet, const char* pEnd) {
    auto(iter, pSet->begin());
    printf("[ ");
    for (; iter != pSet->end(); ++iter) {
        printf("%i ", *iter);
    }
    printf("]%s", pEnd);
}

double AsgnT::get_median_clause_adj_activity() {
    garr<double> acts = garr<double>();
    foreach (it, TempLearntClauses) {
        GsT* pCurLs = *it;
        if (pCurLs->InUse) continue;
        acts.append(pCurLs->get_adj_act());
    }
    double median = acts.find_median_destructively();
    acts.dealloc();
    return median;
}

void AsgnT::delete_excess_learned_clauses() {
    int NumLearned = TempLearntClauses.size() - NumLearntLocked;
    if (NumLearned < max_learnts) return;
    max_learnts += (max_learnts / 128) + 1;
    double median = get_median_clause_adj_activity();
    garr<GsT*> killset = garr<GsT*>();
    foreach (it, TempLearntClauses) {
        GsT* pCurLs = *it;
        if (pCurLs->InUse) continue;
        if (pCurLs->get_adj_act() < median) {
            killset.append(pCurLs);
        }
    }
    int NumDel = 0;
    foreach (it, killset) {
        GsT* pCurLs = *it;
        if (pCurLs->InUse) continue;
        if (GloDbgLev >= 2000) {
            int time_unused = NumBigBt - pCurLs->LastUse;
            double norm_act = pCurLs->activity / cla_inc;
            printf("Deleting sequent %6i (size %4i, NumWatFix %4i, norm_act %#7.2e, adj_act %7.2e, last use %5i ago).\n", 
                pCurLs->id, pCurLs->ReqLits.n, pCurLs->NumFixes, norm_act, pCurLs->get_adj_act(), time_unused);
        }
        NumDel++;
        TempLearntClauses.erase(*it);
        pCurLs->dealloc();
    }
    killset.dealloc();
    if (GloDbgLev >= 1900) {
        printf("Deleted %i of %i available sequents, norm_cutoff = %.2e.\n", 
            NumDel, NumLearned, median / cla_inc);
    }
}


void GsT::dump() {
    DumpRaw();
}

void GsT::DumpBrief() {
    printf("  ReqLits "); DumpLitsEnc(ReqLits, ",  ");
    printf("  Imps "); DumpLitsEnc(ImpLits, "\n");
}

void GsT::DumpRaw() {
    if (this == NULL) {printf("(null)\n"); return;}
    if (this == (GsT*)(1)) {printf("(GsT *)(0x1)\n"); return;}
    if (pGloProb->clauses[id] != this) {printf("############ DEAD ##############\n");}
    printf("pGloProb->clauses.arr[%i] (%s).  ", id, (IsLearned ? "learned" : "original"));
    printf("winner=%i('%c').\n", FreeFmla, (FreeFmla==NO_WINR ? '-' : (FreeFmla ? 'E' : 'A')));
    printf("InUse = %i, LastUse = %u, activity = %#.4g\n", InUse, LastUse, activity);
    // printf("GateLit=%i.\n", GateLit);
    printf("ReqLits: "); DumpLitsEnc(ReqLits);
    printf("ImpLits: "); DumpLitsEnc(ImpLits);
    printf("Watch: [%i, %i], %i.\n", 
        decv(WatchLosrAE(0)), decv(WatchLosrAE(1)),
        decv(WatchedRes));
}


        
int AsgnT::ChlitByDL(int dl) {
    return chlits.flat[dl];
}

int AsgnT::QbFromDL(int dl) {
    return QbFromLit[ChlitByDL(dl)];
}

int AsgnT::QtypeFromDL(int dl) {
    return GloQtype(QbFromDL(dl));
}

int AsgnT::GlitToMinChrono(int lit, int ply) {
    int ret = LitToChrono[lit];
    return ret;
}



/*****************************************************************************
* Functions for pending queue of game-states.  }}}{{{1
*****************************************************************************/
void AsgnT::AddPendingLs(GsT* pCurLs, int score) {
    PendingExecs[score].append(AntecedT(pCurLs));
}

// void AsgnT::AddPendingMonot(int flit) {
//     PendingExecs[MONO_GST_PRI].append(AntecedT(0, 0, MONO_ANTECED, flit));
// }

AntecedT AsgnT::PopFirstPendingLs(int flag) {
    if (!PendingExecs[CONFLICT_GST_PRI].is_empty()) {
        return PendingExecs[CONFLICT_GST_PRI].pop();
    }
    if (flag) return AntecedT();
    // if (!PendingExecs[MONO_GST_PRI].is_empty()) {
    //     return PendingExecs[MONO_GST_PRI].pop();
    // }
    if (!PendingExecs[FORC_GST_PRI].is_empty()) {
        return PendingExecs[FORC_GST_PRI].pop();
    }
    // if (!PendingExecs[WIN_GST_PRI].is_empty()) {
    //     return PendingExecs[WIN_GST_PRI].pop();
    // }
    return AntecedT();
}

bool AsgnT::HasPendingForced() {
    if (!PendingExecs[CONFLICT_GST_PRI].is_empty()) return true;
    //if (!PendingExecs[MONO_GST_PRI].is_empty()) return true;
    if (!PendingExecs[FORC_GST_PRI].is_empty()) return true;
    return false;
}


/*****************************************************************************
* Functions for testing if a game-state matches yet.  }}}{{{1
*****************************************************************************/
int AsgnT::HasConflict() {
    //int old = HasLit(TargGate) || HasLit(negv(TargGate));
    int ret = (pConflictGs != NULL);
    assert(ret == (FreeFmla != NOT_DONE));
    //assert(ret == old);
    return ret;
}


/*****************************************************************************
* Functions for watched literals.  }}}{{{1
*****************************************************************************/

// int PickWatchedLitsAE(GsT* pCurLs, int NewWatches[2]) {
// 
//     int* PrevWatched = pCurLs->WatchRawArr;
//         
//     int iw=0;
// 
//     //#define IsUnassigned(lit) (LitToDL[lit] == NO_DECLEV)
//     #define IsUnassigned(lit) ((LitToDL[lit] == NO_DECLEV) || WaitingPropSet.count(lit))
// 
//     int NumLits  =  pCurLs->ReqLits.n;
//     int* PossLits = pCurLs->ReqLits.arr;
//     int start = 0;
//     // if (HasLit(negv(PrevWatched[0])) || HasLit(negv(PrevWatched[1]))) {
//     //     NewWatches[0] = PrevWatched[0];
//     //     NewWatches[1] = PrevWatched[1];
//     //     return 3;
//     // }
//     for (int i=0; i <= 1; i++) {
//         int CurLit = PrevWatched[i];
//         if (IsUnassigned(CurLit)) {
//             NewWatches[iw++] = CurLit;
//         }
//     }
//     if (iw == 2) return 2;
//     start = pCurLs->WatStartPos;
//     for (int ixLit = start; ixLit < NumLits; ixLit++) {
//         int CurLit = PossLits[ixLit];
//         if (IsUnassigned(CurLit)) {
//             if (iw==1 && NewWatches[0] == CurLit) continue;
//             NewWatches[iw++] = CurLit;
//             if (iw == 2) {
//                 pCurLs->WatStartPos = ixLit + 1;
//                 return 2;
//             }
//         } 
//     }
//     for (int ixLit = 0; ixLit < start; ixLit++) {
//         int CurLit = PossLits[ixLit];
//         if (IsUnassigned(CurLit)) {
//             if (iw==1 && NewWatches[0] == CurLit) continue;
//             NewWatches[iw++] = CurLit;
//             if (iw == 2) {
//                 pCurLs->WatStartPos = ixLit + 1;
//                 return 2;
//             }
//         } 
//     }
// 
//     if (NewWatches[0] == 0) {
//         NewWatches[0] = PrevWatched[0];
//         NewWatches[1] = PrevWatched[1];
//     } else  {
//         #ifndef NO_SAN_CHK
//         assert(NewWatches[1] == 0);
//         #endif
//         if (PrevWatched[0] == NewWatches[0]) {
//             NewWatches[1] = PrevWatched[1];
//         } else {
//             NewWatches[1] = PrevWatched[0];
//         }
//     } 
//     #ifndef NO_SAN_CHK
//     if (NewWatches[0] != 0) assert(NewWatches[0] != NewWatches[1]);
//     #endif
//     return 0;
// }


extern inline
int FixOneWatchedLit(GsT* pCurLs, int OldLit) {
    int iOld;
    if      (pCurLs->WatchRawAE[0] == OldLit) {iOld = 0;}
    else if (pCurLs->WatchRawAE[1] == OldLit) {iOld = 1;}
    else {
        assert(OldLit == pCurLs->WatchedRes);
        pGloAsgn->FixWatchedRes(pCurLs);
        return 0;
    }
    int OtherLit = pCurLs->WatchRawAE[1 - iOld];

    if (LitToDL[negv(OtherLit)] != NO_DECLEV) {
        return 0xDEAD;
    }

    NumWatchFixes++;
    // if (1 && pCurLs->NumFix++ > 40) {
    //     pCurLs->NumFix = 0;
    //     vector<pair<double, LitT> > ord;
    //     foreach (itLit, pCurLs->ReqLits) {
    //         ord.push_back(pair<double,LitT>(LitHit[*itLit] / 1000 - LitHit[negv(*itLit)], *itLit));
    //     }
    //     sort(ord.begin(), ord.end());
    //     int i = 0;
    //     foreach (itPair, ord) {
    //         pCurLs->WatOrd[i++] = itPair->second;
    //     }
    //     pCurLs->WatStopBt = 0;
    // }

    if (pCurLs->WatStopBt < NumStepBt) {
        pCurLs->WatStopBt = NumStepBt;
        pCurLs->WatStartPos = 0;
    }
    int NumLits  =  pCurLs->ReqLits.n;
    int* PossLits = pCurLs->ReqLits.arr; // NOTE!!! ReqLits changed in clean_up_sequents
    int ixLit;
    for (ixLit = 0; ixLit < pCurLs->WatStartPos; ixLit++) {
        int CurLit = PossLits[ixLit];
        assert(!(LitToDL[CurLit] == NO_DECLEV && CurLit != OtherLit));
    }
    for (ixLit = pCurLs->WatStartPos; ixLit < NumLits; ixLit++) {
        int CurLit = PossLits[ixLit];
        if (LitToDL[CurLit] == NO_DECLEV && CurLit != OtherLit) {
            pGloAsgn->LitWatchErase(OldLit, pCurLs, pCurLs->itWatchRawAE[iOld]);
            pCurLs->itWatchRawAE[iOld] = pGloAsgn->LitWatchInsert(CurLit, pCurLs);
            pCurLs->WatchRawAE[iOld] = CurLit;
            //pCurLs->WatStartPos = ixLit;
            return 0;
        }
    }
    //pCurLs->WatStartPos = ixLit;
    if (iOld==0 && LitToDL[OtherLit] == NO_DECLEV) {
        swap(pCurLs->WatchRawAE[0], pCurLs->WatchRawAE[1]);
        swap(pCurLs->itWatchRawAE[0], pCurLs->itWatchRawAE[1]);
    }
    return 0;
}


int FixOneWatchedLitBig(GsT* pCurLs, int OldLit) __attribute__ ((noinline));
int FixOneWatchedLitBig(GsT* pCurLs, int OldLit) {
    return FixOneWatchedLit(pCurLs, OldLit);
}

int FixOneWatchedLitSmall(GsT* pCurLs, int OldLit) __attribute__ ((noinline));
int FixOneWatchedLitSmall(GsT* pCurLs, int OldLit) {
    return FixOneWatchedLit(pCurLs, OldLit);
}

// void AsgnT::FixItWatch(GsT* pCurLs, int* NewWatchesB) {
//     GwSetT_erase_iterator itWatchNew[2] = {
//         pCurLs->itWatchRawAE[0],
//         pCurLs->itWatchRawAE[1],
//     };
//     GwSetT_erase_iterator* itWatchOld = pCurLs->itWatchRawAE;
//     for (int i=0; i < 2; i++) {
//         int cur = pCurLs->WatchRawAE[i];
//         if (cur==0) continue;
//         for (int j=0; j < 2; j++) {if (NewWatchesB[j] == cur) goto SkipDel;}
//         LitWatchErase(cur, pCurLs, pCurLs->itWatchRawAE[i]);
//         itWatchNew[i] = GwSetT_erase_iterator();
//         SkipDel:;
//     }
//     for (int i=0; i < 2; i++) {
//         int cur = NewWatchesB[i];
//         if (cur==0) continue;
//         for (int j=0; j < 2; j++) {
//             if (pCurLs->WatchRawAE[j] == cur) {
//                 itWatchNew[i] = itWatchOld[j];
//                 goto SkipAdd;
//             }
//         }
//         itWatchNew[i] = pGloAsgn->LitWatchInsert(cur, pCurLs);
//         SkipAdd:;
//     }
//     for (int i=0; i < 2; i++) {
//         if (NewWatchesB[i] == 0) {
//             #ifndef NO_SAN_CHK
//             assert(pCurLs->WatchRawAE[i] == 0); 
//             #endif
//             continue;
//         }
//         pCurLs->WatchRawAE[i] = NewWatchesB[i];
//         pCurLs->itWatchRawAE[i] = itWatchNew[i];
//     }
// }
// 
// void AsgnT::FixWatchedLits(GsT* pCurLs, int OldLit=-1) {
//     #ifndef NO_SAN_CHK
//     assert(pCurLs->id != 0);  // Can't insert into LitWatch without an id!
//     #endif
//     int NewWatchesB[2] = {0,0};
//     //pCurLs->NumFixes++;
//     if (!USE_PROP_SET) {
//         PickWatchedLitsAE(pCurLs, NewWatchesB);
//         FixItWatch(pCurLs, NewWatchesB);
//     } else if (OldLit == -1) {
//         int OldWat[2] = {pCurLs->WatchRawAE[0], pCurLs->WatchRawAE[1]};
//         if (pCurLs->ReqLits.n >= 1 && HasLit(OldWat[0])) {FixOneWatchedLit(pCurLs, OldWat[0]);}
//         if (pCurLs->ReqLits.n >= 2 && HasLit(OldWat[1])) {FixOneWatchedLit(pCurLs, OldWat[1]);}
//     } else if (OldLit != pCurLs->WatchedRes) {
//         int ok = FixOneWatchedLit(pCurLs, OldLit);
//         if (ok == 0xDEAD) {return;}
//     }
// 
//     FixWatchedRes(pCurLs);
// }

inline void AsgnT::FastFixWatch(GsT* pCurLs, int OldLit) {
    int IsRes = (OldLit == pCurLs->WatchedRes);
    int ok;
    if (pCurLs->ReqLits.n > 20) {
        ok = FixOneWatchedLitBig(pCurLs, OldLit);
    } else {
        ok = FixOneWatchedLitSmall(pCurLs, OldLit);
    }
    assert(pCurLs->WatStartPos == pCurLs->ReqLits.n || pCurLs->ReqLits[pCurLs->WatStartPos] <= max(pCurLs->WatchRawAE[0], pCurLs->WatchRawAE[1]));
    if (ok == 0xDEAD) {return;}
    if (!IsRes) {
        FixWatchedRes(pCurLs);
    }
}

void AsgnT::SimpFixWatch(GsT* pCurLs) {
    int OldWat[2] = {pCurLs->WatchRawAE[0], pCurLs->WatchRawAE[1]};
    if (pCurLs->ReqLits.n >= 1 && HasLit(OldWat[0])) {FixOneWatchedLitSmall(pCurLs, OldWat[0]);}
    if (pCurLs->ReqLits.n >= 2 && HasLit(OldWat[1])) {FixOneWatchedLitSmall(pCurLs, OldWat[1]);}
    FixWatchedRes(pCurLs);
}

void AsgnT::FixWatchedRes(GsT* pCurLs) {
    int ResLitW = 0;
    //if (pCurLs->ImpLits.n > 0 && pCurLs->Is_ExactlyOne_LosrLit_Free()) 
    if (pCurLs->ImpLits.n > 0 && pCurLs->Is_ExactlyOne_LosrLit_Free()) {
        int LastTrig = pCurLs->WatchLosrAE(0);
        ResLitW = GetFirstFreeLit(pCurLs->ImpLits);
        int OldRes = pCurLs->WatchedRes;
        if (OldRes != ResLitW && ResLitW != 0) {
            if (QbFromLit[ResLitW] < QbFromLit[LastTrig]) {
                #ifndef NO_SAN_CHK
                assert(ResLitW == PickNewResWatch(pCurLs));
                #endif
                if (OldRes != 0) {LitWatchErase(OldRes, pCurLs, pCurLs->itWatchRes);}
                pCurLs->itWatchRes = LitWatchInsert(ResLitW, pCurLs);
                pCurLs->WatchedRes = ResLitW;
            }
        }
    }
}

void AsgnT::EraseWatchedLits(GsT* pCurLs) {
    for (int i=0; i < 2; i++) {
        int cur = pCurLs->WatchRawAE[i];
        if (cur==0) continue;
        LitWatchErase(cur, pCurLs, pCurLs->itWatchRawAE[i]);
    }
    for (int i=0; i < 2; i++) {
        pCurLs->WatchRawAE[i] = 0;
    }
    int OldRes = pCurLs->WatchedRes;
    if (OldRes != 0) {
        LitWatchErase(OldRes, pCurLs, pCurLs->itWatchRes);
        pCurLs->WatchedRes = 0;
    }
}


int AsgnT::PickNewResWatch(GsT* pCurLs) {
    int ResLitW = 0;
    if (!pCurLs->Is_ExactlyOne_LosrLit_Free()) return 0;
    int LastTrig = pCurLs->WatchLosrAE(0);
    ResLitW = GetFirstFreeLit(pCurLs->ImpLits);
    if (ResLitW==0) return 0;
    if (QbFromLit[ResLitW] < QbFromLit[LastTrig]) {return ResLitW;}
    else return 0;
}

void AsgnT::ChangeWatchedFut(GsT* pCurLs, int NewFut) {
    int OldFut = pCurLs->WatchedRes;
    if (OldFut == NewFut) return;
    if (OldFut != 0) {pGloAsgn->LitWatchErase(OldFut, pCurLs, pCurLs->itWatchRes);}
    pCurLs->itWatchRes = pGloAsgn->LitWatchInsert(NewFut, pCurLs);
    pCurLs->WatchedRes = NewFut;
}


/*****************************************************************************
* Stuff  }}}{{{1
*****************************************************************************/

int AsgnT::Backtrack(bool KeepDepsLocked=false) {
    LastPickGate = 0;
    GloStep++;
    NumStepBt++;
    // while (!PendingDeadPars.empty()) {
    //     PendingDeadPars.pop();
    // }
    //PendingMonotoneLits.clear();
    int UndoChlit = GetLastChlit();
    uint CurDL = LitToDL[UndoChlit];
    if (GloDbgLev >= 2100) printf("Popping %i (DL %i) (step %i)\n", decv(UndoChlit), CurDL, GloStep);
    assert(CurDL != ZERO_DECLEV);

    int UndoDL = LitToDL[UndoChlit];
    garri& CurUndoList = UndoListByDL[UndoDL];
    //CurUndoList.append(UndoChlit);
    
    (UndoChlit == chlits.pop()) || die();

    bool qPrintRemoved = (GloDbgLev >= 2110);
    if (qPrintRemoved) printf("Removing [ ");
    for (int i = CurUndoList.n - 1; i >= 0; i--) {
        int flit = CurUndoList[i];
        if (qPrintRemoved) printf("%i ", decv(flit));
        assert(LitToDL[flit] == CurDL);
        // if (flit != UndoChlit) {
        //     if (DepByLit[flit].IsUnitProp()) assert(!GetDepByLit(flit)->IsBlocked());
        // }
        //FreeVars.insert(absv(flit));
        LitToDL[flit] = NO_DECLEV;
        LitToChrono[flit] = ((ChronoT) NO_DECLEV);
        PolPref[CurPathNum][absv(flit)] = (1 ^ signv(flit));
        InsertVarActOrder(absv(flit));
        //FreeVarsByQb[QbFromLit[flit]].insert(absv(flit));
        CurChrono--;
        ResetDepByLit(flit, KeepDepsLocked);
    }
    if (pConflictGs) {
        // int HasAll = true;
        // foreach (itLit, pConflictGs->ReqLits) {
        //     if (!HasLit(*itLit)) {
        //         HasAll = false;
        //         break;
        //     }
        // }
        int HasAll = HasLit(pConflictGs->WatchRawAE[0]);
        if (!HasAll) {
            pConflictGs->InUse &= ~CurPathMask;
            pConflictGs = NULL;
            pGloAsgn->FreeFmla = NOT_DONE;
            LitsInConflict.clear();
        }
    }

    for (int i=0; i <= MAX_GST_PRI; i++) {
        PendingExecs[i].clear();
    }
    if (qPrintRemoved) printf("]\n");
    UndoListByDL[UndoDL].n = 0;
    
    // if (UndoTarg[UndoDL] != 0) {
    //     TargGate = UndoTarg[UndoDL];
    //     UndoTarg[UndoDL] = 0;
    // }
    // if (GetLastChlit() == 0) assert(TargGate == pGloProb->OutGlit);

    return UndoChlit;
}


void AsgnT::AddUndoLit(int chlit, int flit) {
    int UndoDL = LitToDL[chlit];
    //if (GloSanChk > 0) assert(UndoListByDL[UndoDL].indexof(flit) == -1);
    UndoListByDL[UndoDL].append(flit);
    LitToChrono[flit] = CurChrono;
    ChronoToLit[CurChrono] = flit;
}

garri& AsgnT::GetCurUndoList() {
    return UndoListByDL[GetCurDL()];
}

void AsgnT::AppendChlit(int NewChlit, bool IsRedo=false) {
    CurChrono++;
    NumAppends++;
    assert(!HasLit(negv(NewChlit)));
    assert(!HasLit(NewChlit));
    assert(GetDepByLit(NewChlit)== NULL);
    //assert(chlits.flat.n > 0 && QbFromLit[NewChlit] >= QbFromLit[GetLastChlit()]);
    uint declev = chlits.flat.n;
    if (GloDbgLev >= 2005 && !IsRedo)  printf("Chose Lit %i (at DL #%i)\n", decv(NewChlit), declev);
    chlits.append(NewChlit);
    //FreeVars.erase(absv(NewChlit));
    LitToDL[NewChlit] = declev;
    LitToChrono[NewChlit] = CurChrono;
    //LitToAbsTime[NewChlit] = NumAppends;
    ChronoToLit[CurChrono] = NewChlit;
    UndoListByDL[declev].n = 0;
    UndoListByDL[declev].append(NewChlit);
    assert(ChlitByDL(declev) == NewChlit);
    if ((LitHit[NewChlit] += LitHitInc) > 1e20) {cut_lit_hit();}
    //LastHitDL[declev] = *pNumClauses;
    //FreeVarsByQb[QbFromLit[NewChlit]].erase(absv(NewChlit));
}

int StopPoint;

void AsgnT::AppendForcedLit(int lit, GsT* pRef) {
    CurChrono++;
    NumAppends++;
    /*if (StopPoint == NumAppends) {
        stop();
    }*/
    uint NewDL = GetCurDL();
    int UndoChlit = ChlitByDL(NewDL);
    assert(UndoChlit == GetLastChlit());
    assert(!HasLit(negv(lit)));
    assert(!HasLit(lit));
    //assert(pRef->HasLit(negv(lit)) || pRef->HasLit(lit) || pRef->GateLit==lit);
    SetDepByLit(lit, pRef);
    AddUndoLit(UndoChlit, lit);
    //FreeVars.erase(absv(lit));
    LitToDL[lit] = LitToDL[UndoChlit];
    //LitToAbsTime[lit] = NumAppends;
    if (GloDbgLev >= 2051) {
        int doit = (GloDbgLev >= 2050 || IsInputLit(lit));
        if (doit) {
            printf("Appending literal %i: forced at DL %i (chlit %i) ", 
                    decv(lit), LitToDL[UndoChlit], decv(UndoChlit));
            if (pRef) {printf("by clause[%i].\n", pRef->id);}
            else {printf("by something other than unit prop.\n");}
        }
    }
    if ((LitHit[lit] += LitHitInc) > 1e20) {cut_lit_hit();}
    //FreeVarsByQb[QbFromLit[lit]].erase(absv(lit));
}



/*****************************************************************************
* Monotone literal rule.  }}}{{{1
*****************************************************************************/

#define NO_ALIVE_CACHE 0
#define USE_ALIVE_CACHE 1

// void AsgnT::WatchParSanityChk() {
//     for (int lit=2; lit <= pGloProb->LastInpLit; lit++) {
//         PickParent(lit, 1);
//     }
// }
// 
// void AsgnT::PickParent(int lit, int ChkOnly=0) {
//     if (ChkOnly && (HasLit(lit) || HasLit(negv(lit)))) return;
//     if (ChkOnly==1 && LastChkPar[lit]==NumAppends) return;
//     if (ChkOnly) LastChkPar[lit] = NumAppends;
//     garri& poss = pGloProb->ParentLit[lit];
//     int OldPar = LitToLiveParent[lit];
//     if (IsAlive(OldPar)) {
//         return;
//     }
//     for (int i=0; i < poss.n; i++) {
//         int CurPar = poss[i];
//         if (ChkOnly) {PickParent(CurPar, ChkOnly);}
//         if (IsAlive(CurPar)) {
//             if (ChkOnly) {
//                 if (LitToLiveParent[lit] != CurPar) {printf("...\n");}
//                 assert(LitToLiveParent[lit] == CurPar);
//                 return;
//             }
//             LitToLiveParent[lit] = CurPar;
//             ParentToWatchers[OldPar].erase(lit);
//             ParentToWatchers[CurPar].insert(lit);
//             return;
//         }
//     }
// }
// 
// bool AsgnT::IsAlive(int lit) {
//     if (lit == pGloProb->OutGlit) return true;
//     if (HasLit(lit) || HasLit(negv(lit))) return false;
//     LvCacheEntry& cache = LvCache[lit];
//     if (true || cache.timestamp < NumMonoProps) {
//         if (GloSanChk >= 4) PickParent(LitToLiveParent[lit], 1);
//         cache.IsAlive = IsAlive(LitToLiveParent[lit]);
//         cache.timestamp = NumMonoProps;
//     } else {
//         assert(cache.IsAlive == IsAlive(LitToLiveParent[lit]));
//     }
//     return cache.IsAlive;
// }
// 
// int NumChFix = 0;
// 
// void AsgnT::FixChildrenWatches(int ParLit, bool AndProp) {
//     if (IsInputLit(ParLit)) {return;}
//     NumChFix++;
//     int PrintFix = (GloDbgLev >= 2012);
//     if (PrintFix) printf("FixChild [%5i]: %i: ", NumChFix, ParLit);
//     const int n = ParentToWatchers[ParLit].size();
//     garri FrobLits = garri(n, n);
//     int i=0;
//     foreach (iter, ParentToWatchers[ParLit]) {FrobLits[i++] = (*iter);}
//     assert(n == i);
//     for (i=0; i < n; i++) {
//         int CurLit = FrobLits[i];
//         if (PrintFix) printf(" %i", CurLit);
//         if (HasLit(CurLit) || HasLit(negv(CurLit))) continue;
//         PickParent(CurLit);
//         if (PrintFix) printf("(%i)", LitToLiveParent[CurLit]);
//         if (!IsAlive(CurLit)) {
//             if (IsGateLit(CurLit)) {
//                 PendingDeadPars.push(CurLit);
//                 //FixChildrenWatches(CurLit, AndProp);
//             } else {
//                 assert(IsInputLit(CurLit));
//                 if (AndProp) { //  && IsAlive(negv(CurLit))
//                     if (GloSanChk >= 3) PickParent(CurLit, 1);
//                     //PickParent(negv(CurLit), 1);
//                     int flit = (QtypeFromLit(CurLit)=='E' ? negv(CurLit) : CurLit);
//                     AddPendingMonot(flit);//.second ? printf("YEAH!\n") : 0;
//                     //PendingMonotoneLits.insert(flit);//.second ? printf("YEAH!\n") : 0;
//                 }
//             }
//         }
//     }
//     if (PrintFix) printf("\n");
//     FrobLits.dealloc();
// }


/*****************************************************************************
* Stuff  }}}{{{1
*****************************************************************************/

// int AsgnT::ChooseLitByTree(int qb, int gate, int* pNumAttempts) {
//     const int MAX_ATTEMPTS = 20;
//     int qtype = GloQtype(qb);
//     int ply = (qtype=='E' ? 1 : (qtype=='A' ? 0 : die()));
//     if (qb != QbFromLit[GetLastChlit()]) {
//         LastPickGate = 0;
//     }
//     if (HasLit(LastPickGate) || HasLit(negv(LastPickGate))) {
//         LastPickGate = 0;
//     }
//     if (gate == 0) {
//         if (LastPickGate != 0) {gate = LastPickGate;}
//         else {gate = TargGate ^ (1 - ply);}
//     }
//     int IsOr = !IsAndGate(gate);
//     LastPickGate = 0;
//     GsT* pCurLs = AbsGlitToGs(gate);
//     if (pCurLs->pChoices == 0) {
//         pCurLs->pChoices = new garri(0, pCurLs->ReqLits.n);
//     }
//     garri& choices = *pCurLs->pChoices;
//     choices.n = 0;
//     int suic = rand() % 2;
//     foreach (itTrig, pCurLs->TrigAE[ply]) {
//         int CurLit = *itTrig ^ IsOr;
//         if (QbFromLit[CurLit] > qb) break;
//         if (HasLit(CurLit) || HasLit(negv(CurLit))) continue;
//         if (suic) CurLit = negv(CurLit);
//         choices.append(CurLit);
//     }
//     if (IsAndGate(gate) && choices.n >= 2) {
//         LastPickGate = gate;
//         int PickLit = choices[rand() % choices.n];
//         return PickLit;
//     }
//     foreach (itTrig, pCurLs->TrigG) {
//         int CurLit = *itTrig ^ IsOr;
//         if (HasLit(CurLit) || HasLit(negv(CurLit))) continue;
//         if (suic) CurLit = negv(CurLit);
//         choices.append(CurLit);
//     }
//     while (1) {
//         if (choices.n == 0) return 0;
//         int r = rand() % (2*choices.n);
//         int ixChoice = (r >> 1);
//         bool GiveUp = (r & 1) && (absv(gate) != absv(TargGate));
//         int PickLit = choices.RemoveAndFill(ixChoice);
//         if (IsGateLit(PickLit)) {
//             PickLit = ChooseLitByTree(qb, PickLit, pNumAttempts);
//         } else {
//             return PickLit;
//         }
//         if (PickLit != 0) {
//             return PickLit;
//         }
//         else {
//             if (GiveUp) return 0;
//             if ((*pNumAttempts)++ >= MAX_ATTEMPTS) return 0;
//         }
//     }
// }

int NextChlit_InOrder;

void AsgnT::RandomizePolPref() {
    for (int lit=0; lit <= pGloProb->LastPlit; lit++) {
        PolPref[CurPathNum][lit] = rand() % 2;
    }
}
    
int AsgnT::ChooseRandLit(int preferred, bool NoExpu=false) {
    int UseVsids = 0;
    NextChlit_InOrder = rand() % 2;
    if (rand() % 64 != 0 && NumBigBt > 32) {
        UseVsids = 1;
    }

    NumDecs++;
    if (UseVsids) {
        int ret = 0;
        while (ret == 0) {
            ret = VarActOrder.removeMin();
            if (HasLitOrNeg(ret)) {ret = 0;}
        }
        assert(absv(ret) == ret);
        return ret ^ (1 ^ PolPref[CurPathNum][ret]);
    }

    //UseVsids = 0;  // TODO: Remove this after fixing VSIDS
    static garri choices;
    choices.n = 0;
    // int MaxQb = 0;
    // int MinQb = (1 << 30);
    // int MinLit = 0;
    // int MinQtype = 0;
    QuantPfxT CurPfx = QuantPfx;
    assert(CurPfx.blks.n != 0);
    int PrevQb = QbFromLit[GetLastChlit()];
    int CurQb = 0;
    //int StartLit = CurPfx.blks[0].lits[0];
    //if (QbFromLit[StartLit] <= PrevQb && PrevQb <= CurPfx.blks.last().ix) {
    //    StartLit = pGloProb->QuantBlk[PrevQb].lits[0];
    //}
    //for (int lit=StartLit; lit <= pGloProb->LastInpLit; lit+=2) {
    //    if (QbFromLit[lit] == 0) continue;
    //    if (HasLitOrNeg(lit) || IsAlreadyExp(lit)) continue;
    //    if (MinQb == (1 << 30)) {
    //        MinLit = lit;
    //        MinQb = QbFromLit[lit];
    //        MinQtype = QtypeFromLit(MinLit);
    //    }
    //    if (QtypeFromLit(lit) != MinQtype) {
    //        break;
    //    }
    //    MaxQb = QbFromLit[lit];
    //}
    //if (!AllowExpu && MinQb > CurPfx.blks.last().ix) {
    //    if (!(HasLit(TargGate) || HasLit(negv(TargGate)))) {
    //        //PickNewTarg(); // for debugging
    //    }
    //    for (int g = pGloProb->FirstGateVar; g <= pGloProb->LastGlit; g += 2) {
    //        assert(HasLit(ghosted(g, 0)) == HasLit(ghosted(g, 1)));
    //        assert(HasLit(ghosted(negv(g), 0)) == HasLit(ghosted(negv(g), 0)));
    //    }
    //    assert(0);
    //    return 0;  // This shouldn't happen for prenex instances.
    //}
    //if (MinQb == (1<<30)) {die("Internal error: No free literals remaining?!\n");}

    // int TreeLit = 0;
    //
    // if (LastPickGate != 0 || (rand() % 256) >= 128) {
    //     //if (LastPickGate) stop();
    //     int tmp = 0;
    //     TreeLit = ChooseLitByTree(MinQb, 0, &tmp);
    //     if (TreeLit != 0 && !IsAlreadyExp(TreeLit)) {
    //         AppendChlit(TreeLit);
    //         return TreeLit;
    //     }
    // }

    if (!HasLitOrNeg(preferred) && !IsAlreadyExp(preferred)) {
        choices.n = 1;
        choices[0] = preferred;
        goto ChooseUp;
    }

    foreach (itQb, pGloProb->QuantBlk) {
        if (itQb->ix < PrevQb) continue;
        // int ixQb = itQb->ix;
        // foreach (itLit, FreeVarsByQb[ixQb]) 
        foreach (itLit, itQb->lits) {
            int lit = *itLit;
            //if (QbFromLit[lit] == 0) {abort(); continue;}
            if (CurQb != 0) {
                if (QbFromLit[lit] != CurQb) {break;};
            } 
            if (HasLitOrNeg(lit) || IsAlreadyExp(lit)) {
                //if (CurQb) assert(FreeVarsByQb[CurQb].count(lit) == 0);
                continue;
            }
            if (CurQb == 0) {
                CurQb = QbFromLit[lit];
            }
            //assert(FreeVarsByQb[CurQb].count(lit) == 1);
            // if (QbFromLit[lit] != MinQb) break;
            // if (QbFromLit[lit] > MaxQb && QtypeNumFromLit(lit) != LastWinr) {
            //     continue;
            // }
            // if (UseVsids) {
            //     int* AltVsids = vsids;
            //     //int* AltVsids = TimesInConflict;
            //     if (choices.n==0 || AltVsids[lit] >= AltVsids[choices[0]]) {
            //         if (choices.n==0 || AltVsids[lit] > AltVsids[choices[0]]) {
            //             choices.n = 0;
            //         }
            //     } else {
            //         continue;
            //     }
            // }
            choices.append(lit);
            choices.append(negv(lit));
        }
    }



    int chosen;
    ChooseUp:
    {
        //printf("Choices: "); DumpLitsEnc(choices);
        int x = rand();
        if (choices.n == 0) {
            return 0;
        }
        chosen = choices[x % choices.n];
        // // static garr<GsT*> blocking;
        // // blocking.n = 0;
        // // foreach (itWat, LitWatch[(chosen)]) {
        // //     blocking.append(itWat->second);
        // // }
        // choices.n = 0;
        // foreach (itWat, LitWatch[chosen]) {
        //     GsT* pCurLs = itWat->second;
        //     if (pCurLs->WatchedRes == chosen) continue;
        //     int winr = 1 - QtypeNumFromLit(chosen);
        //     if (HasDeadLits(pCurLs->GetWinrLits(winr))) continue;
        //     if (!HasLit(Phantomize(negv(pCurLs->GateLit), 1 - winr))) continue;
        //     ap<int> WinrLits = *pCurLs->GetWinrLits(winr);
        //     bool GotOne = false;
        //     foreach (itWinrLit, WinrLits) {
        //         if (HasLitOrNeg(*itWinrLit)) continue;
        //         if (IsAlreadyExp(*itWinrLit)) continue;
        //         if (QbFromLit[*itWinrLit] < QbFromLit[chosen]) {
        //             choices.append(*itWinrLit);
        //             choices.append(negv(*itWinrLit));
        //             GotOne = true;
        //         }
        //     }
        //     if (GotOne && pCurLs->OrigOrLrn == ORIG_TAG) {
        //         volnop();
        //     }
        // }
        // if (choices.n > 0) {
        //     assert(0);
        //     assert(QbFromLit[chosen] > MaxQb);
        //     NextChlit_InOrder = false;
        //     goto ChooseUp;
        // }
        // // if (QbFromLit[chosen] > MaxQb) {
        // //     return ChooseRandLit(0, true);
        // // }
    }
    //expu = 0;
    //printf("Choosing %i.\n", decv(chosen));
    return chosen;
}

int AsgnT::GetOutOrderChlit() {
    // int MinQb = (1 << 30);
    // for (int lit=2; lit <= pGloProb->LastInpLit; lit+=2) {
    //     if (HasLitOrNeg(lit) || IsAlreadyExp(lit)) continue;
    //     MinQb = QbFromLit[lit];
    //     break;
    // }
    garri& blk = chlits.flat;
    int CurQb = 0;
    int HeadChlit = 0;
    for (int i=0; i < blk.n; i++) {
        int chlit = blk[i];
        if (CurQb == QbFromLit[chlit]) continue;
        if (CurQb < QbFromLit[chlit]) {
            CurQb = QbFromLit[chlit];
            HeadChlit = chlit;
        } else {
            return HeadChlit;
        }
    }
    return 0;
}


void AsgnT::DumpCurUndo() {
    printf("%i forced lits: ", GetCurUndoList().n);
    DumpLitsEnc(GetCurUndoList());
}

void AsgnT::dump(int GatesToo) {
    printf("Lits:\n");
    garri& blk = chlits.flat;
    int nch = 0;
    for (int i=0; i < blk.n; i++) {
        int chlit = blk[i];
        uint dl = LitToDL[chlit];
        nch = printf("    %6i (QB %i%c, DL %2i): ", decv(chlit), QbFromLit[chlit], tolower(QtypeFromLit(chlit)), dl);

        garri& CurUndo = UndoListByDL[LitToDL[chlit]];
        if (CurUndo.n == 0) {
            printf("(none)\n");
        } else {
            const char* sep = "";
            printf("[");
            foreach (it, CurUndo) {
                if (!GatesToo && IsReallyGateLit(*it)) continue;
                nch += printf("%s", sep);
                nch += DumpLitRaw(*it);
                if (nch > LINE_LEN - 7) {
                    if (nch < LINE_LEN) printf("\n");
                    nch = 0;
                }
                sep = " ";
            }
            printf("]\n");
        }
    }
}
void AsgnT::dump() {
    dump(false);
}

void AsgnT::DumpChlits() {
    garri& blk = chlits.flat;
    printf("[ ");
    for (int i=0; i < blk.n; i++) {
        int chlit = blk[i];
        if (chlit != 0) printf("%i ", decv(chlit));
    }
    printf("]\n");
}

void AsgnT::DumpLits(int narf) {
    garri& blk = chlits.flat;
    printf("[ ");
    for (int i=0; i < blk.n; i++) {
        int chlit = blk[i];
        printf("[ ");
        garri& CurUndo = UndoListByDL[LitToDL[chlit]];
        foreach (it, CurUndo) {
            printf("%i ", *it);
        }
        printf("] ");
    }
    printf("]\n");
}

void AsgnT::DumpQuantPfx() {
    foreach (it, pGloProb->QuantBlk) {
        printf("%2i(%c): ", it->ix, it->qtype);
        DumpLitsEnc(it->lits);
    }
}


void AsgnT::restart() {
    // TODO eventually: optimize this; don't need to pop each level individually.
    while (chlits.flat.n > 1) {
        Backtrack();
    }
    clean_up_sequents();
}

void AsgnT::clean_up_sequents() {
    assert(GetCurDL() == 0);
    foreach (itCl, pGloProb->clauses) {
        GsT* pCurLs = *itCl;
        if (!pCurLs) continue;
        if (pCurLs->InUse) continue;
        bool RemovedLit = false;
        foreach (itLit, pCurLs->ReqLits) {
            if (HasLit(negv(*itLit))) {
                pCurLs->dealloc();
                break;
            }
            if (HasLit(*itLit) && *itLit != pCurLs->WatchRawAE[0] && *itLit != pCurLs->WatchRawAE[1]) {
                *itLit = 0;
                // NOTE: Must also remove it from WatOrd!!!!!!!!
                RemovedLit = true;
            }
        }
        if (RemovedLit) {
            pCurLs->ReqLits.remove_value(0);
            pCurLs->WatStartPos = 0;
            pCurLs->WatStopBt = 0;
        }
    }
}



/*****************************************************************************
* Sanity Check for WatchedLits.    }}}{{{1
*****************************************************************************/
void AsgnT::WatchSanityChk1(GsT* pCurLs) {
        if (pCurLs->inactive) {
            assert(pCurLs->WatchRawAE[0] == 0);
            return;
        }
        //int NewAE[4] = {0,0,0,0};
        int* OldAE = pCurLs->WatchRawArr;
        if (HasLit(OldAE[0]) || HasLit(OldAE[1])) {
            int NumFree = 0;
            foreach (itLit, pCurLs->ReqLits) {
                if (HasLit(negv(*itLit))) {return;}
                if (HasLit(*itLit)) {NumFree++;}
            }
        }
        // PickWatchedLitsAE(pCurLs, NewAE);
        // for (int j=0; j < 2; j++) {
        //     if (NewAE[j] == 0) NewAE[j] = OldAE[j];
        // }
        // if (HasLit(NewAE[0]) ^ HasLit(NewAE[1])) {
        //     assert(NewAE[0] == OldAE[0]);
        //     assert(NewAE[1] == OldAE[1]);
        // }
        if (HasLit(OldAE[0]) || HasLit(OldAE[1])) {
            int OldRes = pCurLs->WatchedRes;
            int NewRes = PickNewResWatch(pCurLs);
            if (NewRes != 0 && OldRes != NewRes) {
                //int losr = 1 - pCurLs->FreeFmla;
                //int LastTrig = NewAE[losr*2 + 0];
                if (!HasLit(NewRes)) assert(!HasLit(OldRes));
            }
            // if (FindFullCtlrs && !HasLit(negv(pCurLs->FirstGateWatch()))) {
            //     assert(pCurLs->ctlr == pCurLs->GetFullCtlr(0,0));
            //     if (pCurLs->ctlr != 0 /*&& i < pGloProb->NumDoneClauses*/) {
            //         assert(FullCtlrToGSes[pCurLs->ctlr].count(pCurLs) == 1);
            //         map<LitT,LitT> NewSemis = map<LitT,LitT>();
            //         if (FindSemiCtlrs) {
            //             GetSemiCtlrs(pCurLs->ctlr, 0, NewSemis); 
            //             foreach (it, NewSemis) {
            //                 assert(pCurLs->ctlrs.count(it->first) == 1);
            //             }
            //         }
            //         //assert(pCurLs->ctlrs == NewSemis);
            //     }
            // }
        }
}
void AsgnT::WatchSanityChk3() {
    for (int i=1; i < pGloProb->clauses.n; i++) {
        GsT* pCurLs = pGloProb->clauses[i];
        if (pCurLs == NULL) continue;
        if (pCurLs->ReqLits.n == 1) continue;
        WatchSanityChk1(pCurLs);
    }
}

void AsgnT::LatentExecsChk() {
    for (int i=1; i < pGloProb->NumDoneClauses; i++) {
        GsT* pCurLs = pGloProb->clauses[i];
        if (pCurLs == NULL) continue;
        if (pCurLs->inactive) {
            assert(pCurLs->WatchRawAE[0] == 0);
            continue;
        }
        int ret = ExecLitSet(pCurLs, EM_DRYRUN);
        assert(ret <= 0);
    }
}

short AsgnT::ReEval(int v) {
    const short UNKNOWN = 7000;
    const short UNASGN  =  900;
    if (v == 0) {
        for (int i=0; i <= pGloProb->LastPlit; i++) {
            ReEvalArr[i] = UNKNOWN;
        }
        ReEval(pGloProb->OutGlit);
        return 0;
    }
    assert(v != 206);
    if (IsInputLit(v)) {
        int ret;
        if (HasLit(v)) ret = 1;
        else if (HasLit(negv(v))) ret = 0;
        else ret = UNASGN;
        ReEvalArr[v] = ret;
        return ret;
    }
    if (ReEvalArr[v] != UNKNOWN) {
        return ReEvalArr[v];
    }
    if (IsOrGate(v)) {
        int ret = negv(ReEval(negv(v)));
        if (ret > 1) {ret = absv(ret);}
        ReEvalArr[v] = ret;
        return ret;
    }
    GsT* pCurLs = GlitToGs[v];
    int ret = 1;
    foreach (itLit, pCurLs->ReqLits) {
        int val = ReEval(*itLit);
        if      (val == UNASGN) {ret = UNASGN;}
        else if (val == 0) {ret = 0; break;}
        else {assert(val == 1);}
    }
    ReEvalArr[v] = ret;
    return ret;
}


/*****************************************************************************
* Functions for Boolean Constraint Propagation proper.    }}}{{{1
*****************************************************************************/



int AsgnT::HasDeadLits(ap<int>* pTrigs) {
    foreach (it, *pTrigs) {
        if (HasLit(negv(*it))) return true;
    }
    return false;
}

// int AsgnT::IsResCompat(ap<int>& TrigLits, ap<int>& ImpLits) {
//     auto(itResCur, ImpLits.begin());
//     auto(itResEnd, ImpLits.end());
//     foreach (itWinrLit, TrigLits) {
//         int negflit = negv(*itWinrLit);
//         itResCur = lower_bound(itResCur, itResEnd, negflit);
//         if (itResCur == itResEnd) break;
//         if (*itResCur == negflit) return false;
//     }
//     return true;
// }

// int AsgnT::InCurQuantPfx(int lit, int TestTarg=0) {
//     if (TestTarg==0) TestTarg = TargGate;
//     QuantPfxT CurPfx = QuantPfx[absv(TestTarg)];
//     if (CurPfx.blks.n == 0) return false;
//     int MinQb = CurPfx.blks[0].ix;
//     int MaxQb = CurPfx.blks.last().ix;
//     return (MinQb <= QbFromLit[lit] && QbFromLit[lit] <= MaxQb);
// }
// 
// int AsgnT::UnderQuantPfx(int lit, int TestTarg) {
//     QuantPfxT CurPfx = QuantPfx[absv(TestTarg)];
//     return (CurPfx.blks[0].ix <= QbFromLit[lit]);
// }


int AsgnT::ExecLitSet(GsT* pCurLs, int mode) {
    if (pConflictGs) {
        static int AlreadyWarned = false;
        if (AlreadyWarned == false) {
            AlreadyWarned = true;
            fprintf(stderr, "WARNING: ExecLitSet called when conflict exists!\n");
        }
        return -25;
    }
    if (mode > EM_SCHED && pCurLs->id >= pGloProb->NumDoneClauses) {abort();}
    if (pCurLs->TrigRipe()==0) return -40;
    if (HasLit(negv(pCurLs->WatchRawAE[0])) || HasLit(negv(pCurLs->WatchRawAE[1]))) {
        return -50;
    }
    if (USE_PROP_SET) {
        if (WaitingPropSet.count(pCurLs->WatchRawAE[0]) ||
            WaitingPropSet.count(pCurLs->WatchRawAE[1]))
        {
            return -30;
        }
    }
    if (pCurLs->HasBlockedImps()) return -60;
    if (NO_UNIV_REDUC && !HasLit(pCurLs->WatchedRes)) return -65;
    //if (HasDeadLits(&pCurLs->ReqLits)) abort();
    NumExecAttempts++;
    int IsDry = (mode == EM_DRYRUN);
    if (pCurLs->HasAllWatchedTrigs()) {
        // if (GloSanChk >= 1) {
        //     assert(GetFirstFreeLit(pCurLs->ReqLits)==0);
        // }
        if (GetFirstFreeLit(pCurLs->ReqLits) != 0)  {return -70;}
        if (IsDry) return 1;
        if (!UseAltProp && mode <= EM_SCHED) {
            if (!IsDry) pGloAsgn->AddPendingLs(pCurLs, CONFLICT_GST_PRI);
            return 1;
        }
        if (HasConflict()) {
            return 1;
        }
        int ok = 1;
        // if (pConflictGs != NULL && pConflictGs != pCurLs) {
        //     ok = 2;
        //     if (pGloAsgn->FreeFmla != pConflictGs->FreeFmla) {ok = false;}
        //     if (pCurLs->ReqLits.n + 10 >= pConflictGs->ReqLits.n) {ok = false;}
        //     if (ok && GloDbgLev >= 100) {
        //         printf("\nExisting:\n");
        //         pConflictGs->dump();
        //         printf("\nAlternative:\n");
        //         pCurLs->dump();
        //         volnop();
        //     }
        // }
        if (ok) {
            pConflictGs = pCurLs;
            assert((pCurLs->InUse & CurPathMask) == 0);
            pCurLs->InUse |= CurPathMask;
            pGloAsgn->FreeFmla = pConflictGs->FreeFmla;
        }
        return 1; 
    } 
    if (mode > EM_SCHED && !PendingExecs[CONFLICT_GST_PRI].is_empty()) {
        pGloAsgn->AddPendingLs(pCurLs, FORC_GST_PRI);
        abort();
        return 0;
    }
    return ExecForced(pCurLs, mode);

}


int AsgnT::ExecForced(GsT* pCurLs, int mode) {

        int IsDry = (mode == EM_DRYRUN);
        
        int winr, losr;
        winr = pCurLs->FreeFmla;
        losr = (1 - pCurLs->FreeFmla);

        if ((pCurLs->TrigRipe() == 1)) {
            
            int ForcedLit = negv(pCurLs->WatchLosrAE(0));
            if (GetUpstreamMissingLit(ForcedLit, pCurLs, winr, false) != 0) {return -__LINE__;}
            if (IsDry) {return ForcedLit;}
            if (QbFromLit[ForcedLit] < MinPropQb) {
                return -__LINE__;
            }
            int cnt = 0;
            foreach (itLit, pCurLs->ReqLits) {
                if (!HasLit(*itLit)) cnt++;
            }
            assert(cnt == 1);
            if (!UseAltProp && mode <= EM_SCHED) {
                if (!IsDry) {
                    pGloAsgn->AddPendingLs(pCurLs, FORC_GST_PRI); 
                }
                return ForcedLit;
            }
            if (UseAltProp) {
                NumProps++;
                pGloAsgn->AppendForcedLit(ForcedLit, pCurLs);
                WaitingPropLits.append(ForcedLit);
                if (USE_PROP_SET) {
                    WaitingPropSet.insert(ForcedLit);
                }
            }
            return ForcedLit;

        } else if (pCurLs->TrigRipe() == 2) {
            return 0;
        } else {
            return -490;
        }
}


// bool FnGreaterGsByAct(const GsT* const& pLs1, const GsT* const& pLs2) {
//     return (pLs1->activity > pLs2->activity);
// }


void AsgnT::SchedForProp(int CurLit) {
    static garr<GsT*> CurLitWatch;  // static for optimization
    assert(HasLit(CurLit));
    assert(CurLitWatch.n == 0);
    CurLitWatch.n = 0;
    foreach (it, pGloAsgn->LitWatch[CurLit]) {
        #if USE_WATCH_SET
        CurLitWatch.append((*it).second);
        #elif USE_WATCH_ARR
        if (*it != NULL) {
            CurLitWatch.append((*it));
        }
        #elif USE_WATCH_LIST
        CurLitWatch.append((*it));
        #endif
    }
    //sort(CurLitWatch.begin(), CurLitWatch.end(), GsGtByAct);
    
    //init_garr_from_set(&CurLitWatch, pGloAsgn->LitWatch[CurLit]);
    //sort(CurLitWatch.begin(), CurLitWatch.end(), FnGreaterGsById);
    //sort(CurLitWatch.begin(), CurLitWatch.end(), FnLessGsById);
    //sort(CurLitWatch.begin(), CurLitWatch.end(), FnGreaterGsByAct);
    int OldN = CurLitWatch.n;
    if (GloDbgLev >= 1800 && CurLitWatch.n != OldN) {
        printf("[%i] ", OldN - CurLitWatch.n);
    }
    for (int ixLs=0; ixLs < CurLitWatch.n; ixLs++) {
        GsT* pCurLs = CurLitWatch[ixLs];
        if (!UseAltProp) {
            pGloAsgn->AddPendingLs(pCurLs, FORC_GST_PRI); 
        }
        assert(CurLit > 0);
        FastFixWatch(pCurLs, CurLit);
        ExecLitSet(pCurLs, EM_SCHED);
        if (HasConflict()) break;
    }
    CurLitWatch.n = 0;
}

int AsgnT::Propagate() {
    static garri LitsForced;

    garr<GsT*> PendingList = garr<GsT*>();

    if (UseAltProp) {
        
        int first = 0;
        while (true) {
            int flit=0;
            for (int i=first; i < WaitingPropLits.n; i++) {
                flit = WaitingPropLits[i];
                if (flit != 0) {
                    WaitingPropLits[i] = 0;
                    if (USE_PROP_SET) {
                        WaitingPropSet.erase(flit);
                    }
                    first = i + 1;
                    break;
                } else {
                    assert(0);
                }
            }
            if (HasConflict()) {
                if (USE_PROP_SET) {
                    //WaitingPropSet.clear();
                    WaitingPropSet.erase_vals(WaitingPropLits);
                }
                WaitingPropLits.n = 0;
                break;
            }
            if (flit == 0) {
                if (USE_PROP_SET) {
                    //WaitingPropSet.clear();
                    WaitingPropSet.erase_vals(WaitingPropLits);
                }
                WaitingPropLits.n = 0;
                return 0;
            }
            SchedForProp(flit);
        }
        
    } else {

        while (true) {
            bool flag = false;
            while (true) {
                AntecedT ante = PopFirstPendingLs(false);
                if (ante == EmptyAnteced) break;
                GsT* pCurLs = ante.pDepGs;
                int muffin = pCurLs->WatchRawAE[0];
                int flit = ExecLitSet(pCurLs, EM_WIN);
                if (flit == 1) {
                    assert(HasConflict());
                } else if (flit > 1) {
                    assert(flit == negv(muffin));
                    flag = true;
                    NumProps++;
                    pGloAsgn->AppendForcedLit(flit, pCurLs);
                    SchedForProp(flit);
                }
                if (HasConflict()) {
                    break;
                }
            }
            if (flag == false) break;
            if (HasConflict()) {
                break;
            }
        }
    }
    return 0;
}



/*****************************************************************************
* Stuff for learning a new game state.    }}}{{{1
*****************************************************************************/


bool DumpResols = true;
//typedef pair<ChronoT,LitT> ChronoLitPair;

struct LitsByChronoIterT {
    set<ChronoT>::iterator it;
    LitT operator*() {return pGloAsgn->ChronoToLit[*it];}
    LitsByChronoIterT& operator--() {--it; return *this;}
    LitsByChronoIterT& operator++() {++it; return *this;}
    bool operator==(LitsByChronoIterT other) {return (it == other.it);}
    bool operator!=(LitsByChronoIterT other) {return (it != other.it);}
    LitsByChronoIterT(set<ChronoT>::iterator _it) : it(_it) { }; 

};

struct LitsByChronoT {
    set<ChronoT> s;
    
    void insert(LitT lit) {s.insert(pGloAsgn->LitToChrono[lit]);}
    void erase(LitsByChronoIterT& it) {s.erase(it.it);}
    size_t size() {return s.size();}
    bool empty() {return s.empty();}
    
    LitsByChronoIterT begin() {return LitsByChronoIterT(s.begin());}
    LitsByChronoIterT end()   {return LitsByChronoIterT(s.end());}
    
    //void erase(int lit) {s.erase(ChronoLitPair(pGloAsgn->LitToChrono[lit], lit));}
    //size_t count(LitT lit) {return s.count(ChronoLitPair(pGloAsgn->LitToChrono[lit], lit));}
};

//typedef set<LitT,struct LitLessByChrono> LitSetByChronoT;
typedef LitsByChronoT LitSetByChronoT;

class TrigHeapT {
    public:
    garr<ChronoT> h;
    void insert(LitT lit) {
        int chrono = pGloAsgn->LitToChrono[lit];
        assert(chrono != NO_DECLEV);
        h.append(chrono);
        push_heap(h.begin(), h.end());
    }

    int pop() {
        assert(h.n > 0);
        ChronoT cret = h[0];
        while (h.n > 0 && h[0] == cret) {
            pop_heap(h.begin(), h.end());
            h.n--;
        }
        return pGloAsgn->ChronoToLit[cret];
    }

    int peek() {
        return pGloAsgn->ChronoToLit[h[0]];
    }

    void dealloc() {
        h.dealloc();
    }

    void dump();

};

void TrigHeapT::dump() {
    garri tmp = h.dup();
    sort(tmp.begin(), tmp.end());
    foreach (it, tmp) {
        *it = pGloAsgn->ChronoToLit[*it];
    }
    DumpLitsEnc(tmp, "\n");
    tmp.dealloc();
}

garri TrigImps;

class GsBuilder {
    public:
    set<LitT> ImpLits;
    hash_set<LitT> RecTrigLits;
    TrigHeapT TrigHeap;
    garri NonElimLits;  // Used only in minimization
    int winr;
    int losr;
    int WinrLetr;

    void init() {
        TrigHeap.h.init();
        TrigImps.n = 0;
    }

    void dealloc() {
        TrigHeap.dealloc();
    }

    /*void init(ap<LitT> BaseLits) {
        int n = BaseLits.n;
        TrigHeap.h.init(0, n + (n >> 2) + 64);
        foreach (itLit, BaseLits) {
            assert(pGloAsgn->HasLit(*itLit));
            TrigInsert(*itLit);
        }
    }*/
    
    void TrigInsert(int lit) {
        // if (pGloAsgn->IsChlit(lit)) {
        //     AddVsids.insert(lit);
        // }
        assert(pGloAsgn->HasLit(lit));
        if (QtypeFromLit(lit) == WinrLetr) {
            if (QbFromLit[lit] >= QuantPfx.blks[0].ix) {
                ImpInsert(lit);
                return;
            }
        }
        // if (LitToDL[lit] == 0) {
        //     return;
        // }
        if (RecTrigLits.count(lit)) {
            return;
        } else {
            RecTrigLits.insert(lit);
            pGloAsgn->bump_var_act(absv(lit));
            TrigHeap.insert(lit);
        }
    }

    void ImpInsert(int lit) {
        assert(QbFromLit[lit] >= QuantPfx.blks[0].ix);
        assert(QtypeFromLit(lit) == WinrLetr);
        ImpLits.insert(lit);
        //assert(!pGloAsgn->HasLit(negv(lit)));
    }

    LitT PopTrig() {
        int ret = TrigHeap.pop();
        //if (GloDbgLev >= 1990) printf("POP  %i\n", ret);
        return ret;
    }

    LitT PeekTrig() {
        int ret = TrigHeap.peek();
        return ret;
    }

    bool TrigEmpty() {
        return TrigHeap.h.n == 0;
    }

    void CopyTrigsTo(garri& TrigLits);
};

void GsBuilder::CopyTrigsTo(garri& TrigLits) {
    int NumLits = TrigLits.n;
    assert(NumLits == TrigHeap.h.n);
    for (int i=0; i < NumLits; i++) {
        TrigLits[i] = pGloAsgn->ChronoToLit[TrigHeap.h[i]];
    }
}

bool AsgnT::HasUIP(GsBuilder& gb) {
    LitT top = gb.TrigHeap.pop();
    LitT penult = gb.TrigHeap.peek();
    gb.TrigHeap.insert(top);
    if (gb.TrigHeap.h.n == 1) return true;
    return (LitToDL[top] != LitToDL[penult]);
}

int AsgnT::GetChronoLastLit(ap<int> Lits) {
    int BestLit=0;
    for (int i=0; i < Lits.n; i++) {
        int CurLit = Lits[i];
        if (LitToChrono[BestLit] < LitToChrono[CurLit]) {
            BestLit = CurLit;
        }
    }
    return BestLit;
}

int AsgnT::GetChronoLastTwoLits(ap<int> Lits, int* p2nd) {
    int BestLit=0, PenLit=0;
    for (int i=0; i < Lits.n; i++) {
        int CurLit = Lits[i];
        if (LitToChrono[BestLit] < LitToChrono[CurLit]) {
            PenLit = BestLit;
            BestLit = CurLit;
        } else if (LitToChrono[PenLit] < LitToChrono[CurLit]) {
            PenLit = CurLit;
        }
    }
    *p2nd = PenLit;
    return BestLit;
}


void AsgnT::PrintDischargeMsg(char const* indent, int CurLit, GsT* pCurLs, int WithLit=0) {
    // if (*pNumClauses == 930 && CurLit==738) {
    //     volnop();
    // }
    /*if (pCurLs->id <= LastInitGs) {
        if (absv(CurLit) == absv(pCurLs->GateLit)) {
            if (pCurLs->ReqLits.n == 1) {
                    fprintf(PrfLog, "%s(gate_disjunct  %5i  %5i)\n", 
                            indent, decv(CurLit), decv(pCurLs->ReqLits[0]));
            } else {
                assert(pCurLs->OrigOrLrn == ORIG_TAG);
                if (WithLit != 0) {
                    fprintf(PrfLog, "%s(gate_disjunct  %5i  %5i)\n", 
                            indent, decv(CurLit), decv(WithLit));
                } else {
                    fprintf(PrfLog, "%s(gate_conjuncts %5i)\n", 
                            indent, decv(CurLit));
                }
            }
        } else {
                    assert(PlyFromLit(CurLit) == 1 - FreeFmla);
                    fprintf(PrfLog, "%s(via_phantom_gs %5i  (defn %i))\n", 
                            indent, decv(Unphantomize(CurLit)), decv(pCurLs->GateLit));
        }
    }
    else */ {
                    fprintf(PrfLog, "%s(via_learned_gs %5i  (LrnGS %i))\n", 
                            indent, decv(CurLit), pCurLs->id);
    }
}

void AsgnT::Resolve(GsBuilder& gb, int CurLit) {
    if (QtypeFromLit(CurLit) == GetWinner()) {
        assert(QbFromDL(LitToDL[CurLit]) <= QbFromLit[CurLit]);
        gb.ImpInsert(CurLit);
        return;
    }
    assert(!IsChlit(CurLit));
    GsT* pCurLs = GetDepByLit(CurLit);
    assert(pCurLs->InUse);
    assert(pCurLs->ReqLits.arr != NULL);
    pCurLs->bump_activity();
    NumResolves++;
    if (NumResolves > pGloProb->LastPlit) {
        die("Detected an infinite loop in constructing a game-state.\n");
    }
    if (PrfLog) PrintDischargeMsg("  ", CurLit, pCurLs);
    int HasSpecial = 0;
    for (int i=0; i < pCurLs->ReqLits.n; i++) {
        int lit = pCurLs->ReqLits[i];
        if (lit == negv(CurLit)) {HasSpecial=lit; continue;}
        assert(PlyFromLit(CurLit) == gb.losr);
        if (HasLit(lit) && (LitToChrono[lit] < LitToChrono[CurLit])) {
            gb.TrigInsert(lit);
        } else {
            assert(0);
            assert(QbFromLit[lit] > QbFromLit[CurLit]);
            assert(PlyFromLit(lit) == gb.winr);
            gb.ImpInsert(lit);
        }

    }
    assert(HasSpecial);
    assert(pGloAsgn->FreeFmla == pCurLs->FreeFmla);
    for (int i=0; i < pCurLs->ImpLits.n; i++) {
        int CurImp = pCurLs->ImpLits[i];
        gb.ImpInsert(CurImp);
        if (QbFromLit[CurLit] < QbFromLit[CurImp]) {
            TrigImps.append(CurImp);
        }
    }
}

int NumRedunCacheHits;
int GloUIP;


/*****************************************************************************
* MakeLearnedGs: Analyze conflict and learns a new game state.    }}}{{{1
*****************************************************************************/
void AsgnT::MakeLearnedGs(garr<GsT*>& MoreLrn) {
    GsBuilder gb = GsBuilder();
    NumResolves = 0;
    gb.winr = FreeFmla;
    gb.losr = 1 - FreeFmla;
    gb.WinrLetr = GetWinner();
    assert(pGloAsgn->HasConflict());

    if (PrfLog) {
        fprintf(PrfLog, "(setf (LrnGS %i) (quote (\n", *pNumClauses);
    }
    pConflictGs->bump_activity();
    GsT* pInitGs = pConflictGs;
    {
        assert(HasConflict());
        if (PrfLog) {fprintf(PrfLog, "  (from_gs     %5i)\n", pInitGs->id);}
        assert(pInitGs != NULL);
        assert(pInitGs->HasAllWatchedTrigs());
        LitsInConflict.insert(GetChronoLastLit(pInitGs->ReqLits));
        foreach (it, pInitGs->ReqLits) {
            gb.TrigInsert(*it);
        }
        foreach (it, pInitGs->ImpLits) {
            gb.ImpInsert(*it);
        }
    }

    if (DumpResols && GloDbgLev >= 2010) {
        if (GloDbgLev >= 2015) {printf("Lits: "); DumpLits(0);}
        //DumpLitsEnc(gb.ConflictAsgn, "\n");
    }

    uint CurDL = GetCurDL();
    int NumCur = 0;
    foreach (itLit, pConflictGs->ReqLits) {
        if (LitToDL[*itLit] == CurDL) {
            NumCur++;
        }
    }
    if (NumCur == 1) {
        volnop();
    }

    bool AlreadyLearnedUIP = false;
    bool AlreadyLearnedSuic = false;
    int CurLit=0;
    int elig = true; //false;
    int winner = GetWinner();
    while (true) {
        if (gb.TrigEmpty()) break;
        CurLit = gb.PeekTrig();
        //TimesInConflict[CurLit]++;
        int QtypeCurDL = GloQtype(QbFromDL(LitToDL[CurLit]));
        int ok;
        int QtypeCurLit = QtypeFromLit(CurLit);
        ok = (QtypeCurLit != winner && QtypeCurDL != winner) && HasUIP(gb) && (CurDL > 0);
        if (ok && elig && 1) {
            foreach (itImp, gb.ImpLits) {
                if (QbFromLit[CurLit] < QbFromLit[*itImp]) continue;
                if (LitToDL[CurLit] <= LitToDL[*itImp]) {
                    if (AllowExpu && IsChlit(CurLit)) break;
                    goto TryAgainUIP;
                }
            }
            if (IsChlit(CurLit)) {
                assert(QtypeNumFromLit(CurLit) == gb.losr);
                break;
            }
            if (!AlreadyLearnedUIP && LitsInConflict.count(CurLit)==0) {
                MoreLrn[LRN_TYPE_2] = CreateGsFromGb(gb);
                MoreLrn[LRN_TYPE_2]->LrnType = LRN_TYPE_2;
                GloUIP = gb.TrigHeap.peek();
                AlreadyLearnedUIP = true;
                AlreadyLearnedSuic = true;
                if (PrfLog) {
                    fprintf(PrfLog, ")))\n");
                    fprintf(PrfLog, "(setf (LrnGS %i) (quote (\n", *pNumClauses+1);
                    fprintf(PrfLog, "  (via_learned_gs (LrnGS %i))\n", *pNumClauses);
                }
                break;
                continue;
            }
            TryAgainUIP:;
        }
        elig = true;
        gb.PopTrig();
        Resolve(gb, CurLit);
    }
    if (PrfLog) {
        fprintf(PrfLog, ")))\n");
    }

    // if (TrigImps.n > 0) {
    //     stop();
    // }
    GsT* pLrn = MoreLrn[LRN_TYPE_2];
    if (MoreLrn[LRN_TYPE_2] == NULL) {
        MoreLrn[LRN_TYPE_1] = CreateGsFromGb(gb);
        MoreLrn[LRN_TYPE_1]->LrnType = LRN_TYPE_1;
        pLrn = MoreLrn[LRN_TYPE_1];
        GloUIP = gb.TrigHeap.peek();
    }
    if (0) {pLrn->dump();}
    gb.dealloc();
    return;

}




/*****************************************************************************
* More stuff related to learning.    }}}{{{1
*****************************************************************************/

// bool AsgnT::IsRedundant(int CurLit, GsBuilder& gb, map<LitT,bool>& cache, ap<LitT>& TrigLits) {
//     int ret2 = IsRedundant2(CurLit, gb, cache, TrigLits);
//     //int ret1 = IsRedundant1(CurLit, gb, cache, TrigLits);
//     //assert(ret1==ret2);
//     cache[CurLit] = ret2;
//     return ret2;
// }
// bool AsgnT::IsRedundant2(int CurLit, GsBuilder& gb, map<LitT,bool>& cache, ap<LitT>& TrigLits) {
//     // Precond: IsInputLit(CurLit)==false
//     auto(CacheIter, cache.find(CurLit));
//     if (CacheIter != cache.end()) {NumRedunCacheHits++; return CacheIter->second;}
//     assert(0); //assert(IsGateLit(CurLit));
//     int IsOr = !IsAndGate(CurLit);
//     GsT* pCurLs = GlitToGs[CurLit ^ IsOr];
//     GsT* pDepLs = GetDepByLit(CurLit);
//     int ret = !IsOr;
//     for (int i=0; i < pCurLs->ReqLits.n; i++) {
//         int lit = pCurLs->ReqLits[i] ^ IsOr;
//         if (lit == negv(CurLit)) {abort();}
//         int HasIt = false;
//         if (binary_search(TrigLits.begin(), TrigLits.end(), lit) == true) {HasIt=1;}
//         else if (IsInputLit(lit) && gb.ImpLits.count(lit) == 1 && gb.ImpLits.count(negv(lit)) == 0) {HasIt=1;}
//         else if (!IsInputLit(lit) && IsRedundant(lit, gb, cache, TrigLits)) {HasIt=1;}
//         else {HasIt = 0;}
//         if (IsOr) {
//             if (!HasIt) continue;
//             else {ret=true; break;}
//         } else {
//             if (HasIt) continue;
//             else {ret=false; break;}
//         }
//     }
//     return ret;
// }
// 
// bool AsgnT::IsRedundant1(int CurLit, GsBuilder& gb, map<LitT,bool>& cache, ap<LitT>& TrigLits) {
//     // Precond: IsInputLit(CurLit)==false
//     auto(CacheIter, cache.find(CurLit));
//     if (CacheIter != cache.end()) {NumRedunCacheHits++; return CacheIter->second;}
//     GsT* pCurLs = GetDepByLit(CurLit);
//     int ret = true;
//     for (int i=0; i < pCurLs->ReqLits.n; i++) {
//         int lit = pCurLs->ReqLits[i];
//         if (lit == negv(CurLit)) {continue;}
//         if (binary_search(TrigLits.begin(), TrigLits.end(), lit) == true) {continue;}
//         if (gb.ImpLits.count(lit) == 1 && gb.ImpLits.count(negv(lit)) == 0) {continue;}
//         if (!IsInputLit(lit) && IsRedundant(lit, gb, cache, TrigLits)) {continue;}
//         ret = false;
//         break;
//     }
//     return ret;
// }

#define REDUN_MODE_1 1
#define REDUN_MODE_2 2

inline bool AsgnT::IsRedundantBySeq(int CurLit, GsBuilder& gb, GsT* pDepLs, int mode) {
    bool HasForced = false;
    int ForcedQb = QbFromLit[CurLit];
    foreach (it, pDepLs->ReqLits) {
        if (*it == negv(CurLit)) {HasForced = true; continue;}
        int ChronoInv = (LitToChrono[CurLit] < LitToChrono[*it]);
        if (mode == REDUN_MODE_1) {
            if (gb.RecTrigLits.count(*it) == 0) {return false;}
            if (ChronoInv) {return false;}
        } else if (mode == REDUN_MODE_2) {
            if (gb.NonElimLits.bin_search(*it) == false) {return false;}
            if (ForcedQb < QbFromLit[*it]) {return false;}
            if (ForcedQb == QbFromLit[*it] && ChronoInv) {return false;}
        }
    }
    assert(HasForced);
    foreach (it, pDepLs->ImpLits) {
        if (gb.ImpLits.count(*it) == 0) {
            return false;
        }
    }
    if (GloDbgLev >= 2001) 
        printf("Removed %i via pGloProb->clauses.arr[%i]\n", CurLit, pDepLs->id);
    return true;
}

bool AsgnT::IsRedundant(int CurLit, GsBuilder& gb, int mode) {
    if (LitToDL[CurLit] == 0) return true;
    GsT* pDepLs = GetDepByLit(CurLit);
    if (pDepLs == NULL) {
        assert(IsChlit(CurLit));
        return false;
    }
    if (IsRedundantBySeq(CurLit, gb, pDepLs, mode)) return true;
    if (mode == REDUN_MODE_2) {
        foreach (it, pGloAsgn->LitWatch[negv(CurLit)]) {
            GsT* pCurLs = *it;
            if (pCurLs == NULL) continue;
            if (IsRedundantBySeq(CurLit, gb, pCurLs, mode)) {
                return true;
            }
        }
    }
    return false;
}

void AsgnT::MinimizeConflict(GsBuilder& gb, garri& TrigLits) {
    /* Sorensson & Biere, "Minimizing Learned Clauses", SAT 2009. */
    map<LitT, bool> cache;
    //garri orig = TrigLits.dup();
    for (int i=0; i < TrigLits.n; i++) {
        int CurLit = TrigLits[i];
        if (IsRedundant(CurLit, gb, REDUN_MODE_1)) {
            TrigLits[i] = 0;
        }
    }
    int NumOrig = TrigLits.n;
    TrigLits.remove_value(0);
    int delta = NumOrig - TrigLits.n;
    if (delta) {
        //printf("Removed %i\n", delta);
    }
    //orig.dealloc();
}

void GsT::InitWatchedLitsPart1() {
    GsT* pCurLs = this;
    assert(LitToDL[0] == 0);
    //pCurLs->WatchLosrAE(0) = pCurLs->WatchLosrAE(1) = 0;
    int WatchW[2] = {0,0};
    //int ixWat[3] = {0,0,0};
    for (int i=0; i < pCurLs->ReqLits.n; i++) {
        int wlit = pCurLs->ReqLits[i];
        if (LitToDL[wlit] > LitToDL[WatchW[0]]) {
            WatchW[1] = WatchW[0];  //ixWat[1] = ixWat[0];
            WatchW[0] = wlit;       //ixWat[0] = i;
        } else
        if (LitToDL[wlit] > LitToDL[WatchW[1]]) {
            WatchW[1] = wlit;       //ixWat[1] = i;
        }
    }
    if (pCurLs->ImpLits.n > 0) {
        int BestFut = pCurLs->ImpLits[0];
        foreach (itFut, pCurLs->ImpLits) {
            if (LitToDL[*itFut] > LitToDL[BestFut]) {
                BestFut = *itFut;
            }
        }
        pCurLs->WatchedRes = BestFut;
    }
    pCurLs->WatOrd.alloc(max(2, pCurLs->ReqLits.n));
    for (int i=0; i < ReqLits.n; i++) {
        WatOrd[i] = ReqLits[i];
        // if (i > 0 && WatOrd[i] == WatchW[0]) {swap(WatOrd[i], WatOrd[0]);}
        // if (i > 1 && WatOrd[i] == WatchW[1]) {swap(WatOrd[i], WatOrd[1]);}
    }
    pCurLs->WatchRawAE[0] = WatchW[0];
    pCurLs->WatchRawAE[1] = WatchW[1];
    // if (pCurLs->WatOrd.n >= 1) {swap(pCurLs->WatOrd[0], pCurLs->WatOrd[ixWat[0]]);}
    // if (pCurLs->WatOrd.n >= 2) {swap(pCurLs->WatOrd[1], pCurLs->WatOrd[ixWat[1]]);}
    if (pCurLs->WatchLosrAE(0) != 0) assert(pCurLs->WatchLosrAE(0) != pCurLs->WatchLosrAE(1));
}

GsT* AsgnT::CreateGsFromGb(GsBuilder& gb) {
    GsT* pCurLs = talloc(GsT, 1);
    pCurLs->init();

    int NumLits = gb.TrigHeap.h.n;
    garri TrigLits = garri(NumLits);
    gb.CopyTrigsTo(TrigLits);
    MinimizeConflict(gb, TrigLits);
    timsort(TrigLits);
    TrigLits.remove_adj_dups();
    int MaxLosrQb = 0;
    foreach (itLit, TrigLits) {
        MaxLosrQb = max(MaxLosrQb, (int)QbFromLit[*itLit]);
    }
    if (MaxLosrQb == 0) {
        MaxLosrQb = MAX_QB;
    }

    int NumImps = gb.ImpLits.size();
    pCurLs->ImpLits.alloc(NumImps);
    //{auto(itLit, gb.ImpLits.begin());
    //for (int i=0; i < NumImps; i++) {
    //    pCurLs->ImpLits[i] = *(itLit++);
    //}}
    {
    int i = 0;
    foreach (itLit, gb.ImpLits) {
        if (MaxLosrQb < QbFromLit[*itLit]) {continue;}
        assert(i < pCurLs->ImpLits.n);
        pCurLs->ImpLits[i++] = *itLit;
    }
    if (pCurLs->ImpLits.n != i) {
        NumReducs++;
    }
    pCurLs->ImpLits.n = i;
    }
    pCurLs->ReqLits = dup_as_ap(TrigLits);
    TrigLits.dealloc();


    pCurLs->FreeFmla = this->FreeFmla;
    //pCurLs->GateLit = TargGate ^ 1 ^ this->FreeFmla;
    //assert(HasLit(pCurLs->GateLit));
    assert(FreeFmla==true || FreeFmla==false);

    //assert(IsResCompat(pCurLs->TrigAE[FreeFmla], pCurLs->ImpLits));

    timsort(pCurLs->ReqLits);
    //pCurLs->ReqLits.sort(CmpLitByPoit);
    pCurLs->InitTrigLits(pCurLs->ReqLits);
    pCurLs->InitWatchedLitsPart1();

    if (NumLits == 0) {
        return pCurLs;
    }
    
    volnop();

    int BestImp = 0;
    int LastTrig = pCurLs->WatchLosrAE(0);
    int Penult  =  pCurLs->WatchLosrAE(1);
    if (!gb.ImpLits.empty()) {BestImp = *gb.ImpLits.begin();}
    foreach (itLit, gb.ImpLits) {
        if (QbFromLit[*itLit] > QbFromLit[LastTrig]) continue;
        if (LitToDL[*itLit] > LitToDL[Penult]) {
            if (LitToDL[*itLit] > LitToDL[BestImp]) {
                BestImp = *itLit;
            }
        }
    }
    pCurLs->WatchedRes = BestImp;
    if (pCurLs->ImpLits.n > 0) assert(pCurLs->WatchedRes != 0);
    if (GloSanChk >= 3) {
        int WinrLetr = GetWinner();
        foreach (it, pCurLs->ImpLits) {
            assert(QtypeFromLit(*it) == WinrLetr);
        }
    }

    if (GloSanChk >= 2) {
        vector<LitT>* pRawInts = pCurLs->AsLitVec();
        auto(itOld, DupLrnTest.find(*pRawInts));
        if (itOld != DupLrnTest.end()) {
            GsT* pOld = itOld->second;
            pOld->PrintIndex();
            printf("THIS SHOULDN'T BE HAPPENING\n");
            assert(0);
        }
        DupLrnTest[*pRawInts] = pCurLs;
        delete pRawInts;
    }
    pCurLs->IsLearned = true;
    pCurLs->LastUse = NumBigBt;
    pCurLs->bump_activity();
    /*{
    int* pCur1 = pCurLs->ReqLits.begin();  int* pEnd1 = pCurLs->ReqLits.begin();
    int* pCur2 = pCurLs->ImpLits.begin();  int* pEnd2 = pCurLs->ImpLits.begin();
    while (pCur1 != pEnd1 && pCur2 != pEnd2) {
        if      (*pCur1 < *pCur2) {++pCur1;}
        else if (*pCur2 < *pCur1) {++pCur2;}
        else {
            if (GloDbgLev >= 1950) {
                printf("NARF! %i\n", *pCur1);
            }
            break;
        }
    }
    }*/

    return pCurLs;
}




void AsgnT::AssignLrnId(GsT* pCurLs) {
    pCurLs->id = (pGloProb->clauses.n);
    pGloProb->clauses.append(pCurLs);
    if (pCurLs->ReqLits.n > 3) {
        TempLearntClauses.insert(pCurLs);
    }
}

void AsgnT::AddNewLitSet(GsT* pCurLs) {
    //int NumLits = pCurLs->ReqLits.n;
    //assert(NumLits > 0);
    if (pCurLs->ReqLits.n <= 1) {
        //stop();
    }
    pCurLs->RegisterLitsHave();
    int* NewWatchesB = pCurLs->WatchRawArr;
    for (int i=0; i < 2; i++) {
        if (NewWatchesB[i] == 0) continue;
        pCurLs->itWatchRawAE[i] = pGloAsgn->LitWatchInsert(pCurLs->WatchRawAE[i], pCurLs);
    }
    pCurLs->itWatchRes = pGloAsgn->LitWatchInsert(pCurLs->WatchedRes, pCurLs);

}

void AsgnT::PrintNewLitSet(GsT* pCurLs, const char* extra, int DeltaBT) {
    if (GloDbgLev >= 1900) {
        printf("%s Lrn GS #%i (L%i). ", extra, pCurLs->id, pCurLs->LrnType);
    }
    if (GloDbgLev >= 1800) {
        int winner = (pCurLs->FreeFmla==true ? 'E' : 'A');
        printf("(");
        int DeltaQB = QbFromLit[LastChoice] - QbFromLit[GetLastChlit()];
        printf("%c wins at QB %2i%c, ", winner, QbFromLit[LastChoice], tolower(QtypeFromLit(LastChoice)));
        if (DeltaBT) printf(UTF8_DELTA "DL %4i, " UTF8_DELTA "QB %2i, %3i resolvs, ", 
            DeltaBT, DeltaQB, NumResolves);
        //printf("%4i%c, ", LastChoice, tolower(QtypeFromLit(LastChoice)));
        if (!DeltaBT) printf("                  ");
        printf(")  ");
    }
    if (GloDbgLev >= 2000) {
        DumpLitsEnc(pCurLs->ReqLits, " , imps ");
        DumpLitsEnc(pCurLs->ImpLits, " ");
    }
    if (GloDbgLev >= 1950 && GloDbgLev < 2000) {
        printf("A ");
        DumpLitsEnc(pCurLs->ReqLits, ", E ");
    }
    if (1900 <= GloDbgLev && GloDbgLev < 1950) {
        if (pCurLs->ReqLits.n <= 1) {
            DumpLitsEnc(pCurLs->ReqLits, "");
            //stop();
        } else {
            int NumLits = pCurLs->ReqLits.n;
            printf("   ReqLits.n = %i ", NumLits);
        }
    }
    if (GloDbgLev >= 1800) {
        const char* sep;
        if (GloDbgLev >= 2000) {sep = "\n\n"; NeedNewl=false;}
        else if (GloDbgLev >= 1900) {sep = "\n"; NeedNewl=false;}
        else if (NumBigBt % 2 == 0) {sep = "\n"; NeedNewl=false;}
        else {sep = ""; NeedNewl = true;}
        printf("%s", sep);
    }

}



// uint AsgnT::FindBtLev(GsT* pLrn) {
//     /* Find the backtrack decision level. */
//     int winr = FreeFmla;
//     int losr = 1 - FreeFmla;
//     uint BtDL;
//     int MaxQb;
// 
//     int FinalLit = pLrn->WatchAE(losr, 0);
//     if (LitToDL[   pLrn->WatchAE(winr, 0)] > LitToDL[FinalLit]) {
//         FinalLit = pLrn->WatchAE(winr, 0);
//     }
//     if (IsGateLit(FinalLit) || QbFromLit[FinalLit] < QuantPfx.blks[0].ix) {
//         BtDL = LitToDL[FinalLit];
//         return BtDL;
//     }
// 
//     BtDL = LitToDL[pLrn->WatchAE(losr, 1)];
//     MaxQb = QbFromLit[pLrn->WatchAE(losr, 0)];
//     assert(IsInputLit(pLrn->WatchAE(losr, 0)));
//     foreach (it, pLrn->ImpLits) {
//         uint ImpDL = LitToDL[*it];
//         if (BtDL < ImpDL && QbFromLit[*it] < MaxQb) {
//             BtDL = ImpDL;
//         }
//     }
//     foreach (it, pLrn->TrigAE[winr]) {
//         uint ImpDL = LitToDL[*it];
//         if (BtDL < ImpDL && QbFromLit[*it] < MaxQb) {
//             BtDL = ImpDL;
//         }
//     }
//     return BtDL;
// }
    
GsT* pOldConflict = NULL;

/*****************************************************************************
* LearnAndBacktrack.    }}}{{{1
*****************************************************************************/
GsT* AsgnT::LearnAndBacktrack(int GotTwo) {
    int winr = FreeFmla;
    int LastChlit = GetLastChlit();
    int ElimQb = QbFromLit[LastChlit];
    pOldConflict = pConflictGs;
    bool DoCegar = (
        HasMoreVars() &&
        (AllowCegar >= 1) && LastChlit != 0 && 
        !pGloProb->NoMoreVars && 
        true);
    if (DoCegar) {
        if (winr == QtypeNumFromLit(LastChlit)) {
            DoCegar = true;
        } else {
            DoCegar = 
            (LitToDL[pConflictGs->WatchRawAE[0]] != LitToDL[pConflictGs->WatchRawAE[1]] &&
             QbFromLit[pConflictGs->WatchedRes] < QbFromLit[pConflictGs->WatchRawAE[0]]);
            if (DoCegar) {
                ElimQb = QbFromLit[pConflictGs->WatchedRes];
            }
        }
        if (ElimQb <= 1) {
            DoCegar = false;
        }
        //if (ElimQb != InnermostInputQb - 1) {DoCegar = false;}
        if (ElimQb < InnermostInputQb - 3) {   // TODO: Remove this!!!
            DoCegar = false;
        }
    }
    // if (DoCegar && winr != QtypeNumFromLit(LastChlit)) {
    //     ElimQb++;
    // }
    
    //CegarLits.clear();
    // {
    //     bool HasAnyWinr = false;
    //     if (winr == 0) {
    //         foreach (itLit, pGloProb->QuantBlk[2].lits) {
    //             if (!IsInputLit(*itLit)) continue;
    //             if (HasLitOrNeg(*itLit)) {
    //                 HasAnyWinr = true;
    //             }
    //         }
    //     }
    //     if ((HasAnyWinr && !DoCegar)) {
    //         foreach (itLit, pGloProb->QuantBlk[2].lits) {
    //             if (!IsInputLit(*itLit)) continue;
    //             if (HasLit(*itLit)) {
    //                 printf("LitToDL[%i] = %i\n", *itLit, LitToDL[*itLit]);
    //             }
    //         }
    //         stop();
    //     }
    // }
    static garri AsgnLits = garri(0,0);
    AsgnLits.n = 0;

    garr<GsT*> MoreLrn; MoreLrn.init(LRN_TYPE_4+1);
    for (int i=0; i < LRN_TYPE_4+1; i++) {MoreLrn[i] = NULL;}
    //GsT* pLrnGood = 
    pGloAsgn->MakeLearnedGs(MoreLrn);

    //if (pLrn == NULL) {assert(0);}
    /*
    uint BtDL = FindBtLev(pLrn);
    for (int i=LRN_TYPE_2; i < LRN_TYPE_4+1; i++) {
        GsT* pLrnX = MoreLrn[i];
        if (pLrnX != NULL) {
            uint BtLX = FindBtLev(MoreLrn[i]);
            if (BtLX < BtDL) {
                BtDL = BtLX;
                pLrn = MoreLrn[i];
            }
        }
    }
    int DeltaBT = GetCurDL() - BtDL;
    */
    for (int i=LRN_TYPE_2; i >= LRN_TYPE_1; i--) {
        GsT* pLrnX = MoreLrn[i];
        if (pLrnX) AssignLrnId(pLrnX);
    }
    if (GloDbgLev >= 2004) {
        DumpLits(0);
    }
    /*
    int FinalLit = pLrn->WatchAE(losr, 0);
    if (LitToDL[   pLrn->WatchAE(winr, 0)] > LitToDL[FinalLit]) {
        FinalLit = pLrn->WatchAE(winr, 0);
    }
    bool IsFinal = (QbFromDL(LitToDL[FinalLit]) < QuantPfx.blks[0].ix);
    if (IsGateLit(FinalLit)) assert(IsFinal);
    */



    GsT* pLrn = NULL;
    foreach (itGs, MoreLrn) {
        GsT* pLrnX = *itGs;
        if (pLrnX == NULL) continue;
        assert(pLrn == NULL);
        pLrn = pLrnX;
        AddNewLitSet(pLrnX);
        if (pLrnX->ReqLits.n == 0) {
            PrintNewLitSet(pLrnX, "$", 0);
            return pLrnX;
        }
    }



    //int InitTarg = TargGate;
    int PrevLit;
    uint InitDL = GetCurDL();

    int FirstOut = GetOutOrderChlit();
    if (FirstOut) {
        volnop();
    }
        

    if (winr == LastWinr) {
        WinStreak++;
    } else {
        WinStreak = 0;
    }
    LastWinr = winr;
    LastWinrChlits.n = 0;
    InitChoices.n = 0;

    unsigned int TargDL = LitToDL[pLrn->WatchRawAE[1]];
    if ((QbFromLit[pLrn->WatchedRes] < QbFromLit[pLrn->WatchRawAE[0]]) || NO_UNIV_REDUC) {
        TargDL = max(TargDL, LitToDL[pLrn->WatchedRes]);
    }


    /* Make AsgnLits */
    if (DoCegar) {
        for (int i=LRN_TYPE_2; i >= LRN_TYPE_1; i--) {
            GsT* pLrnX = MoreLrn[i];
            if (pLrnX == NULL) continue;
            foreach (itLit, pLrnX->ImpLits) {
                if (!((QtypeNumFromLit(*itLit)==winr) && IsInputLit(*itLit))) {
                    continue;
                }
                if (QbFromLit[*itLit] != ElimQb) continue;
                AsgnLits.append(*itLit);
            }
        }
        
        foreach (itChlit, chlits.flat) {
            garri& CurUndo = UndoListByDL[LitToDL[*itChlit]];
            foreach (itLit, CurUndo) {
                int lit = *itLit;
                if (lit == 0) continue;
                if (!IsInputLit(lit)) {continue;}
                //if (QtypeNumFromLit(lit) != winr) {continue;}
                //if (QbFromLit[lit] != ElimQb) continue;
                if (QbFromLit[lit] == ElimQb - 1) continue;
                //if (QbFromDL(TargDL) <= QbFromLit[lit] && QbFromLit[lit] < ElimQb) continue;
                if (ElimQb < QbFromLit[lit]) continue;
                if (lit > pGloProb->RawLastInpLit) continue;
                AsgnLits.append(lit);
            }
        }
    
        timsort(AsgnLits);
        AsgnLits.remove_adj_dups();
        if (ElimQb == 3) {
            volnop();
        }
        LitT prev = 0;
        foreach (itLit, AsgnLits) {
            if (*itLit == negv(prev)) {
                DoCegar = false;
                if (GloDbgLev >= 1100) {
                    printf("Cegar aborted due to contradictory winner lits (QB %i, contra lit %i)\n",
                        ElimQb, absv(prev));
                    break;
                }
            }
            prev = *itLit;
        }
    }



    if (1) {
        while (GetCurDL() != TargDL) {
            pGloAsgn->Backtrack();
        }
        int ret = ExecLitSet(pLrn, EM_SCHED); assert(ret > 0);
        pGloAsgn->Propagate();
    } else {
        while (true) {
            if (GetLastChlit() == ZERO_LIT) break;
            if (GloDbgLev >= 2009) printf("Popping %i (DL %i)\n", GetLastChlit(), GetCurDL());
            PrevLit = pGloAsgn->Backtrack();
            if (pConflictGs != NULL) {continue;}
            bool IsAsserting = 0;
            bool IsBlocked = 0;
            for (int i=LRN_TYPE_1; i <= LRN_TYPE_3; i++) {
                GsT* pLrnX = MoreLrn[i];
                if (pLrnX == NULL) continue;
                if (pLrnX->HasBlockedImps()) {
                    IsBlocked = 1;
                    // stop();
                    break;
                }
                IsAsserting |= (pLrnX->HasAllWatchedTrigs() || ExecLitSet(pLrnX, EM_DRYRUN) > 0);
                if (IsAsserting) continue;
            }
            if (IsBlocked) continue;
            if ((IsAsserting && GetCurDL() > 0)) {
                continue;
            } else {
                pGloProb->NumDoneClauses = pGloProb->clauses.n;
                int ret = ExecLitSet(pLrn, EM_SCHED); 
                if (ret > 0) {
                    pGloAsgn->Propagate();
                } else {
                    AppendChlit(PrevLit);
                    SchedForProp(PrevLit);
                    pGloAsgn->Propagate();
                }
                break;
            }
        }
    }
    assert(GetCurDL() == TargDL);
    int DeltaBT = InitDL - GetCurDL();
    assert(DeltaBT >= 0);
    int IsFinal = false; 
    bool PrintedOne = false;
    foreach (itGs, MoreLrn) {
        if (*itGs == NULL) continue;
        if (!PrintedOne) {
            PrintNewLitSet(*itGs, (IsFinal ? "$" : (GotTwo ? "*" : "+")), DeltaBT);
        } else {
            PrintNewLitSet(*itGs, "\\", 0);
        }
    }
    if (GloDbgLev >= 2051) {
        printf("CONFLICT!\n");
    }
    if (0) {
        foreach (itGs, MoreLrn) {
            if (*itGs == NULL) continue;
            GsT* pLrnX = *itGs;
            for (int ixCl = LastOrigClauseID; ixCl < pGloProb->clauses.n; ixCl++) {
                GsT* pTestCl = pGloProb->clauses[ixCl];
                if (pTestCl == NULL) continue;
                if (pTestCl == pLrnX) continue;
                if (pTestCl->is_subsumed_by(pLrnX)) {
                    if (GloDbgLev >= 100) {
                        printf(" %5i subsumed by %5i\n", pTestCl->id, pLrnX->id);
                    }
                    //pTestCl->dealloc();
                }
            }
        }
    }
    if (GloDbgLev >= 2004) {
        DumpLits(0);
    }
    NumBigBt++;
    
    if (DeltaBT == 0) {
        abort();
        //restart();
    }

    if (GloSanChk >= 3 && !HasConflict()) WatchSanityChk3();
    if (GloSanChk >= 4 && !HasConflict()) LatentExecsChk();
    if (GloDbgLev >= 2012) printf("Done Lrn #%i\n", pGloProb->clauses.n-1);

    MoreLrn.dealloc();


    if (DoCegar) {
        map<int,int> *hit = new map<int,int>();

        if (NumBigBt == 961) {
            volnop();
        }

        garri OldChlits = chlits.flat.dup();
        restart();   // TODO: Don't restart!

        pGloProb->NumDoneClauses = pGloProb->clauses.n;
        GsT* CegarGs = NULL;
        int OldNumVars = pGloProb->NextNewVar;
        if (AllowCegar >= 1) {
            try {
                CegarGs = Augment(pGloProb->OutGlit, AsgnLits, hit, ElimQb);
                //stop();
            } catch (char const* s) {
                static int AlreadyWarned = false;
                pGloProb->NoMoreVars = true;
                if (AlreadyWarned == false) {
                    AlreadyWarned = true;
                    fprintf(stderr, "\nRan out space for new variables!\n");
                }
            }
        }
        if (CegarGs) {
            // if (GloDbgLev >= 2000) {
            //     printf("AsgnLits: "); DumpLitsEnc(AsgnLits, "\n");
            //     foreach (itPair, *hit) {
            //         int OldGlit = itPair->first;
            //         int NewGlit = itPair->second;
            //         if (OldGlit == NewGlit) continue;
            //         if      (NewGlit == 0) {continue; printf("%4i => false\n", OldGlit);}
            //         else if (NewGlit == 1) {continue; printf("%4i => true\n", OldGlit);}
            //         else if (IsInputLit(NewGlit)) {
            //             continue; //printf("%4i => %5i   (input lit)\n", OldGlit, NewGlit);
            //         } else {
            //             printf("%4i => %5i: ", OldGlit, NewGlit);
            //             DumpLitsEnc(GlitToGs[NewGlit]->ReqLits, " from ");
            //             if (IsOrGate(OldGlit)) {printf(" -");}
            //             DumpLitsEnc(AbsGlitToGs(OldGlit)->ReqLits);
            //         }
            //     }
            // } 
            if (GloDbgLev >= 1100) {
                int NumNewVars = (pGloProb->NextNewVar - OldNumVars)/2;
                printf("CEGAR %i: QB=%i, NumNewVars=%i, NumAsgnVars=%i\n", 
                    NumCegLrn, ElimQb, NumNewVars, AsgnLits.n);
            }
        }
        delete hit;

        if (CegarGs) {
            NumCegLrn++;
            ExecLitSet(CegarGs, EM_SCHED);
        } else if (AllowCegar >= 1) {
            static int AlreadyWarned = false;
            if (AlreadyWarned == false) {
                AlreadyWarned = true;
                if (!pGloProb->NoMoreVars) {
                    printf("WARNING: CEGAR failed!\n");
                } else if (GloDbgLev >= 100) {
                    printf("(Out of variables, no more CEGAR.)\n");
                }
            }
        }
        Propagate();

        if (GloSanChk >= 4 && !pGloAsgn->HasConflict()) LatentExecsChk();
        foreach (itChlit, OldChlits) {
            LitT chlit = *itChlit;
            if (HasConflict()) {break;}
            if (HasLit(negv(chlit))) {break;}
            if (HasLit(chlit)) {continue;}
            AppendChlit(chlit);
            SchedForProp(chlit);
            Propagate();
        }
        OldChlits.dealloc();
        if (GloSanChk >= 3 && !pGloAsgn->HasConflict()) WatchSanityChk3();
        if (GloSanChk >= 4 && !pGloAsgn->HasConflict()) LatentExecsChk();
        
    }

    return NULL;
}


// int AsgnT::PickNewTarg() {
//     foreach (itBlk, QuantPfx.blks) {
//         foreach (itLit, (*itBlk).lits) {
//             assert(HasLit(*itLit) || HasLit(negv(*itLit)));
//         }
//     }
//     static garri choices;
//     assert(choices.n == 0);
//     int IsNeg = !IsAndGate(TargGate);
//     foreach (it, GlitToGs[TargGate ^ IsNeg]->TrigG) {
//         int CurLit = *it ^ IsNeg;
//         if (HasLit(CurLit) || HasLit(negv(CurLit))) continue;
//         choices.append(CurLit);
//     }
//     int ret = choices[rand() % choices.n];
//     choices.n = 0;
//     return ret;
// }

// bool AsgnT::OccursOnlyOnce(int lit) {
//     garri pars = pGloProb->ParentLit[lit];
//     if (pars.n == 0) return true;
//     int count = 0;
//     foreach (it, pars) {
//         int CurPar = *it;
//         if (LitToLiveParent[CurPar] != 0) {count++;}
//     }
//     if (count > 1) return false;
//     return OccursOnlyOnce(pars[0]);
// }

void AsgnT::PreSolve() {
    /*for (int i=2; i <= pGloProb->LastInpLit; i += 2) {
        FreeVars.insert(i);
    }*/
    TargGate = pGloProb->OutGlit;
    LastDefnGs = pGloProb->clauses.n - 1;
    // foreach (it, BattleVars) {
    //     pGloProb->AddTopGs(*it);
    //     pGloProb->AddTopGs(negv(*it));
    //     BattleVarList.append(*it);
    // }
    if (!NoUnivGhost) {
        NewLrnGs_Top(ghosted(pGloProb->OutGlit, 0), 1);
        if (HasConflict()) {return;}
        NewLrnGs_Top(ghosted(negv(pGloProb->OutGlit), 1), 0);
        if (HasConflict()) {return;}
    }
    LastInitGs = pGloProb->clauses.n - 1;

    // {
    // assert(pGloProb->QuantBlk.n >= 2);
    // garri SplitLits = garri();
    // foreach (itLit, pGloProb->QuantBlk[2].lits) {
    //     if (IsInputLit(*itLit)) {
    //         SplitLits.append(*itLit);
    //     }
    // }
    // foreach (itLit, SplitLits) {
    //     SplitOnVar(*itLit);
    // }
    // }

    pGloProb->NumDoneClauses = pGloProb->clauses.n;
    pGloAsgn->Propagate();
    foreach (it, pGloProb->clauses) {
        if (*it == NULL) continue;
        ExecLitSet(*it, EM_SCHED);
        pGloAsgn->Propagate();
        if (HasConflict()) {return;}
    }
    WatchSanityChk3();
    if (!pGloAsgn->HasConflict()) LatentExecsChk();

    // if (0) {
    //     for (int lit=0; lit <= pGloProb->LastInpLit; lit++) {
    //         //if (LitToLiveParent[lit] == 0) 
    //         if (IsAlive(lit) && !IsAlive(negv(lit))) {
    //             die_f("Error: Input formula has a monotone literal: %i.\n", lit);
    //             /*if (HasLit(lit) || HasLit(negv(lit))) continue;
    //             int qtype = QtypeFromLit(lit);
    //             int flit = (qtype=='A' ? lit : (qtype=='E' ? negv(lit) : die()));
    //             GsT* DummyLs = MakeDummyOneLitClause(flit);
    //             AppendForcedLit(flit, DummyLs);
    //             pGloAsgn->SchedForProp(flit);*/
    //         }
    //     }
    // }
    // if (0) {
    //     for (int lit=0; lit <= pGloProb->LastInpLit; lit++) {
    //         garri pars = pGloProb->ParentLit[lit];
    //         if (pars.n == 0) continue;
    //         if (!OccursOnlyOnce(lit)) continue;
    //         GsT* pCurLs = pGloProb->LitHave[lit][0];
    //         if (pCurLs->ReqLits.n != 2) continue;
    //         int owner = QtypeNumFromLit(lit);
    //         int OtherLit = pCurLs->ReqLits[0] + pCurLs->ReqLits[1] - lit;
    //         if (!IsInputLit(OtherLit)) continue;
    //         if (QbFromLit[OtherLit] <= QbFromLit[lit]) continue;
    //         GsT* pNewLs = pGloProb->AddTwoLitLrnGs(OtherLit, (lit ^ owner), owner);
    //         assert(pNewLs);
    //         stop();
    //     }
    // }
    pGloAsgn->Propagate();
    WatchSanityChk3();

    // if (!NoExpand) ConsiderExpansion(true);
    // DoExpansion();  //FirstTime
    // DoExpansion();
    // if (!NoExpand) ConsiderExpansion(true);
    // if (!pGloAsgn->HasConflict()) LatentExecsChk();
    //stop();


    //IsRestarting = 0;
    ixChoi = 0;
    NumSinceCut = 0;
    NumUntilRestart = 2;
    LastPrintTime = GetUserTimeMicro();
    NumRestarts = 0;

    if (0) {
        int ElimQb = 2;
        assert(pGloProb->QuantBlk.n > ElimQb);
        int n = 0;
        foreach (itLit, pGloProb->QuantBlk[ElimQb].lits) {
            if (IsInputLit(*itLit)) {n++;}
        }
        for (int i=0; i < (1 << n); i++) {
            garri AsgnLits = garri(n);
            for (int j=0; j < n; j++) {
                AsgnLits[j] = pGloProb->QuantBlk[ElimQb].lits[j] ^ ((i >> j) % 2);
            }
            DumpLitsEnc(AsgnLits);
            map<int,int> *hit = new map<int,int>();
            if (1) {
                Augment(pGloProb->OutGlit, AsgnLits, hit, ElimQb);
            }
            if (0) {
                CegarExtend(AsgnLits, hit, ElimQb);
            }
            delete hit;
        }
        Propagate();
    }

    clean_up_sequents();
    
    //TargStack.init();
}

// void AsgnT::DoExpansion() {
//     int start = 2;
//     while (1 && !NoExpand) {
//         //  printf("Exiting for profiling purposes!\n");
//         //  exit(0);
//         start = ConsiderExpansion(false, start);
//         if (start==0) break;
//         pGloAsgn->Propagate();
//         WatchSanityChk3();
//     }
// }

// void AsgnT::StableSortByQb(garri lits) {
//     static garri* LitsInQb = NULL;
//     if (LitsInQb == NULL) {
//         LitsInQb = talloc(garri, pGloProb->QuantBlk.n);
//         for (int qb=1; qb <= InnermostQB; qb++) {
//             LitsInQb[qb].init(0, pGloProb->QuantBlk[qb].lits.n);
//         }
//     }
//     for (int qb=1; qb <= InnermostQB; qb++) {
//         LitsInQb[qb].n = 0;
//     }
//     foreach (itLit, lits) {
//         int lit = *itLit;
//         LitsInQb[QbFromLit[lit]].append(lit);
//     }
//     lits.n = 0;
//     for (int qb=1; qb <= InnermostQB; qb++) {
//         lits.extend(LitsInQb[qb]);
//     }
// }

int AsgnT::choose_lit_and_prop(int choice) {
    int NewLit = pGloAsgn->ChooseRandLit(choice);
    if (GloDbgLev >= 2002) {
        printf("(pick %i) ", NewLit);
    }
    LastChoice = NewLit;
    AppendChlit(NewLit);
    pGloAsgn->SchedForProp(NewLit);
    pGloAsgn->Propagate();
    return NewLit;
}

template <typename T>
class CircBufr {
    public:
    int size;
    int cur_ix;
    T* buffer;

    CircBufr(int arg_size) {
        size = arg_size;
        buffer = talloc(T, size);
        cur_ix = 0;
    }
    ~CircBufr() {
        free(buffer);
        buffer = NULL;
        size = 0;
    }
    T& prev_val() {
        return buffer[(cur_ix - 1 + size) % size];
    }
    T& cur_val() {
        return buffer[cur_ix];
    }
    void advance() {
        cur_ix = (cur_ix + 1) % size;
    }

};


/*****************************************************************************
* Top-level solver function.    }}}{{{1
*****************************************************************************/
FmlaT AsgnT::solve() {
    // {
    //     garri AsgnLits = garri(1,1);
    //     AsgnLits[0] = 182;
    //     map<int,int> *hit = new map<int,int>();
    //     Augment(pGloProb->OutGlit, AsgnLits, hit, 0);
    //     Propagate();
    // }

    // if (0) {
    //     garri CegAsgn = garri(0);
    //     CegAsgn.append(14);
    //     CegarExtend(CegAsgn);
    //     CegAsgn.n = 0;
    //     CegAsgn.append(negv(14));
    //     CegarExtend(CegAsgn);
    // }

    long long LastPrintBt = 0;
    long long LastPrintProps = 0;
    auto(AgeCutoff, CircBufr<int>(4));
    while (1) {
        while (!pGloAsgn->HasConflict()) {
            int choice = 0;
            // if (UndoTarg[GetCurDL()+1] == 0) {
            //     UndoTarg[GetCurDL()+1] = TargGate;
            // }
            while (ixChoi < InitChoices.n) {
                choice = InitChoices[ixChoi];
                if (pGloAsgn->HasLit(choice)) {ixChoi++; continue;}
                if (pGloAsgn->HasLit(negv(choice))) {
                    InitChoices.n = 0;
                }
                break;
            }
            if (GloSanChk >= 5 && !pGloAsgn->HasConflict()) LatentExecsChk();
            choose_lit_and_prop(choice) || die_f("\nERROR: NO CHOICES!  (NumBt = %i)\n", NumBigBt);
            GloStep++;
        }
        int GotOne = false;
        int GotTwo = false;
        GsT* pFinal;
        //if (GloSanChk >= 3) WatchSanityChk3();
        while (pGloAsgn->HasConflict()) {
            if (GotOne) GotTwo = true;
            GotOne = true;

            pFinal = LearnAndBacktrack(GotTwo);
            cla_decay_activity();
            var_decay_activity();
            LitHitInc *= 1.01;


            if (--learntsize_adjust_cnt == 0){
                learntsize_adjust_confl *= learntsize_adjust_inc;
                learntsize_adjust_cnt    = (int)learntsize_adjust_confl;
                max_learnts             *= learntsize_inc;
            }


            if (pFinal != NULL) {
                if (pFinal->ReqLits.n == 0) {
                    return pFinal->FreeFmla;
                }
            }

            delete_excess_learned_clauses();

            NumUntilRestart--;
            NumSinceCut++;
            if (GloTimeOut != 0  &&  (GetUserTimeMicro() - StartTime) > GloTimeOut) {
                PrintStats();
                fprintf(stderr, "Timed out!  (%d backtracks)\n", NumBigBt);
                exit(0);
            }
        }

        if (GotOne) {
            /* Random restarts. */
            //GsT* pLrn = pGloProb->clauses.last();
            if (NumUntilRestart <= 0 || !NoRestart) {
                if (GloDbgLev >= 100) {
                    long long CurTime = GetUserTimeMicro();
                    int DeltaTime = (CurTime - LastPrintTime);
                    if (GloDbgLev >= 2000 || DeltaTime >= 500*1000) {
                        LastPrintTime = CurTime;
                        if (NeedNewl) {printf("\n"); NeedNewl=false;}
                        printf("### RESTART %3i ###  ", NumRestarts);
                        printf("(Bt %5i, C %2i, speed:%6.0f conflicts/sec, %7.0f props/sec)\n",
                            NumBigBt, NumCegLrn,
                            ((double)(NumBigBt - LastPrintBt) / (double)DeltaTime * 1000000),
                            ((double)(NumProps - LastPrintProps) / (double)DeltaTime * 1000000)
                            );
                        LastPrintBt = NumBigBt;
                        LastPrintProps = NumProps;
                    }
                }
            }
            if (NumUntilRestart <= 0 && !NoRestart) {
                NumUntilRestart = RestartCycle;
                NumRestarts++;
                if (NumRestarts <= 4) {
                    NumUntilRestart = 8;
                } else {
                    RestartCycle += (int) sqrt(RestartCycle / 16);
                }
                if (GloDbgLev >= 2000) printf("\n");
                if (1 || NumSinceCut >= 8) {
                    NumSinceCut = 0;
                }
                AgeCutoff.cur_val() = NumBigBt;
                AgeCutoff.advance();
                pGloAsgn->restart();
                CurPathNum = (CurPathNum + 1) % (MaxPathNum + 1);
                CurPathMask = PathMaskFromNum(CurPathNum);
                if (CurPathNum == PathNum1) {
                    RandomizePolPref();
                }
                InitChoices.n = 0;
                LastWinrChlits.n = 0;
                // int cutoff = max(AgeCutoff.cur_val(), 200);
                // int NumDel = 0;
                // foreach (itCl, pGloProb->clauses) {
                //     GsT* pCurLs = *itCl;
                //     if (!pCurLs) continue;
                //     if (!pCurLs->IsLearned) continue;
                //     if (pCurLs->ReqLits.n <= 6) continue;
                //     int time_unused = NumBigBt - pCurLs->LastUse;
                //     if (time_unused > cutoff) {
                //         if (pCurLs->InUse) {
                //             assert(LitToDL[(negv(pCurLs->WatchRawAE[0]))] == 0);
                //             continue;
                //         }
                //         if (GloDbgLev >= 100) {
                //             printf("Deleting sequent %i (size %i, last use %i ago).\n", 
                //                 pCurLs->id, pCurLs->ReqLits.n, time_unused);
                //         }
                //         pCurLs->dealloc();
                //         NumDel++;
                //     }
                // }
                // if (GloDbgLev >= 100) {
                //     printf("Deleted %i sequents.\n", NumDel);
                // }
                // // if (NumRestarts % 5 == 0) {
                // //     if (!NoExpand) ConsiderExpansion(true);
                // //     DoExpansion();
                // // }
                // /*while (TargStack.n > 0) {
                //     TargGate = TargStack.PopLast();
                // }*/
                // //runs.clear();
            } else {
                expu = AllowExpu;
            }

            /*const uint RunSize = 3;
            if (runs.size() < RunSize && pGloAsgn->chlits.flat.n > 2) {
                pair<int, vector<int> > tmp;
                tmp.first = pLrn->ReqLits.n;
                for (int i=0; i < pGloAsgn->chlits.flat.n; i++) {
                    tmp.second.push_back(pGloAsgn->chlits.flat[i]);
                }
                runs.insert(tmp);
                pGloAsgn->restart();
                if (runs.size() == RunSize) {
                    InitChoices.n = 0;
                    foreach (iter, runs.begin()->second) {
                        InitChoices.append(*iter);
                    }
                    ixChoi = 1;
                    pGloAsgn->restart();
                    if (GloDbgLev >= 2000) printf("Selected choosen run.\n");
                }
            }*/

            /*InitChoices = pGloAsgn->chlits.flat.dup();
            ixChoi = 1;
            pGloAsgn->restart();*/
        }
    }
    assert(false);
}


/*****************************************************************************
* Helper funcs for parsing the input file.    }}}{{{1
*****************************************************************************/


/*void AddWatchOrig(GsT *pLitSet, int k, int lit) {
    assert(k==0 || k==1);
    assert(pLitSet->ReqLits.indexof(lit) != -1);
    pLitSet->WatchedLit[k] = lit;
    pGloAsgn->LitWatch[lit].insert(pLitSet);
}*/





// void CalcLitToBat(int glit, int bat) {
//     if (IsInputLit(glit)) {
//         LitToBat[glit]       = pGloProb->QuantBlk[QbFromLit[glit]].bat;
//         LitToBat[negv(glit)] = pGloProb->QuantBlk[QbFromLit[glit]].bat;
//         return;
//     }
//     if (QuantPfx[glit].blks.n > 0) {
//         bat = glit;
//     }
//     if (LitToBat[glit] != 0) {
//         if (LitToBat[glit] != bat) {
//             die_f("Error: Gate %i appears under both %i and %i.\n", glit, LitToBat[glit], bat);
//         }
//         return;
//     }
//     LitToBat[glit] = bat;
//     LitToBat[negv(glit)] = bat;
//     foreach (it, AbsGlitToGs(glit)->ReqLits) {
//         CalcLitToBat(*it, bat);
//     }
// }

/*void ProbInst::CalcGlitToBat(int CurLit, int CurBat) {
    if (IsInputLit(CurLit)) {
        CurBat = pGloProb->QuantBlk[MinQbFromGlit[CurLit]].bat;
        GlitToBat[CurLit] = CurBat;
        GlitToBat[negv(CurLit)] = CurBat;
        return;
    }
    if (pGloAsgn->IsBattleLit(CurLit)) {CurBat = CurLit;}
    if (GlitToBat[CurLit] != 0 && GlitToBat[CurLit] != CurBat) {
        die_f("Error in input file: Gate %i occurs under both %i and %i.\n",
            GlitToBat[CurLit], CurBat);
    }
    GlitToBat[CurLit] = CurBat;
    GlitToBat[negv(CurLit)] = CurBat;
    foreach (it, AbsGlitToGs(CurLit)->TrigG) {
        CalcGlitToBat(*it, absv(CurBat));
    }
}*/

void DieParse(const char* pErrMsg, Lexer* lex, ...) {
    fprintf(stderr, "Error on line %i: ", lex->LineNum);
    va_list ap;
    va_start(ap, lex);
    vfprintf(stderr, pErrMsg, ap);
    va_end(ap);
    printf("\n");
    abort();
}


void AsgnT::NewDefnGsAlt(LitT ixExpr, garri args) {
    static bool WarnDuplicate = true;
    args = args.dup();
    timsort(args);
    //args.sort(CmpLitByNarf);
    args.remove_adj_dups();
    auto(entry, HashedApInt(args));
    if (ArgsToGlit.count(entry)) {
        int old = ArgsToGlit[entry];
        //args = garri(1);
        if (WarnDuplicate && GloDbgLev >= 20) {
            WarnDuplicate = false;
            fprintf(stderr, "Redundant gate: %i has same inputs as %i.\n", ixExpr, old);
        }
        //args[0] = old;
    } else {
        ArgsToGlit[entry] = ixExpr;
    }

    garri lits = garri(0, args.n + 1);
    for (int ply=0; ply <= 1; ply++) {
        if (ply == 0 and NoUnivGhost) continue;
        lits.n = 0;
        int GhostLit = ghosted(negv(ixExpr), ply);
        lits.append(GhostLit);
        for (int i=0; i < args.n; i++) {
            lits.append(ghosted_or_pass(args[i], ply));
        }
        GsT* pCurLs = NewLrnGs_Mixed(lits, 1 - ply);
        pCurLs->GlitDefd = GhostLit;
        GhostToGs[ghosted(ixExpr, ply)] = pCurLs;
        GhostToGs[negv(ghosted(ixExpr, ply))] = NON_INPUT;
        if (IsInCegar) TempNewClauses.append(pCurLs);
    }
    for (int i=0; i < args.n; i++) {
        for (int ply=0; ply <= 1; ply++) {
            if (ply == 0 and NoUnivGhost) continue;
            lits.n = 0;
            int GhostGate = ghosted(ixExpr, ply);
            lits.append(ghosted_or_pass(negv(args[i]), ply));
            lits.append(GhostGate);
            GsT* pCurLs = NewLrnGs_Mixed(lits, 1 - ply);
            pCurLs->GlitDefd = GhostGate;
            if (IsInCegar) TempNewClauses.append(pCurLs);
        }
    }
    lits.dealloc();
    args.dealloc();
    return;
}


GsT* AsgnT::NewLrnGs_Top(LitT top, int winr) {
    garri args = garri(1, 1);
    args[0] = top;
    GsT* ret = NewLrnGs_Mixed(args, winr);
    args.dealloc();
    return ret;
}

GsT* AsgnT::NewLrnGs_Mixed(garri AsgnLits, int winr) {
    static garri args;
    static garri WinrLits;
    args.n = 0;
    WinrLits.n = 0;
    foreach (itLit, AsgnLits) {
        if (QtypeNumFromLit(*itLit) == winr) {
            WinrLits.append(*itLit);
        } else {
            args.append(*itLit);
        }
    }
    return NewLrnGs(args, WinrLits, winr);
}

GsT* AsgnT::NewLrnGs(garri args, garri WinrLits, int winr) {
    assert(winr==1 || winr==0);
    GsT* pCurLs = talloc(GsT, 1);
    pCurLs->id = pGloProb->clauses.n;
    pGloProb->clauses.append(pCurLs);
    pCurLs->init();
    pCurLs->FreeFmla = winr;
    {
        garri tmp = args.dup();
        timsort(tmp);
        //tmp.sort(CmpLitByPoit);
        tmp.remove_adj_dups();
        pCurLs->ReqLits = dup_as_ap(tmp);
        tmp.dealloc();
        pCurLs->InitTrigLits(pCurLs->ReqLits);
        pCurLs->RegisterLitsHave();
    }
    {
        garri tmp = WinrLits.dup();
        timsort(tmp);
        //tmp.sort(CmpLitByPoit);
        tmp.remove_adj_dups();
        pCurLs->ImpLits = dup_as_ap(tmp);
        tmp.dealloc();
    }
    pCurLs->InitWatchedLitsPart1();
    AddNewLitSet(pCurLs);
    pGloProb->NumDoneClauses = pGloProb->clauses.n;
    ExecLitSet(pCurLs, EM_SCHED);
    Propagate();
    return pCurLs;
}


/****************************************************************************/
void add_var_to_qblock(int qvar, int ixQb) {
    pGloProb->QuantBlk[ixQb].lits.append(qvar);
    assert(qvar == absv(qvar));
    assert(QbFromLit[qvar] == 0);
    QbFromLit[qvar] = ixQb;
    QbFromLit[negv(qvar)] = ixQb;
    pGloAsgn->InsertVarActOrder(qvar);
}

void AddQuantBlock(int qtype, garri qvars) {
    assert(qtype=='A' || qtype=='E');
    int QuantGate = pGloProb->OutGlit;
    int ixQb = pGloProb->QuantBlk.n;
    assert(qtype != pGloProb->QuantBlk.last().qtype);
    pGloProb->QuantBlk.AppendBlank();
    pGloProb->QuantBlk[ixQb].qtype = qtype;
    pGloProb->QuantBlk[ixQb].ix = ixQb;
    pGloProb->QuantBlk[ixQb].bat = QuantGate;
    pGloProb->LastQtypeBlock[QtypeNumFromLetr(qtype)] = ixQb;
    VarActOrder.add_qb(ixQb);
    foreach (it, qvars) {
        int qvar = *it;
        add_var_to_qblock(qvar, ixQb);
    }
    QuantPfx.blks.append(pGloProb->QuantBlk[ixQb]);
    if (ixQb > MAX_QB) {
        die("Too many quantification blocks.");
    }
}

void add_gate_to_qb(int ixExpr, garri args) {
    for (int qtype=0; qtype <= 1; qtype++) {
        int MaxQb = 0;
        foreach (itArg, args) {
            int CurQb = QbFromLit[ghosted_or_pass(*itArg, qtype)];
            assert(CurQb != 0);
            if (MaxQb < CurQb) {
                MaxQb = CurQb;
            }
        }
        if (MaxQb > InnermostInputQb) {
            MaxQb = InnermostInputQb;
        }
        //MaxQb = InnermostInputQb + 1;
        int CurQb = MaxQb + ((pGloProb->QuantBlk[MaxQb].qtype == QtypeLetrFromNum(qtype)) ? 0 : 1);
        assert(pGloProb->QuantBlk[CurQb].qtype == QtypeLetrFromNum(qtype));
        add_var_to_qblock(ghosted(absv(ixExpr), qtype), CurQb);
    }
}

double CalcGateExpFactor() {
    double THRES1 = 10000;
    double THRES2 = 500000;
    double pct = 0;
    if (false) {
    } else if (pGloProb->LastOrigGlit < THRES1) {
        pct = 1;
    } else if (pGloProb->LastOrigGlit < THRES2) {
        pct = ((THRES2 - pGloProb->LastOrigGlit) / THRES2);
        pct = pct * sqrt(pct);
    }
    double factor = 1 + pct * 24;
    return factor;
}

/*****************************************************************************
* ParseFile.    }}}{{{1
*****************************************************************************/
ProbInst ParseFile(char* filename) {
    //pGloProb = talloc(ProbInst,1);
    pNumClauses = &pGloProb->clauses.n;
    ProbInst& prob = *pGloProb;
    Lexer lex(filename);

    if (StrEqc(lex.CurTok.p, "CktQBF")) {lex.AdvTok(); lex.ReadLineSep();}
    prob.LastInpLit = lex.ReadIntParamEnc("LastInputVar") | 1;
    if (StrEqc(lex.CurTok.p, "FirstGateVar")) {lex.ReadIntParamEnc("FirstGateVar");}
    prob.LastOrigGlit = (lex.ReadIntParamEnc("LastGateVar")) | 1;
    prob.OutGlit = lex.ReadIntParamEnc("OutputGateLit");
    if (StrEqc(lex.CurTok.p, "PreprocTimeMilli")) {
        lex.AdvTok();
        prob.PreprocTimeMilli = lex.ReadIntRaw();
        lex.ReadLineSep();
    }
    int RawLastInpLit = prob.LastInpLit;
    prob.RawLastInpLit = RawLastInpLit;
    prob.OutGlit = frob_raw_gate(prob.OutGlit);
    prob.LastOrigGlit = frob_raw_gate(absv(prob.LastOrigGlit)) + 3;
    prob.LastGlit = (int)(prob.LastOrigGlit * CalcGateExpFactor() + 10000) | 1;
    prob.LastInpLit = prob.LastGlit;
    prob.FirstGateVar = prob.LastGlit;
    prob.NextNewVar = prob.LastOrigGlit + 1;
    prob.NoMoreVars = false;



    prob.LastPlit = prob.LastGlit;

    prob.clauses.init(0, prob.LastGlit*2 + 4000);
    prob.clauses.AppendBlank();
    prob.LitHave = talloc(garr<GsT*>, prob.LastPlit+1);
    //OrigGateFromDeriv = talloc(int, prob.LastPlit+1);
    ReEvalArr = talloc(short, prob.LastPlit+1);
    //TimesInConflict = talloc(int, prob.LastGlit+1);
    //vsids = talloc(VsidsT, prob.LastPlit+1);
    VarActivity = garr<double>(prob.LastPlit+1);
    GlitToGs = talloc(GsT*, prob.LastPlit+1);
    GhostToGs = garr<GsT*>(prob.LastPlit+1);
    //prob.ParentLit = talloc(garri, prob.LastPlit+1);
    QbFromLit = talloc(short, prob.LastPlit+1);
    //for (int i=prob.NextNewVar; i <= prob.LastGlit; i++) {QbFromLit[i] = MAX_QB;}
    //GlitToBat = talloc(int, prob.LastPlit+1);
    prob.QuantBlk.init(0);
    prob.QuantBlk.AppendBlank();
    assert(prob.QuantBlk[0].ix == 0);

    for (int lit = 0; lit <= prob.LastOrigGlit; lit++) {
        GhostToGs[lit] = (lit <= prob.RawLastInpLit ? INPUT_LIT : NON_INPUT);
    }

    
    //pGloAsgn = new AsgnT;
    pGloAsgn->init();
    pGloAsgn->LitWatch = new GwSetT[prob.LastPlit+1];
    #if USE_WATCH_ARR
    for (int ix = 0; ix <= prob.LastPlit; ix++) {
        pGloAsgn->LitWatch[ix].wlit = ix;
    }
    #endif

    /* Read quantifier blocks */
    int PrevMaxLit = 0;
    int CurMaxLit = 0;
    int PrevQvar = 0;
    while (true) {
        if (lex.CurTok.p == NULL) {DieParse("Unexpected end-of-file.", &lex, 0);}
        if (lex.CurTokType == INT_TOK_TYPE) {break;}
        int QuantGate=0;
        sscanf(lex.CurLine.p, "<q gate=%u", &QuantGate);
        if (QuantGate == 0) DieParse("Expected a line beginning with \"<q gate=\" followed by a positive integer.", &lex, 0);
        QuantGate = frob_raw_gate(QuantGate);
        if (QuantGate != prob.OutGlit) DieParse("Non-prenex forms are no longer supported.", &lex, 0);
        //pGloAsgn->BattleVars.insert(QuantGate);
        lex.SkipLine();
        while (true) {
            if (strncmp(lex.CurLine.p, "</q>", 4) == 0) {
                lex.SkipLine();
                break;
            }
            char qtype=0;
            if      (StrEqc(lex.CurTok.p, "A")) {qtype = 'A';}
            else if (StrEqc(lex.CurTok.p, "E")) {qtype = 'E';}
            else {DieParse("Expected 'a' or 'e', but found '%.1s'.", &lex, lex.CurTok.p);}
            lex.AdvTok();

            garri qvars;
            qvars.init();
            while (lex.CurTokType == INT_TOK_TYPE) {
                int qvar = lex.ReadIntEnc();
                if (qvar != absv(qvar)) {DieParse("Variable number is not even.", &lex, 0);}
                if (qvar != PrevQvar+2) {DieParse("CurQvar = %i != %i = PrevQvar+2\n", &lex, qvar, PrevQvar+2);}
                PrevQvar += 2;
                if (!prob.IsValidInpVar(qvar)) {DieParse("Variable number is out of range.", &lex, 0);}
                if (qvar < PrevMaxLit) {
                    DieParse("Vars quantified at an inner QB must be numbered higher than those quantified at an outer QB.", &lex, 0);
                }
                qvars.append(qvar);
                CurMaxLit = max(CurMaxLit, qvar);
            }
            PrevMaxLit = CurMaxLit;
            lex.ReadLineSep();

            AddQuantBlock(qtype, qvars);
            // foreach (itLit, qvars) {
            //     OrigGateFromDeriv[*itLit] = *itLit;
            //     OrigGateFromDeriv[negv(*itLit)] = negv(*itLit);
            // }
            qvars.dealloc();
        }
    }




    /* Read propositional formula */
    while (lex.CurTok.p != NULL) {
        int ixExpr = lex.ReadIntEnc();   
        if (ixExpr != absv(ixExpr)) {DieParse("Expr number is not even.", &lex, 0);}
        ixExpr = frob_raw_gate(ixExpr);
        assert(prob.IsValidExprIx(ixExpr));
        lex.ChkTok("=");
        const char* sOp = lex.ReadOpIdfr();
        garri args = lex.ReadArgs();
        if (StrEqc(sOp, "AND")) {
        } else if (StrEqc(sOp, "OR")) {
            // [x = a || b || c] is equivalent to [~x = ~a && ~b && ~c].
            ixExpr = negv(ixExpr);
            sOp = "AND";
            foreach (it, args) {*it = negv(*it);}
        } else {
            die_f("Error on line %i: Unexpected '%s'.\n", lex.LineNum, sOp);
        }
        {
            set<int> TmpSet(args.begin(), args.end());
            for (int i=0; i < args.n; i++) {
                if (TmpSet.count(negv(args[i])) != 0) {
                    DieParse("Defn contains contradictory literals: %i and its negation.", &lex, args[i]);
                }
            }
            if (TmpSet.size() < (uint) args.n) {
                DieParse("Duplicated argument.", &lex);
            }
        }
        if (args.n == 1) {
            DieParse("Operator needs at least two distinct arguments.", &lex, 0);
        }
        foreach (it, args) {
            if (*it > RawLastInpLit) {
                *it = frob_raw_gate(*it);
            }
        }
        prob.FirstGateVar = min(prob.FirstGateVar, absv(ixExpr));
        //printf("%i\n", ixExpr);
        assert(StrEqc(sOp, "AND"));
        for (int i=0; i < args.n; i++) {
            if (args[i] >= ixExpr) {DieParse("Argument must be less than parent.", &lex, 0);}
        }

        //pGloAsgn->NewDefnGs(ixExpr, args, false);
        GateDefn.setval(ixExpr, args);
        lex.ReadLineSep();
        //args.dealloc();
    }

    InnermostInputQb = prob.QuantBlk.n - 1;
    
    for (int q=0; q <=1; q++) {
        garri qvars = garri();
        AddQuantBlock('A'+'E' - prob.QuantBlk.last().qtype, qvars);
        AddQuantBlock('A'+'E' - prob.QuantBlk.last().qtype, qvars);
        int CurQtype = QtypeNumFromLetr(prob.QuantBlk.last().qtype);
        ExtraQb[CurQtype] = prob.QuantBlk.last().ix;
    }
    

    for (int ixExpr = 0; ixExpr <  GateDefn.n; ixExpr++) {
        garri args = GateDefn[ixExpr];
        if (args.arr == NULL) continue;
        for (int qtype=0; qtype <= 1; qtype++) {
            int qvar = ghosted(absv(ixExpr), qtype);
            assert(QbFromLit[qvar] == 0);
        }
    }
    for (int ixExpr = 0; ixExpr <  GateDefn.n; ixExpr++) {
        if (NoUnivGhost) break;
        garri args = GateDefn[ixExpr];
        if (args.arr == NULL) continue;
        add_gate_to_qb(ixExpr, args);
    }

    // pGloAsgn->DumpQuantPfx();

    for (int ixExpr = GateDefn.n - 1; ixExpr >= 0; ixExpr -= 1) {
        garri args = GateDefn[ixExpr];
        if (args.arr == NULL) continue;
        if (!NoUnivGhost) {
            pGloAsgn->NewDefnGsAlt(ixExpr, args);
        } else {
            if (ixExpr == prob.OutGlit) continue;
            assert(GateDefn[prob.OutGlit].bin_search(negv(ixExpr)));
            garri NegArgs = garri(args.n);
            for (int i=0; i < args.n; i++) {
                NegArgs[i] = negv(args[i]);
            }
            GsT* pCurLs = pGloAsgn->NewLrnGs_Mixed(args, 0);
            pCurLs->IsLearned = false;
            NegArgs.dealloc();
        }
        // OrigGateFromDeriv[ixExpr] = ixExpr;
        // OrigGateFromDeriv[negv(ixExpr)] = negv(ixExpr);
    }
     
    int NumDefn = GateDefn.n;
    for (int ixExpr = 0; ixExpr < NumDefn; ixExpr++) {
        if ((ixExpr & 2) != 0) continue;
        garri args = GateDefn[ixExpr];
        if (args.arr == NULL) continue;
        assert(ghosted(ixExpr, 0) == ixExpr);
        garri new_args(args.n);
        for (int i=0; i < args.n; i++) {
            new_args[i] = ghosted_or_pass(args[i], 1);
        }
        GateDefn.setval(ghosted(ixExpr, 1), new_args);
    }

    pGloAsgn->init2();
    max_learnts = ((double) pGloProb->clauses.n) * learntsize_factor;

    LastOrigClauseID = pGloProb->clauses.n;

    InnermostQB = prob.QuantBlk.n - 1;
    
    //pGloAsgn->InitPart2();

    return prob;
}


void AsgnT::PrintStats() {
    double x;
    double CpuTime = (GetUserTimeMicro() - StartTime) / 1000000.0;
    x=NumBigBt;          printf("Conflicts:    %10.0f  (%9.0f / sec)\n", x, x / CpuTime);
    x=NumDecs;           printf("Decisions:    %10.0f  (%9.0f / sec)\n", x, x / CpuTime);
    x=NumProps;          printf("Propagations: %10.0f  (%9.0f / sec)\n", x, x / CpuTime);
    x=NumWatchFixes;     printf("Watch fixes:  %10.0f  (%9.0f / sec)\n", x, x / CpuTime);
    x=NumWatchCleanups;  printf("Watch cleans: %10.0f  (%9.0f / sec)\n", x, x / CpuTime);
    x=max_learnts;       printf("max_learnts:  %10.0f  \n", x);
    x=NumReducs;         printf("NumReducs:    %10.0f  \n", x);
}


/*****************************************************************************
* main.    }}}{{{1
*****************************************************************************/

int main(int argc, char** argv) {
    
    GloRandSeed = 1;
    int ShowTime = 1;

    /*set<int> a; a.insert(7); a.insert(5);
    set<int> b; b.insert(5); b.insert(7);
    set<set<int> > foo;
    foo.insert(a);
    printf("ASDF %i\n", foo.size());*/

    if (argc < 2 || (StrEqc(argv[1], "-h") || StrEqc(argv[1], "--help"))) {
        fprintf(stderr, "Usage: %s infile [OPTIONS]\n", argv[0]);
        fprintf(stderr, 
            "Options:\n"
            " '-s N'        Use N as the seed for the random number generator.\n"
            " '-stime'      Seed the random number generator using the current time.\n"
            " '-d N'        Print debugging info; higher N for more info.\n"
            " '-s-cnf'      Print 's cnf 0' if instance is false, else print 's cnf 1'.\n"
            " '-cegar 0'    Don't use CEGAR learning.\n"
            " '-cegar 1'    Use CEGAR learning.\n"
            " '-q1', '-q2'  Quiet modes.\n"
            //" '-NoExpand'   Don't expand; only use DPLL.\n"
            " '-plog FILE'  Write a proof log to FILE\n"
            " '-SanChk N'   Higher N means more sanity checks.  Default is __.\n"
            " '-TimeOut N'  Abort after N seconds.\n"
            " '-pt'         Include preprocessing time in time limit.\n"
            " '-NoTime'     Don't show elapsed time. (For regression testing.)\n"
            " '-Restart N'  Restart every N learned game states.\n"
            //" '-Repeat N'   Repeat N times (for profiling purposes).\n"
            " See source code for additional options.\n"
        );
        exit(1);
    }


    char* in_filename = argv[1];
    if (!StrEqc(in_filename, "-") && access(in_filename, F_OK) == -1) {
        fprintf(stderr, "File '%s' does not exist.\n", in_filename);
        exit(1);
    }

    /* Parse command-line options. */
    int NumRepeat = 1;
    bool AddPreprocTime = false;
    bool CondensedOut = false;
    bool QuietMode = false;
    bool s_cnf = false;
    char* PlogName = NULL;
    int ixArg = 2;
    for (; ixArg < argc; ixArg++) {
        char* sArg = argv[ixArg];
        if (sArg[0] == '-' && sArg[1] == '-') {
            sArg += 1;
        }
        if (false) {}
        else if (StrEqc(sArg, "-pt")) {AddPreprocTime = true;}
        else if (StrEqc(sArg, "-s-cnf")) {s_cnf = true;}
        else if (StrEqc(sArg, "-q2")) {QuietMode = true;}
        else if (StrEqc(sArg, "-q1")) {CondensedOut = true;}
        else if (StrEqc(sArg, "-OldCube")) {OldCubeWay = true;}
        else if (StrEqc(sArg, "-NoRestart")) {NoRestart = true;}
        else if (StrEqc(sArg, "-NoTime")) { ShowTime = 0; }
        else if (StrEqc(sArg, "-NoUnivGhost")) { NoUnivGhost = 1; }
        else if (StrEqc(sArg, "-expu")) { AllowExpu = 1; abort();}
        else if (StrEqc(sArg, "-stime")) { GloRandSeed = RandSeedFromTime(); }
        else if (ixArg+1 >= argc) {
            die_f("The option '%s' doesn't exist, or it requires a parameter.\n", sArg);
        }
        else if (StrEqc(sArg, "-Cegar")) { AllowCegar = atoi(argv[++ixArg]); }
        else if (StrEqc(sArg, "-Expand")) { NoExpand = !atoi(argv[++ixArg]); }
        else if (StrEqc(sArg, "-monotone")) { UseMonotone = atoi(argv[++ixArg]); }
        else if (StrEqc(sArg, "-LrnType4")) { UseLrnType4 = atoi(argv[++ixArg]); }
        else if (StrEqc(sArg, "-plog")) { PlogName = argv[++ixArg]; }
        else if (StrEqc(sArg, "-MaxLearnFactor")) { learntsize_factor = atof(argv[++ixArg]); }
        else if (StrEqc(sArg, "-s") || StrEqc(sArg, "-seed")) {
            assert(++ixArg < argc);
            GloRandSeed = atoi(argv[ixArg]);
        }
        else if (StrEqc(sArg, "-d") || StrEqc(sArg, "-DbgLev")) {
            assert(++ixArg < argc);
            #ifndef GloDbgLev
            GloDbgLev = atoi(argv[ixArg]);
            #else
            die("Must recompile with GloDbgLev not #define'd to use this option.\n");
            #endif
        }
        else if (StrEqc(sArg, "-SanChk")) {
            assert(++ixArg < argc);
            #ifndef NO_SAN_CHK
            GloSanChk = atoi(argv[ixArg]);
            #else
            die("Must recompile without NO_SAN_CHK #define'd to use this option.\n");
            #endif
        }
        else if (StrEqc(sArg, "-RestartCycle")) {
            assert(++ixArg < argc);
            RestartCycle = atoi(argv[ixArg]);
        }
        else if (StrEqc(sArg, "-TimeOut")) {
            assert(++ixArg < argc);
            GloTimeOut = (long long)(atof(argv[ixArg])*1000000);
        }
        else if (StrEqc(sArg, "-Repeat")) {
            assert(++ixArg < argc);
            NumRepeat = atoi(argv[ixArg]);
        }
        else {
            printf("Unrecognized option: '%s'\n", sArg);
            exit(1);
        }
    }

    if (PlogName) {
        PrfLog = fopen(PlogName, "w");
        //fprintf(PrfLog, "");
    }


    if (GloDbgLev > 100) printf(
        "\n\n##############################################################################\n\n");

    /* If we extend the solver to handle QBFs with free variables, then FreeFmla
     * would be a formula in terms of the free variables.  At present, FreeFmla
     * is either true or false. */
    FmlaT FreeFmla = -1;
    while (NumRepeat-- > 0) {
        /* Seed the random number generator. */
        srand(GloRandSeed);
        
        /* Read the input QBF file. */
        try {
            ParseFile(in_filename);
        } catch (char const* pErrMsg) {
            printf("%s\n", pErrMsg);
            fprintf(stderr, "%s\n", pErrMsg);
            exit(1);
        }

        StartTime = GetUserTimeMicro();
        if (AddPreprocTime) {
            StartTime -= (((long long) pGloProb->PreprocTimeMilli) * 1000);
        }
        //CalcLitToBat(pGloProb->OutGlit, 0);
        garr<GsT*>& clauses = pGloProb->clauses;
        for (int i=1; i < clauses.n; i++) {
            pGloAsgn->SimpFixWatch(clauses[i]);
        }
        //pGloAsgn->InitParents();

        QbFromLit[ZERO_LIT] = ZERO_DECLEV;
        DupLrnTest.clear();

        if (GloDbgLev >= 2020) {printf("OutGlit: %i\n", pGloProb->OutGlit);}

        if (!QuietMode) {
            printf("Seed: %2i. ", GloRandSeed);
            fflush(stdout);
        }
        if (GloDbgLev >= 1000) {printf("\n");}

        /* Call the solver procedure. */
        pGloAsgn->FreeFmla = NOT_DONE;
        pGloAsgn->PreSolve();
        if (pGloAsgn->HasConflict()) {
            FreeFmla = pGloAsgn->FreeFmla;
        } else {
            pGloAsgn->FreeFmla = NOT_DONE;
            FreeFmla = pGloAsgn->solve();
        }
        FixNeedNewl();
        double CpuTime = (GetUserTimeMicro() - StartTime) / 1000000.0;
        if (QuietMode) {
            if (!s_cnf) {
                printf("%s\n", (FreeFmla ? "true. " : "false."));
            }
        } else if (CondensedOut) {
            printf("%s Bt:%5i, Dec:%5lli.  T: %0.3f s.\n", 
                (FreeFmla ? "true. " : "false."), NumBigBt, NumDecs,
                ShowTime * (GetUserTimeMicro() - StartTime) / 1000000.0);
        } else {
            printf("%s Bt:%5i, D:%5lli.  R:%4i, P:%7lli, w:%9lli, C:%3i, T: %0.3f s.\n", 
                (FreeFmla ? "true. " : "false."), NumBigBt, NumDecs, pGloAsgn->NumRestarts,
                NumProps,
                NumWatchFixes,
                NumCegLrn,
                ShowTime * CpuTime);
        }
        if (s_cnf) {
            printf("s cnf %s\n", (FreeFmla ? "1" : "0"));
        }

        if (GloDbgLev >= 100) {
            pGloAsgn->PrintStats();
            if (GloDbgLev >= 2000) {
                pGloAsgn->dump(0);
            }
        }
        if (0 && GloDbgLev >= 100) {
            garri& TopLits = pGloAsgn->UndoListByDL[ZERO_DECLEV];
            printf("Top unit lits: [");
            for (int i=0; i < TopLits.n; i++) {
                int CurLit = TopLits[i];
                if (!IsInputLit(CurLit)) continue;
                if (i != 0) printf(", ");
                printf("%i(%c)", decv(CurLit), tolower(QtypeFromLit(CurLit)));
            }
            printf("]\n");
            /*if (pGloAsgn->HasDoneLitSet(0)) {
                printf("Done LitSet: ");
                DumpLitsEnc(pGloAsgn->pBestDoneLs->ReqLits);
            }
            printf("CurAsgn: ");
            pGloAsgn->chlits.dump();*/
        }
    }
    if (FreeFmla != true && FreeFmla != false) {
        return 30;
    }
    return (FreeFmla ? 10 : 20);

    return 0;

}
